#pragma once

#include <js/IJSEngine.h>
#include <js/IJSContext.h>
#include <window/IWindow.h>

class JSEngineFactory;
class CaptureDeviceFactory;
class WindowFactory;
class ARWindow;
class PluginManager;
class SourceFactory;

/**
* CEBhEB
*/
class MainWindow
{
public:

	/**
	* \z܂B
	*/
	MainWindow(HWND hWnd, JSEngineFactory* pJSEngineFactory, CaptureDeviceFactory* pCaptureDeviceFactory, WindowFactory* pWindowFactory, HRESULT* phr);

	/**
	* j܂B
	*/
	~MainWindow(void);

	/**
	* JavaScript̎sJn܂B
	*/
	HRESULT Execute(const char* path);

	/**
	* EBhEB
	*/
	HRESULT WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam, bool* pPreventDefault);

private:

	/**
	* EBhẼbp[B
	*/
	class WindowWrapper : public IWindow, public IJSContextNotify
	{
	public:

		/**
		* \z܂B
		*/
		WindowWrapper(MainWindow* pOwner, IWindow* pBase);

		/**
		* j܂B
		*/
		virtual ~WindowWrapper();

		/**
		* j܂B
		*/
		virtual void Destroy();

		/**
		* eEBhEnhݒ肵܂B
		*/
		virtual HRESULT SetParent(HWND hWnd);

		/**
		* ʒmIuWFNgݒ肵܂B
		*/
		virtual HRESULT SetNotify(IWindowNotify* pNotify);

		/**
		* EBhE𐶐܂B
		*/
		virtual HRESULT Create(int width, int height);

		/**
		* C[vJn܂B
		*/
		virtual HRESULT RunMainLoop();

		/**
		* ĕ`悵܂B
		*/
		virtual HRESULT Redraw();

		/**
		* GLp̃tbNǉ܂B
		*/
		virtual HRESULT AddGLHook(IWindowGLHook* pHook);

		/**
		* GLp̃tbN폜܂B
		*/
		virtual HRESULT RemoveGLHook(IWindowGLHook* pHook);

		/**
		* EBhE܂B
		*/
		virtual HRESULT Close();
		
		/**
		* R\[ɏoȂ܂B
		*/
		virtual HRESULT PrintLine(const char* line);
		
		/**
		* A[g\Ȃ܂B
		*/
		virtual HRESULT Alert(const char* message);
		
		/**
		* Xe[^XbZ[W\Ȃ܂B
		*/
		virtual HRESULT SetStatus(const char* status);

		/**
		* ݂̏|[g܂B
		*/
		virtual HRESULT ReportStatus(ostream* pOut);

		/**
		* A[g\Ȃ܂B
		*/
		virtual HRESULT OnAlert(const char* message);

		/**
		* G[\Ȃ܂B
		*/
		virtual HRESULT OnError(const char* message, const char* filename, unsigned int lineno, const char* token);
		
		/**
		* R\[ɏoȂ܂B
		*/
		virtual HRESULT OnPrintLine(const char* line);

		/**
		* `ݒ肵܂B
		*/
		virtual HRESULT SetRect(int x, int y, int width, int height);

		/**
		* `擾܂B
		*/
		virtual HRESULT GetRect(int* pX, int* pY, int* pWidth, int* pHeight);

		/**
		* ǂݍݗpGLReLXgJn܂B
		*/
		virtual HRESULT BeginGLLoading();

		/**
		* ǂݍݗpGLReLXgI܂B
		*/
		virtual HRESULT EndGLLoading();

	private:

		/**
		* EBhEnhێ܂B
		*/
		MainWindow* m_pOwner;

		/**
		* x[XIuWFNgێ܂B
		*/
		IWindow* m_pBase;

	};

	/**
	* ܂B
	*/
	HRESULT Init(JSEngineFactory* pJSEngineFactory);

	/**
	* ReLXgj܂B
	*/
	void DestroyContext();

	/**
	* TCYύX܂B
	*/
	HRESULT OnSize(int width, int height);

	/**
	* ʂ`悵܂B
	*/
	HRESULT OnPaint(HDC hDC);

	/**
	* TCYXV܂B
	*/
	HRESULT RefreshSize();

	/**
	* Rg[̋`擾܂B
	*/
	HRESULT GetControlRect(int width, int height, RECT* pPathEditRect, RECT* pPathButtonRect, RECT* pDebugEditRect);

	/**
	* XNvgs܂B
	*/
	HRESULT Evaluate(SourceFactory* pSourceFactory, const char* path);

	/**
	* fobOo͂Ȃ܂B
	*/
	void Debug(const char* message);

	/**
	* EBhE̒ۃNXB
	*/
	class AbstractWnd
	{
	public:

		/**
		* \z܂B
		*/
		AbstractWnd(HWND hParentWnd);

		/**
		* j܂B
		*/
		virtual ~AbstractWnd();

		/**
		* `擾܂B
		*/
		virtual HRESULT GetRect(RECT* pRect);

		/**
		* ړ܂B
		*/
		virtual HRESULT Move(const RECT* pRect, bool repaint);

		/**
		* EBhEnh擾܂B
		*/
		virtual HWND GetHWnd();

	protected:

		/**
		* ftHgtHgݒ肵܂B
		*/
		HRESULT SetDefaultFont();

		/**
		* eEBhEnhێ܂B
		*/
		HWND m_hParentWnd;

		/**
		* EBhEnhێ܂B
		*/
		HWND m_hWnd;
	};

	/**
	* {^EBhEB
	*/
	class ButtonWnd : public AbstractWnd
	{
	public:

		/**
		* \z܂B
		*/
		ButtonWnd(HWND hParentWnd);

		/**
		* j܂B
		*/
		virtual ~ButtonWnd();

		/**
		* 쐬܂B
		*/
		HRESULT Create(const char* name, const RECT* pRect, int id);
	};

	/**
	* eLXgEBhEB
	*/
	class TextWnd : public AbstractWnd
	{
	public:

		/**
		* \z܂B
		*/
		TextWnd(HWND hParentWnd);

		/**
		* j܂B
		*/
		virtual ~TextWnd();

		/**
		* sǂݒ肵܂B
		*/
		HRESULT SetMultiline(bool multiline);

		/**
		* 쐬܂B
		*/
		HRESULT Create(const char* defaultValue, const RECT* pRect, int id);

		/**
		* eLXg擾܂B
		*/
		HRESULT GetText(string& text);

		/**
		* eLXgݒ肵܂B
		*/
		HRESULT SetText(const char* message);

		/**
		* so͂܂B
		*/
		HRESULT PrintLine(const char* message);

	private:

		/**
		* sǂێ܂B
		*/
		bool m_bMultiline;

		/**
		* sێ܂B
		*/
		int m_iLines;

	};

	/**
	* ԃEBhEB
	*/
	class StatusWnd : public AbstractWnd
	{
	public:

		/**
		* \z܂B
		*/
		StatusWnd(HWND hParentWnd);

		/**
		* j܂B
		*/
		virtual ~StatusWnd();

		/**
		* 쐬܂B
		*/
		HRESULT Create(int size[], int count, int id);

		/**
		* Xe[^Xݒ肵܂B
		*/
		HRESULT SetText(int index, const char* message);
	};

	/**
	* EBhEnhێ܂B
	*/
	HWND m_hWnd;

	/**
	* JavaScript̃GWێ܂B
	*/
	IJSEngine* m_pJSEngine;

	/**
	* JavaScript̃ReLXgێ܂B
	*/
	IJSContext* m_pJSContext;

	/**
	* EBhEێ܂B
	*/
	ARWindow* m_pARWindow;

	/**
	* vOC}l[Wێ܂B
	*/
	PluginManager* m_pPluginManager;

	/**
	* EBhE(GL)ێ܂B
	*/
	IWindow* m_pGLWindow;

	/**
	* t@Ngێ܂B
	*/
	CaptureDeviceFactory* m_pCaptureDeviceFactory;

	/**
	* t@Ngێ܂B
	*/
	WindowFactory* m_pWindowFactory;

	/**
	* Xe[^ẌB
	*/
	StatusWnd* m_pStatusWnd;

	/**
	* pXҏẄB
	*/
	TextWnd* m_pPathEdit;

	/**
	* pX{^B
	*/
	ButtonWnd* m_pPathButton;

	/**
	* fobO\̈B
	*/
	TextWnd* m_pDebugEdit;

	/**
	* c[o[̃uVB
	*/
	HBRUSH m_hToolbarBrush;

	/**
	* rfIʂ̃uVB
	*/
	HBRUSH m_hVideoBrush;
};
