#include "StdAfx.h"
#include "MMDObject.h"
#include <plugin/ARGUtil.h>
#include <plugin/ARGCritSection.h>
#include "ExternalClock.h"
#include "MMDObjectFactory.h"

#define MEMBER_SET_FACE 1
#define MEMBER_PLAY 2
#define MEMBER_STOP 3
#define MEMBER_FACE 4
#define MEMBER_FACE_COUNT 5
#define MEMBER_IK_TARGET 6
#define MEMBER_IK_COUNT 7
#define MEMBER_SET_IK_ENABLED 8
#define MEMBER_IS_IK_ENABLED 9
#define MEMBER_BONE 10
#define MEMBER_BONE_COUNT 11
#define MEMBER_MAX_FRAME 12
#define MEMBER_CURRENT_FRAME 13
#define MEMBER_PLAYING 14
#define MEMBER_FPS 15
#define MEMBER_LOAD_VMD 16
#define MEMBER_SET_BONE_VISIBLE 17
#define MEMBER_IS_BONE_VISIBLE 18

#define HANDLER_VMD_LOADED "VMDLoaded"

#ifdef __cplusplus
extern "C" {
#endif
GLubyte*	mqoLoadTextureEx(char *texfile,char *alpfile,int *tex_width, int *tex_height, int* tex_mem_width, int* tex_mem_height, unsigned char alpha);
#ifdef __cplusplus
}
#endif

// t[[g`܂B
#define DEFAULT_FPS 24.0f

// \z܂B
MMDObject::MMDObject(MMDObjectFactory* pFactory, IARGPluginEnvironment* pEnv, HRESULT* phr)
{
	m_pFactory = pFactory;
	pFactory->AddRef();
	m_pEnv = pEnv;
	pEnv->AddRef();
	m_pModel = NULL;
	m_pUpdateWorker = NULL;
	m_pClock = NULL;
	m_pExtClock = NULL;
	m_fFps = DEFAULT_FPS;
	m_fOffset = 0.0f;

	HRESULT hr = CreateTypeInfo();
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
MMDObject::~MMDObject(void)
{
	if(m_pUpdateWorker != NULL) {
		HANDLE hExitEvent = NULL;
		m_pUpdateWorker->Stop(& hExitEvent);
		m_pUpdateWorker->Release();
		m_pUpdateWorker = NULL;

		ARGBaseWorker::WaitForDestroy(hExitEvent);
		hExitEvent = NULL;
	}
	if(m_pExtClock != NULL) {
		delete m_pExtClock;
		m_pExtClock = NULL;
	}
	if(m_pClock != NULL) {
		m_pClock->Release();
		m_pClock = NULL;
	}
	if(m_pEnv != NULL) {
		m_pEnv->Release();
		m_pEnv = NULL;
	}
	if(m_pFactory != NULL) {
		m_pFactory->Release();
		m_pFactory = NULL;
	}
	if(m_pModel != NULL) {
		delete m_pModel;
		m_pModel = NULL;
	}
	for(vector<IARGEventDispatcher*>::size_type i = 0; i < m_vmdLoadedDisps.size(); i ++) {
		IARGEventDispatcher* pDisp = m_vmdLoadedDisps[i];
		pDisp->Release();
	}
	m_vmdLoadedDisps.clear();
}

// ǂݏo܂B
HRESULT MMDObject::Load(const char* pmdPath, const char* vmdPath, double scale)
{
	if(pmdPath == NULL || vmdPath == NULL) {
		return E_POINTER;
	}

	MMDModel* pNewModel = new MMDModel();
	HRESULT hr = pNewModel->OpenPMD(pmdPath);
	if(FAILED(hr)) {
		return hr;
	}

	if(strlen(vmdPath) > 0) {
		hr = pNewModel->OpenVMD(vmdPath);
		if(FAILED(hr)) {
			return hr;
		}
	}

	hr = pNewModel->SetScale(scale);
	if(FAILED(hr)) {
		return hr;
	}

	TextureProviderImpl textureProv(pmdPath);

 	hr = pNewModel->Prepare(& textureProv);
	if(FAILED(hr)) {
		return hr;
	}
	m_pmdPath = pmdPath;
	m_vmdPath = vmdPath;

	m_pModel = pNewModel;
	return S_OK;
}

// `悵܂B
STDMETHODIMP MMDObject::Draw()
{
	if(m_pModel == NULL) {
		return S_OK;
	}

	return m_pModel->DrawAsync(m_pUpdateWorker);
}

// AChȌsȂ܂B
STDMETHODIMP MMDObject::OnIdle()
{
	if(m_pUpdateWorker == NULL) {
		return S_OK;
	}
	m_pUpdateWorker->SetIdle();
	return S_OK;
}

// IDispatch̎B
STDMETHODIMP MMDObject::Invoke(DISPID dispIdMember, REFIID riid, LCID  lcid, WORD  wFlags, DISPPARAMS *pDispParams, VARIANT *pVarResult, EXCEPINFO *pExcepInfo, unsigned int *puArgErr)
{
	if(m_pModel == NULL) {
		return E_UNEXPECTED;
	}
	if(dispIdMember == MEMBER_SET_FACE) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 1) {
			return E_INVALIDARG;
		}
		return E_NOTIMPL;
	}else if(dispIdMember == MEMBER_PLAY) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 1) {
			return E_INVALIDARG;
		}
		VARIANTARG* pArg = &(pDispParams->rgvarg[0]);
		if(pArg->vt != VT_R8) {
			return E_INVALIDARG;
		}

		m_fOffset = pArg->dblVal;
		StopUpdateWorker();

		HRESULT hr = m_clock.Play();
		if(FAILED(hr)) {
			return hr;
		}

		return StartUpdateWorker();
	}else if(dispIdMember == MEMBER_STOP) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		return m_clock.Stop();
	}else if(dispIdMember == MEMBER_FACE) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL || pVarResult == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 1) {
			return E_INVALIDARG;
		}
		VARIANTARG* pArg = &(pDispParams->rgvarg[0]);
		if(pArg->vt != VT_I4) {
			return E_INVALIDARG;
		}
		string name = "";
		HRESULT hr = m_pModel->GetFaceName(pArg->intVal, name);
		if(FAILED(hr)) {
			return hr;
		}
		BSTR ret = NULL;
		hr = ARGUtil::GetBSTRFromString(& ret, name.c_str());
		if(FAILED(hr)) {
			return hr;
		}
		
		pVarResult->vt = VT_BSTR;
		pVarResult->bstrVal = ret;
		return S_OK;

	}else if(dispIdMember == MEMBER_FACE_COUNT) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			int count = 0;
			HRESULT hr = m_pModel->GetFaceCount(& count);
			if(FAILED(hr)) {
				return hr;
			}
 
			pVarResult->vt = VT_I4;
			pVarResult->intVal = count;
			return S_OK;
		}else{
			return E_UNEXPECTED;
		}
	}else if(dispIdMember == MEMBER_IK_TARGET) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL || pVarResult == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 1) {
			return E_INVALIDARG;
		}
		VARIANTARG* pArg = &(pDispParams->rgvarg[0]);
		if(pArg->vt != VT_I4) {
			return E_INVALIDARG;
		}
		string name = "";
		HRESULT hr = m_pModel->GetIKTargetName(pArg->intVal, name);
		if(FAILED(hr)) {
			return hr;
		}
		BSTR ret = NULL;
		hr = ARGUtil::GetBSTRFromString(& ret, name.c_str());
		if(FAILED(hr)) {
			return hr;
		}
		
		pVarResult->vt = VT_BSTR;
		pVarResult->bstrVal = ret;
		return S_OK;

	}else if(dispIdMember == MEMBER_IK_COUNT) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			int count = 0;
			HRESULT hr = m_pModel->GetIKCount(& count);
			if(FAILED(hr)) {
				return hr;
			}
 
			pVarResult->vt = VT_I4;
			pVarResult->intVal = count;
			return S_OK;
		}else{
			return E_UNEXPECTED;
		}
	}else if(dispIdMember == MEMBER_SET_IK_ENABLED) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL || pVarResult == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 2) {
			return E_INVALIDARG;
		}
		VARIANTARG* pIndexArg = &(pDispParams->rgvarg[0]);
		if(pIndexArg->vt != VT_I4) {
			return E_INVALIDARG;
		}
		VARIANTARG* pValueArg = &(pDispParams->rgvarg[1]);
		if(pValueArg->vt != VT_BOOL) {
			return E_INVALIDARG;
		}

		bool val = false;
		if(pValueArg->boolVal) {
			val = true;
		}
		return m_pModel->SetIKEnabled(pIndexArg->intVal, val);

	}else if(dispIdMember == MEMBER_IS_IK_ENABLED) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL || pVarResult == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 1) {
			return E_INVALIDARG;
		}
		VARIANTARG* pArg = &(pDispParams->rgvarg[0]);
		if(pArg->vt != VT_I4) {
			return E_INVALIDARG;
		}
		bool value = false;
		HRESULT hr = m_pModel->IsIKEnabled(pArg->intVal, & value);
		if(FAILED(hr)) {
			return hr;
		}
		pVarResult->vt = VT_BOOL;
		pVarResult->boolVal = value;
		return S_OK;

	}else if(dispIdMember == MEMBER_BONE) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL || pVarResult == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 1) {
			return E_INVALIDARG;
		}
		VARIANTARG* pArg = &(pDispParams->rgvarg[0]);
		if(pArg->vt != VT_I4) {
			return E_INVALIDARG;
		}
		string name = "";
		HRESULT hr = m_pModel->GetBoneName(pArg->intVal, name);
		if(FAILED(hr)) {
			return hr;
		}
		BSTR ret = NULL;
		hr = ARGUtil::GetBSTRFromString(& ret, name.c_str());
		if(FAILED(hr)) {
			return hr;
		}
		
		pVarResult->vt = VT_BSTR;
		pVarResult->bstrVal = ret;
		return S_OK;

	}else if(dispIdMember == MEMBER_BONE_COUNT) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			int count = 0;
			HRESULT hr = m_pModel->GetBoneCount(& count);
			if(FAILED(hr)) {
				return hr;
			}
 
			pVarResult->vt = VT_I4;
			pVarResult->intVal = count;
			return S_OK;
		}else{
			return E_UNEXPECTED;
		}
	}else if(dispIdMember == MEMBER_MAX_FRAME) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			int frameNo = 0;
			HRESULT hr = m_pModel->GetMaxFrame(& frameNo);
			if(FAILED(hr)) {
				return hr;
			}
 
			pVarResult->vt = VT_R8;
			pVarResult->dblVal = frameNo;
			return S_OK;
		}else{
			return E_UNEXPECTED;
		}
	}else if(dispIdMember == MEMBER_CURRENT_FRAME) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			float frameNo = 0.0f;
			if(m_pUpdateWorker != NULL) {
				HRESULT hr = m_pUpdateWorker->GetLastFrame(& frameNo);
				if(FAILED(hr)) {
					return hr;
				}
			}
 
			pVarResult->vt = VT_R8;
			pVarResult->dblVal = frameNo;
			return S_OK;
		}else{
			return E_UNEXPECTED;
		}
	}else if(dispIdMember == MEMBER_PLAYING) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			bool val = false;
			HRESULT hr = m_clock.IsPlaying(& val);
			if(FAILED(hr)) {
				return hr;
			}
 
			pVarResult->vt = VT_BOOL;
			pVarResult->boolVal = val;
			return S_OK;
		}else{
			return E_UNEXPECTED;
		}
	}else if(dispIdMember == MEMBER_FPS) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			if(pVarResult == NULL) {
				return E_POINTER;
			}
			pVarResult->vt = VT_R8;
			pVarResult->dblVal = m_fFps;
			return S_OK;

		}else if(wFlags == DISPATCH_PROPERTYPUT) {
			if(pDispParams == NULL) {
				return E_POINTER;
			}
			if(pDispParams->cArgs < 1) {
				return E_INVALIDARG;
			}
			VARIANTARG* pArg = &(pDispParams->rgvarg[0]);
			if(pArg->vt != VT_R8) {
				return E_INVALIDARG;
			}

			m_fFps = pArg->dblVal;
			return RestartUpdateWorker();

		}else{
			return E_UNEXPECTED;
		}
	}else if(dispIdMember == MEMBER_LOAD_VMD) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL || pVarResult == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 1) {
			return E_INVALIDARG;
		}
		VARIANTARG* pArg = &(pDispParams->rgvarg[0]);
		if(pArg->vt != VT_BSTR) {
			return E_INVALIDARG;
		}
		string path = "";
		HRESULT hr = ARGUtil::GetStringFromBSTR(path, pArg->bstrVal);
		if(FAILED(hr)) {
			return hr;
		}

		string absPath = "";
		hr = m_pFactory->GetAbsolutePath(path.c_str(), absPath);
		if(FAILED(hr)) {
			return hr;
		}

		StopUpdateWorker();
		hr = m_pModel->ClearMotion();
		if(FAILED(hr)) {
			return hr;
		}

		m_vmdPath = absPath;

		ARGCritSection cs(& m_vmdLoadedDispsMutex);
		LoadVMDWorker* pWorker = new LoadVMDWorker(m_pModel, absPath.c_str(), m_vmdLoadedDisps);
		hr = m_pEnv->InvokeWorker(pWorker);
		pWorker->Release();
		pWorker = NULL;
		return hr;

	}else if(dispIdMember == MEMBER_SET_BONE_VISIBLE) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL || pVarResult == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 2) {
			return E_INVALIDARG;
		}
		VARIANTARG* pIndexArg = &(pDispParams->rgvarg[0]);
		if(pIndexArg->vt != VT_I4) {
			return E_INVALIDARG;
		}
		VARIANTARG* pValueArg = &(pDispParams->rgvarg[1]);
		if(pValueArg->vt != VT_BOOL) {
			return E_INVALIDARG;
		}

		bool val = false;
		if(pValueArg->boolVal) {
			val = true;
		}
		return m_pModel->SetBoneVisible(pIndexArg->intVal, val);

	}else if(dispIdMember == MEMBER_IS_BONE_VISIBLE) {
		if(wFlags != DISPATCH_METHOD) {
			return E_UNEXPECTED;
		}
		if(pDispParams == NULL || pVarResult == NULL) {
			return E_POINTER;
		}
		if(pDispParams->cArgs < 1) {
			return E_INVALIDARG;
		}
		VARIANTARG* pArg = &(pDispParams->rgvarg[0]);
		if(pArg->vt != VT_I4) {
			return E_INVALIDARG;
		}
		bool value = false;
		HRESULT hr = m_pModel->IsBoneVisible(pArg->intVal, & value);
		if(FAILED(hr)) {
			return hr;
		}
		pVarResult->vt = VT_BOOL;
		pVarResult->boolVal = value;
		return S_OK;

	}else{
		return E_UNEXPECTED;
	}
	return E_NOTIMPL;
}

// pNbNݒ肵܂B
STDMETHODIMP MMDObject::SetSyncClock(IARGSyncClock* pClock)
{
	if(pClock != NULL) {
		pClock->AddRef();
	}
	if(m_pClock != NULL) {
		m_pClock->Release();
		m_pClock = NULL;
	}
	m_pClock = pClock;

	return RestartUpdateWorker();
}

// pNbN擾܂B
STDMETHODIMP MMDObject::GetSyncClock(IARGSyncClock** ppClock)
{
	if(ppClock == NULL) {
		return E_POINTER;
	}
	*ppClock = m_pClock;
	if(m_pClock != NULL) {
		m_pClock->AddRef();
	}
	return S_OK;
}

// Cxg擾܂B
STDMETHODIMP MMDObject::GetEventCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = 1;
	return S_OK;
}

// Cxg擾܂B
STDMETHODIMP MMDObject::GetEvent(int index, char* pEventName, int maxLen, int* pLen)
{
	if(index != 0) {
		return E_INVALIDARG;
	}
	string name = HANDLER_VMD_LOADED;
	if(pEventName == NULL) {
		*pLen = name.length();
		return S_OK;
	}

	strcpy_s(pEventName, maxLen, name.c_str());
	if((int)name.length() < maxLen) {
		pEventName[name.length()] = '\0';
	}
	return S_OK;
}

// CxgfBXpb`ǉ܂B
STDMETHODIMP MMDObject::AddEventDispatcher(const char* name, IARGEventDispatcher* pDispatcher)
{
	if(name == NULL || pDispatcher == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname != HANDLER_VMD_LOADED) {
		return E_INVALIDARG;
	}
	ARGCritSection cs(& m_vmdLoadedDispsMutex);
	m_vmdLoadedDisps.push_back(pDispatcher);
	pDispatcher->AddRef();
	return S_OK;
}

// CxgfBXpb`폜܂B
STDMETHODIMP MMDObject::RemoveEventDispatcher(const char* name, IARGEventDispatcher* pDispatcher)
{
	string sname = name;
	if(sname != HANDLER_VMD_LOADED) {
		return E_INVALIDARG;
	}
	if(pDispatcher != NULL) {
		return E_NOTIMPL;
	}
	ARGCritSection cs(& m_vmdLoadedDispsMutex);
	for(vector<IARGEventDispatcher*>::size_type i = 0; i < m_vmdLoadedDisps.size(); i ++) {
		IARGEventDispatcher* pDisp = m_vmdLoadedDisps[i];
		pDisp->Release();
	}
	m_vmdLoadedDisps.clear();
	return S_OK;
}

// ITypeInfo𐶐܂B
HRESULT MMDObject::CreateTypeInfo()
{
	vector<VARTYPE> emptyp;
	ARGTypeInfo* pTypeInfo = new ARGTypeInfo(IID_IARGObject);
	vector<VARTYPE> setFacep;
	setFacep.push_back(VT_BSTR);
	HRESULT hr = pTypeInfo->AddElement(L"setFace", MEMBER_SET_FACE, INVOKE_FUNC, VT_VOID, setFacep);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	vector<VARTYPE> playp;
	playp.push_back(VT_R8);
	hr = pTypeInfo->AddElement(L"play", MEMBER_PLAY, INVOKE_FUNC, VT_VOID, playp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"stop", MEMBER_STOP, INVOKE_FUNC, VT_VOID, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	vector<VARTYPE> indexp;
	indexp.push_back(VT_I4);
	hr = pTypeInfo->AddElement(L"face", MEMBER_FACE, INVOKE_FUNC, VT_BSTR, indexp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"faceCount", MEMBER_FACE_COUNT, INVOKE_PROPERTYGET, VT_I4, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"ikTarget", MEMBER_IK_TARGET, INVOKE_FUNC, VT_BSTR, indexp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"ikCount", MEMBER_IK_COUNT, INVOKE_PROPERTYGET, VT_I4, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	vector<VARTYPE> setEnabledp;
	setEnabledp.push_back(VT_I4);
	setEnabledp.push_back(VT_BOOL);
	hr = pTypeInfo->AddElement(L"setIKEnabled", MEMBER_SET_IK_ENABLED, INVOKE_FUNC, VT_VOID, setEnabledp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"isIKEnabled", MEMBER_IS_IK_ENABLED, INVOKE_FUNC, VT_BOOL, indexp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"bone", MEMBER_BONE, INVOKE_FUNC, VT_BSTR, indexp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"boneCount", MEMBER_BONE_COUNT, INVOKE_PROPERTYGET, VT_I4, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"maxFrame", MEMBER_MAX_FRAME, INVOKE_PROPERTYGET, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"currentFrame", MEMBER_CURRENT_FRAME, INVOKE_PROPERTYGET, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"playing", MEMBER_PLAYING, INVOKE_PROPERTYGET, VT_BOOL, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"fps", MEMBER_FPS, INVOKE_PROPERTYGET, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"fps", MEMBER_FPS, INVOKE_PROPERTYPUT, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	vector<VARTYPE> loadp;
	loadp.push_back(VT_BSTR);
	hr = pTypeInfo->AddElement(L"loadVMD", MEMBER_LOAD_VMD, INVOKE_FUNC, VT_VOID, loadp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"setBoneVisible", MEMBER_SET_BONE_VISIBLE, INVOKE_FUNC, VT_VOID, setEnabledp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"isBoneVisible", MEMBER_IS_BONE_VISIBLE, INVOKE_FUNC, VT_BOOL, indexp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	m_pTypeInfo = pTypeInfo;
	return S_OK;
}

// [J[XV܂B
HRESULT MMDObject::RestartUpdateWorker()
{
	StopUpdateWorker();

	return StartUpdateWorker();
}

// [J[~܂B
void MMDObject::StopUpdateWorker()
{
	if(m_pUpdateWorker != NULL) {
		HANDLE hExitEvent = NULL;
		m_pUpdateWorker->Stop(& hExitEvent);
		m_pUpdateWorker->Release();
		m_pUpdateWorker = NULL;

		ARGBaseWorker::WaitForDestroy(hExitEvent);
		hExitEvent = NULL;
	}
	if(m_pExtClock != NULL) {
		delete m_pExtClock;
		m_pExtClock = NULL;
	}
}

// [J[Jn܂B
HRESULT MMDObject::StartUpdateWorker()
{
	IClock* pUClock = & m_clock;
	if(m_pClock != NULL) {
		m_pExtClock = new ExternalClock(m_pClock);
		pUClock = m_pExtClock;
	}
	m_pUpdateWorker = new UpdateWorker(m_pModel, pUClock, m_fFps, m_fOffset);
	return m_pEnv->InvokeWorker(m_pUpdateWorker);
}

// \z܂B
MMDObject::TextureProviderImpl::TextureProviderImpl(const char* pmdPath)
{
	string pmdSource = pmdPath;
	string::size_type spos = pmdSource.find_last_of('/');
	string::size_type bpos = pmdSource.find_last_of('\\');
	string base_path;

	if(spos == string::npos && bpos == string::npos) {
		base_path = "";
	}else if(spos != string::npos && bpos != string::npos) {
		if(spos < bpos) {
			base_path = pmdSource.substr(0, bpos + 1);
		}else{
			base_path = pmdSource.substr(0, spos + 1);
		}
	}else if(spos != string::npos) {
		base_path = pmdSource.substr(0, spos + 1);
	}else if(bpos != string::npos) {
		base_path = pmdSource.substr(0, bpos + 1);
	}
	m_basePath = base_path;
}

// j܂B
MMDObject::TextureProviderImpl::~TextureProviderImpl()
{
	;
}

// ǂݏo܂B
HRESULT MMDObject::TextureProviderImpl::Load(const char* filename, TEXTURE_DESC* pDesc)
{
	if(filename == NULL || pDesc == NULL) {
		return E_POINTER;
	}
	GLubyte *image;
	string texpath = m_basePath;
	texpath += filename;
	char stexpath[MAX_PATH + 2];
	strcpy_s(stexpath, MAX_PATH, texpath.c_str());
	for(int i = strlen(stexpath) - 1; i >= 0; i --) {
		if(stexpath[i] == '*') {
			stexpath[i] = '\0';
			break;
		}
	}
	image = mqoLoadTextureEx(stexpath, NULL, &pDesc->tex_width, &pDesc->tex_height, &pDesc->tex_mem_width, &pDesc->tex_mem_height, 255);

	glPixelStorei(GL_UNPACK_ALIGNMENT,4);
	glPixelStorei(GL_PACK_ALIGNMENT,4);
	glGenTextures(1,&pDesc->texture_id);			// eNX`𐶐
	glBindTexture(GL_TEXTURE_2D,pDesc->texture_id);	// eNX`̊蓖

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, pDesc->tex_mem_width, pDesc->tex_mem_height,
					0, GL_RGBA, GL_UNSIGNED_BYTE, image);

	//o^΁Aǂݍ񂾃obt@͕sv
	free(image);
	glBindTexture(GL_TEXTURE_2D,0);	// ftHgeNX`̊蓖
	return S_OK;
}

// \z܂B
MMDObject::UpdateWorker::UpdateWorker(MMDModel* pModel, IClock* pClock, float fps, float offset)
{
	m_pModel = pModel;
	m_pClock = pClock;
	m_hIdleEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	m_fFps = fps;
	m_fOffset = offset;
	m_fLastFrame = m_fOffset;
}

// j܂B
MMDObject::UpdateWorker::~UpdateWorker()
{
	m_pModel = NULL;
	if(m_hIdleEvent != NULL) {
		CloseHandle(m_hIdleEvent);
		m_hIdleEvent = NULL;
	}
}

// s܂B
STDMETHODIMP MMDObject::UpdateWorker::Invoke()
{
	while(IsContinue()) {
		WaitForSingleObject(m_hIdleEvent, 5000);
		float frameNo = 0.0f;
		{
			ARGCritSection cs(& m_clockMutex);
			if(m_pClock == NULL) {
				continue;
			}
			LONGLONG ms = 0;
			HRESULT hr = m_pClock->GetCurrentTimeMS(& ms);
			if(FAILED(hr)) {
				return hr;
			}
			double dsec = ((double)ms) / 1000.0; 
			frameNo = (float)(dsec * m_fFps + m_fOffset);
			m_fLastFrame = frameNo;
		}
		HRESULT hr = m_pModel->UpdateAsync(this, frameNo);
		if(FAILED(hr)) {
			return hr;
		}
	}
	return S_OK;
}

// NeBJZNVJn܂B
void MMDObject::UpdateWorker::Begin()
{
	m_modelMutex.Begin();
}

// NeBJZNVI܂B
void MMDObject::UpdateWorker::End()
{
	m_modelMutex.End();
}

// Ōɏt[ԍ擾܂B
HRESULT MMDObject::UpdateWorker::GetLastFrame(float* pLastFrame)
{
	if(pLastFrame == NULL) {
		return E_POINTER;
	}
	ARGCritSection cs(& m_clockMutex);
	*pLastFrame = m_fLastFrame;
	return S_OK;
}

// ~܂B
HRESULT MMDObject::UpdateWorker::Stop(HANDLE* hExitEvent)
{
	HRESULT hr = ARGBaseWorker::Stop(hExitEvent);
	if(FAILED(hr)) {
		return hr;
	}
	{
		ARGCritSection cs(& m_clockMutex);
		m_pClock = NULL;
	}
	SetEvent(m_hIdleEvent);
	return S_OK;
}

// AChԂł邱Ƃݒ肵܂B
void MMDObject::UpdateWorker::SetIdle()
{
	SetEvent(m_hIdleEvent);
}

// \z܂B
MMDObject::LoadVMDWorker::LoadVMDWorker(MMDModel* pModel, const char* path, vector<IARGEventDispatcher*>& disps)
{
	m_pModel = pModel;
	m_vmdPath = path;
	for(vector<IARGEventDispatcher*>::size_type i = 0; i < disps.size(); i ++) {
		IARGEventDispatcher* pDisp = disps[i];
		pDisp->AddRef();
		m_vmdLoadedDisps.push_back(pDisp);
	}
}

// j܂B
MMDObject::LoadVMDWorker::~LoadVMDWorker()
{
	for(vector<IARGEventDispatcher*>::size_type i = 0; i < m_vmdLoadedDisps.size(); i ++) {
		IARGEventDispatcher* pDisp = m_vmdLoadedDisps[i];
		pDisp->Release();
	}
	m_vmdLoadedDisps.clear();
}

// s܂B
STDMETHODIMP MMDObject::LoadVMDWorker::Invoke()
{
	HRESULT hr = m_pModel->OpenVMD(m_vmdPath.c_str());
	VARIANT vars[2];
	VariantInit(vars + 0);
	VariantInit(vars + 1);

	vars[0].vt = VT_I4;
	vars[0].intVal = hr;
	vars[1].vt = VT_BSTR;
	BSTR bstr = NULL;
	hr = ARGUtil::GetBSTRFromString(& bstr, m_vmdPath.c_str());
	if(FAILED(hr)) {
		return hr;
	}
	vars[1].bstrVal = bstr;

	for(vector<IARGEventDispatcher*>::size_type i = 0; i < m_vmdLoadedDisps.size(); i ++) {
		IARGEventDispatcher* pDisp = m_vmdLoadedDisps[i];
		hr = pDisp->DispatchEvent(vars, 2);
		if(FAILED(hr)) {
			return hr;
		}
	}

	VariantClear(vars + 0);
	VariantClear(vars + 1);
	return S_OK;
}
