#include "StdAfx.h"
#include "MMDObjectFactory.h"
#include <plugin/ARGUtil.h>
#include "MMDObject.h"

// \z܂B
MMDObjectFactory::MMDObjectFactory(IARGPluginEnvironment* pEnv)
{
	m_pEnv = pEnv;
	pEnv->AddRef();
}

// j܂B
MMDObjectFactory::~MMDObjectFactory(void)
{
	if(m_pEnv != NULL) {
		m_pEnv->Release();
		m_pEnv = NULL;
	}
}

// IuWFNgOpenGLgp邩ǂ擾܂B
STDMETHODIMP MMDObjectFactory::UseGL(bool* pValue)
{
	if(pValue == NULL) {
		return E_POINTER;
	}
	*pValue = true;
	return S_OK;
};

// p[^^Cv̌␔擾܂B
STDMETHODIMP MMDObjectFactory::GetParameterTypeCandidateCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = 1;
	return S_OK;
}

// p[^^Cv̌₲Ƃ̈̐擾܂B
STDMETHODIMP MMDObjectFactory::GetParameterTypeCount(int candIndex, int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	if(candIndex == 0) {
		*pCount = 3;
		return S_OK;
	}else{
		return E_INVALIDARG;
	}
}

// p[^^Cv̌₲Ƃ̈^Cv擾܂B
STDMETHODIMP MMDObjectFactory::GetParameterType(int candIndex, int index, VARTYPE* pVarType)
{
	if(pVarType == NULL) {
		return E_POINTER;
	}
	if(candIndex == 0) {
		if(index == 0) {
			*pVarType = VT_BSTR;
			return S_OK;
		}else if(index == 1) {
			*pVarType = VT_BSTR;
			return S_OK;
		}else if(index == 2) {
			*pVarType = VT_R8;
			return S_OK;
		}else{
			return E_INVALIDARG;
		}
	}else{
		return E_INVALIDARG;
	}
}

// IuWFNg𐶐܂B
STDMETHODIMP MMDObjectFactory::Create(VARIANT* pVars, int varCount, IARGObject** ppObject)
{
	if(pVars == NULL || ppObject == NULL) {
		return E_POINTER;
	}
	if(varCount != 3) {
		return E_INVALIDARG;
	}
	VARIANT* pVPMDPath = pVars + 0;
	if(pVPMDPath->vt != VT_BSTR || pVPMDPath->bstrVal == NULL) {
		return E_INVALIDARG;
	}
	string pmdPath;
	HRESULT hr = ARGUtil::GetStringFromBSTR(pmdPath, pVPMDPath->bstrVal);
	if(FAILED(hr)) {
		return hr;
	}

	VARIANT* pVVMDPath = pVars + 1;
	if(pVVMDPath->vt != VT_BSTR && pVVMDPath->vt != VT_EMPTY && pVVMDPath->vt != VT_NULL) {
		return E_INVALIDARG;
	}
	string vmdPath;
	if(pVVMDPath->bstrVal != NULL && pVVMDPath->vt == VT_BSTR) {
		hr = ARGUtil::GetStringFromBSTR(vmdPath, pVVMDPath->bstrVal);
		if(FAILED(hr)) {
			return hr;
		}
	}

	VARIANT* pVScale = pVars + 2;
	if(pVScale->vt != VT_R8) {
		return E_INVALIDARG;
	}
	double scale = pVScale->dblVal;

	string absPMDPath;
	hr = GetAbsolutePath(pmdPath.c_str(), absPMDPath);
	if(FAILED(hr)) {
		return hr;
	}

	string absVMDPath;
	if(vmdPath.length() > 0) {
		hr = GetAbsolutePath(vmdPath.c_str(), absVMDPath);
		if(FAILED(hr)) {
			return hr;
		}
	}

	hr = S_OK;
	MMDObject* pObject = new MMDObject(this, m_pEnv, & hr);
	if(FAILED(hr)) {
		delete pObject;
		return hr;
	}

	hr = pObject->Load(absPMDPath.c_str(), absVMDPath.c_str(), scale);
	if(FAILED(hr)) {
		delete pObject;
		return hr;
	}

	*ppObject = pObject;
	return S_OK;
}

// ΃pX擾܂B
HRESULT MMDObjectFactory::GetAbsolutePath(const char* path, string& absolutePath)
{
	if(path == NULL) {
		return E_POINTER;
	}

	int maxLen = 0;
	HRESULT hr = m_pEnv->GetRealPath(path, NULL, 0, & maxLen);
	if(FAILED(hr)) {
		return hr;
	}

	// I[Ŋm
	char* pAbsolutePath = new char[maxLen + 1];
	ZeroMemory(pAbsolutePath, sizeof(char) * (maxLen + 1));

	hr = m_pEnv->GetRealPath(path, pAbsolutePath, maxLen + 1, & maxLen);
	if(FAILED(hr)) {
		delete[] pAbsolutePath;
		pAbsolutePath = NULL;
		return hr;
	}
	pAbsolutePath[maxLen] = '\0';
	absolutePath = pAbsolutePath;
	delete[] pAbsolutePath;
	pAbsolutePath = NULL;

	return S_OK;
}
