#include "StdAfx.h"
#include "VRMLObject.h"
#include <AR/arvrml.h>

#define MEMBER_X 1
#define MEMBER_Y 2
#define MEMBER_Z 3

// \z܂B
VRMLObject::VRMLObject(IARGPluginEnvironment* pEnv, HRESULT* phr)
{
	m_pEnv = pEnv;
	pEnv->AddRef();
	m_z = 0.0;
	m_x = 0.0;
	m_y = 0.0;
	m_iVRMLId = -1;

	HRESULT hr = CreateTypeInfo();
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
VRMLObject::~VRMLObject(void)
{
	if(m_pEnv != NULL) {
		m_pEnv->Release();
		m_pEnv = NULL;
	}
	if(m_iVRMLId >= 0) {
		arVrmlFree(m_iVRMLId);
		m_iVRMLId = -1;
	}
}

// ǂݏo܂B
HRESULT VRMLObject::Load(const char* path)
{
	if(path == NULL) {
		return E_POINTER;
	}
	m_path = path;

	m_iVRMLId = arVrmlLoadFile(path);
	if(m_iVRMLId < 0) {
		return E_UNEXPECTED;
	}
	return S_OK;
}

// `悵܂B
STDMETHODIMP VRMLObject::Draw()
{
	if(m_iVRMLId < 0) {
		return S_OK;
	}

	glPushMatrix();
    glTranslatef(m_x, m_y, m_z);

	arVrmlDraw(m_iVRMLId);

	glPopMatrix();
	arVrmlTimerUpdate();
	return S_OK;
}

// AChȌsȂ܂B
STDMETHODIMP VRMLObject::OnIdle()
{
	return S_OK;
}

// IDispatch̎B
STDMETHODIMP VRMLObject::Invoke(DISPID dispIdMember, REFIID riid, LCID  lcid, WORD  wFlags, DISPPARAMS *pDispParams, VARIANT *pVarResult, EXCEPINFO *pExcepInfo, unsigned int *puArgErr)
{
	if(dispIdMember == MEMBER_X) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			pVarResult->vt = VT_R8;
			pVarResult->dblVal = m_x;
		}else if(wFlags == DISPATCH_PROPERTYPUT) {
			if(pDispParams == NULL) {
				return E_POINTER;
			}
			if(pDispParams->cArgs < 1) {
				return E_INVALIDARG;
			}
			VARIANTARG* pParamVar = & pDispParams->rgvarg[0];
			if(pParamVar->vt != VT_R8) {
				return E_INVALIDARG;
			}
			m_x = pParamVar->dblVal;
		}else{
			return E_UNEXPECTED;
		}
	}else if(dispIdMember == MEMBER_Y) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			pVarResult->vt = VT_R8;
			pVarResult->dblVal = m_y;
		}else if(wFlags == DISPATCH_PROPERTYPUT) {
			if(pDispParams == NULL) {
				return E_POINTER;
			}
			if(pDispParams->cArgs < 1) {
				return E_INVALIDARG;
			}
			VARIANTARG* pParamVar = & pDispParams->rgvarg[0];
			if(pParamVar->vt != VT_R8) {
				return E_INVALIDARG;
			}
			m_y = pParamVar->dblVal;
		}else{
			return E_UNEXPECTED;
		}
	}else if(dispIdMember == MEMBER_Z) {
		if(wFlags == DISPATCH_PROPERTYGET) {
			pVarResult->vt = VT_R8;
			pVarResult->dblVal = m_z;
		}else if(wFlags == DISPATCH_PROPERTYPUT) {
			if(pDispParams == NULL) {
				return E_POINTER;
			}
			if(pDispParams->cArgs < 1) {
				return E_INVALIDARG;
			}
			VARIANTARG* pParamVar = & pDispParams->rgvarg[0];
			if(pParamVar->vt != VT_R8) {
				return E_INVALIDARG;
			}
			m_z = pParamVar->dblVal;
		}else{
			return E_UNEXPECTED;
		}
	}else{
		return E_UNEXPECTED;
	}
	return S_OK;
}

// ITypeInfo𐶐܂B
HRESULT VRMLObject::CreateTypeInfo()
{
	vector<VARTYPE> emptyp;
	ARGTypeInfo* pTypeInfo = new ARGTypeInfo(IID_IARGObject);
	HRESULT hr = pTypeInfo->AddElement(L"x", MEMBER_X, INVOKE_PROPERTYGET, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"y", MEMBER_Y, INVOKE_PROPERTYGET, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"z", MEMBER_Z, INVOKE_PROPERTYGET, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"x", MEMBER_X, INVOKE_PROPERTYPUT, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"y", MEMBER_Y, INVOKE_PROPERTYPUT, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	hr = pTypeInfo->AddElement(L"z", MEMBER_Z, INVOKE_PROPERTYPUT, VT_R8, emptyp);
	if(FAILED(hr)) {
		delete pTypeInfo;
		pTypeInfo = NULL;
		return hr;
	}

	m_pTypeInfo = pTypeInfo;
	return S_OK;
}
