#include "StdAfx.h"
#include "VRMLObjectFactory.h"
#include <plugin/ARGUtil.h>
#include "VRMLObject.h"

// \z܂B
VRMLObjectFactory::VRMLObjectFactory(IARGPluginEnvironment* pEnv)
{
	m_pEnv = pEnv;
	pEnv->AddRef();
}

// j܂B
VRMLObjectFactory::~VRMLObjectFactory(void)
{
	if(m_pEnv != NULL) {
		m_pEnv->Release();
		m_pEnv = NULL;
	}
}

// p[^^Cv̌␔擾܂B
STDMETHODIMP VRMLObjectFactory::GetParameterTypeCandidateCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = 1;
	return S_OK;
}

// p[^^Cv̌₲Ƃ̈̐擾܂B
STDMETHODIMP VRMLObjectFactory::GetParameterTypeCount(int candIndex, int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	if(candIndex == 0) {
		*pCount = 1;
		return S_OK;
	}else{
		return E_INVALIDARG;
	}
}

// p[^^Cv̌₲Ƃ̈^Cv擾܂B
STDMETHODIMP VRMLObjectFactory::GetParameterType(int candIndex, int index, VARTYPE* pVarType)
{
	if(pVarType == NULL) {
		return E_POINTER;
	}
	if(candIndex == 0) {
		if(index == 0) {
			*pVarType = VT_BSTR;
			return S_OK;
		}else{
			return E_INVALIDARG;
		}
	}else{
		return E_INVALIDARG;
	}
}

// IuWFNg𐶐܂B
STDMETHODIMP VRMLObjectFactory::Create(VARIANT* pVars, int varCount, IARGObject** ppObject)
{
	if(pVars == NULL || ppObject == NULL) {
		return E_POINTER;
	}
	if(varCount != 1) {
		return E_INVALIDARG;
	}
	VARIANT* pVPath = pVars;
	if(pVPath->vt != VT_BSTR || pVPath->bstrVal == NULL) {
		return E_INVALIDARG;
	}

	string path;
	HRESULT hr = ARGUtil::GetStringFromBSTR(path, pVPath->bstrVal);
	if(FAILED(hr)) {
		return hr;
	}

	int maxLen = 0;
	hr = m_pEnv->GetRealPath(path.c_str(), NULL, 0, & maxLen);
	if(FAILED(hr)) {
		return hr;
	}

	// I[m
	char* pAbsolutePath = new char[maxLen + 1];
	ZeroMemory(pAbsolutePath, sizeof(char) * (maxLen + 1));

	hr = m_pEnv->GetRealPath(path.c_str(), pAbsolutePath, maxLen + 1, & maxLen);
	if(FAILED(hr)) {
		delete[] pAbsolutePath;
		pAbsolutePath = NULL;
		return hr;
	}
	pAbsolutePath[maxLen] = '\0';
	path = pAbsolutePath;
	delete[] pAbsolutePath;
	pAbsolutePath = NULL;

	hr = S_OK;
	VRMLObject* pObject = new VRMLObject(m_pEnv, & hr);
	if(FAILED(hr)) {
		delete pObject;
		return hr;
	}
	hr = pObject->Load(path.c_str());
	if(FAILED(hr)) {
		delete pObject;
		return hr;
	}

	*ppObject = pObject;
	return S_OK;
}
