// artktest.cpp : R\[ AvP[ṼGg |Cg`܂B
//
#include "stdafx.h"

#include <ar/TestARPatternFactory.h>

#include <js/JSEngineFactory.h>
#include <common/SourceFactory.h>
#include <js/SourceJSCode.h>
#include <ar/ARWindow.h>
#include <video/CaptureDeviceFactory.h>
#include <window/WindowFactory.h>

#define RETURN(hr) \
	_CrtDumpMemoryLeaks(); \
	return hr;


// EBhEێ܂B
int main(int argc, char **argv)
{
	WindowFactory windowFactory(WINDOW_WIN32, &argc, argv);
	JSEngineFactory engineFactory;

	// 
	IJSEngine* pEngine = NULL;
	HRESULT hr = engineFactory.Create(& pEngine);
	if(FAILED(hr)) {
		RETURN(hr);
	}
 
	IJSContext* pContext = NULL;
	hr = pEngine->CreateContext(& pContext);
	if(FAILED(hr)) {
		pEngine->Destroy();
		pEngine = NULL;
		RETURN(hr);
	}

	ICaptureDevice* pCaptureDevice = NULL;
	CaptureDeviceFactory captureDeviceFactory;
	hr = captureDeviceFactory.Create(& pCaptureDevice);
	if(FAILED(hr)) {
		pEngine->Destroy();
		pEngine = NULL;
		RETURN(hr);
	}

	IWindow* pBaseWindow = NULL;
	hr = windowFactory.Create(& pBaseWindow);
	if(FAILED(hr)) {
		pEngine->Destroy();
		pEngine = NULL;
		pCaptureDevice->Destroy();
		pCaptureDevice = NULL;
		RETURN(hr);
	}

	ARWindow* pWindow = new ARWindow(pBaseWindow, pCaptureDevice, & hr);
	if(FAILED(hr)) {
		pEngine->Destroy();
		pEngine = NULL;
		pCaptureDevice->Destroy();
		pCaptureDevice = NULL;
		pBaseWindow->Destroy();
		pBaseWindow = NULL;
		RETURN(hr);
	}

	hr = pWindow->Attach(pContext);
	if(FAILED(hr)) {
		pContext->Destroy();
		pContext = NULL;
		pEngine->Destroy();
		pEngine = NULL;
		RETURN(hr);
	}

	SourceFactory sourceFactory("", & hr);
	if(FAILED(hr)) {
		pContext->Destroy();
		pContext = NULL;
		pEngine->Destroy();
		pEngine = NULL;
		RETURN(hr);
	}
	{
		ISource* pSource = NULL;
		hr = sourceFactory.Load("Data/test.js", & pSource);
		if(FAILED(hr)) {
			pContext->Destroy();
			pContext = NULL;
			pEngine->Destroy();
			pEngine = NULL;
			RETURN(hr);
		}

		SourceJSCode code(pSource, & hr);
		pSource = NULL;
		if(FAILED(hr)) {
			pContext->Destroy();
			pContext = NULL;
			pEngine->Destroy();
			pEngine = NULL;
			RETURN(hr);
		}

		hr = pContext->Evaluate(& code);
		if(FAILED(hr) || hr == S_FALSE) {
			pContext->Destroy();
			pContext = NULL;
			pEngine->Destroy();
			pEngine = NULL;
			delete pWindow;
			pWindow = NULL;
			RETURN(hr);
		}
	}

	// p^[̏
	/*TestARPatternFactory* pTestPatternFactory = new TestARPatternFactory();
	ARPattern* pPattern = NULL;
	ARGraphics* pGraphics = NULL;
	hr = pTestPatternFactory->CreatePattern(& pPattern, & pGraphics);
	delete pTestPatternFactory;
	pTestPatternFactory = NULL;

    if(FAILED(hr)) {
        printf("pattern load error !!\n");
        return (0);
    }

	hr = pWindow->AddPattern(pPattern);
	if(FAILED(hr)) {
		delete pWindow;
		pWindow = NULL;
		return (0);
	}

	int index = 0;
	hr = pWindow->AddGraphics(pGraphics, & index);
	if(FAILED(hr)) {
		delete pWindow;
		pWindow = NULL;
		return (0);
	}*/

	hr = pWindow->Start();
	if(FAILED(hr)) {
		delete pWindow;
		pWindow = NULL;
		return (0);
	}

	pWindow->RunMainLoop();
	return (0);
}
