#pragma once

#include "ARWindow.h"
#include "ARGraphics.h"
#include "../video/ICaptureDevice.h"
#include "../window/IWindow.h"
#include "../common/AbstractProperty.h"
#include "../plugin/IARGObjectFactory.h"
class ARPattern;
class ARJSWindow;
class IJSContext;
class IImageLoader;
class IPatternDetector;

/**
* AR̃EBhEłB
*/
class ARWindow : public IWindowNotify
{
public:

	/**
	* \z܂B
	*/
	DLL_DECL ARWindow(IWindow* pWindow, ICaptureDevice* pCaptureDevice, HRESULT* phr);

	/**
	* j܂B
	*/
	DLL_DECL ~ARWindow(void);

	/**
	* rfI܂B
	*/
	DLL_DECL HRESULT InitVideo(const char* config);

	/**
	* EBhE܂B
	*/
	DLL_DECL HRESULT InitWindow(const char* config);

	/**
	* EBhEJn܂B
	*/
	DLL_DECL HRESULT Start();

	/**
	* EBhE~܂B
	*/
	DLL_DECL HRESULT Stop();

	/**
	* p^[oǉ܂B
	*/
	DLL_DECL HRESULT AddPatternDetector(const char* name, IPatternDetector* pDetector);

	/**
	* OtBbNǉ܂B
	*/
	DLL_DECL HRESULT AddGraphics(ARGraphics* pGraphics, int* pIndex);

	/**
	* OtBbN폜܂B
	*/
	DLL_DECL HRESULT RemoveGraphics(int index);

	/**
	* C[vȂ܂B
	*/
	DLL_DECL HRESULT RunMainLoop();

	/**
	* JavaScriptReLXgݒ肵܂B
	*/
	DLL_DECL HRESULT Attach(IJSContext* pContext);

	/**
	* EBhE擾܂B
	*/
	HRESULT GetWindow(IWindow** ppWindow);

	/**
	* EBhEB
	*/
	virtual HRESULT OnCreate();

	/**
	* IdleB
	*/
	virtual HRESULT OnIdle();
	
	/**
	* `揈B
	*/
	virtual HRESULT OnDraw();

	/**
	* L[Cxg̏B
	*/
	virtual HRESULT OnKeyEvent(unsigned char key, int x, int y);
	
	/**
	* G[B
	*/
	virtual HRESULT OnError(HRESULT hr);

	/**
	* Xe[^X擾܂B
	*/
	virtual HRESULT GetStatus(string& status);

	/**
	* Xe[^Xݒ肵܂B
	*/
	virtual HRESULT SetStatus(const char* status);

	/**
	* ImageLoaderݒ肵܂B
	*/
	DLL_DECL HRESULT AddImageLoader(const char* name, IImageLoader* pImageLoader);

	/**
	* ImageLoader擾܂B
	*/
	HRESULT GetImageLoader(const char* name, IImageLoader** ppImageLoader);

	/**
	* OtBbÑXg擾܂B
	*/
	HRESULT GetWorlds(IArrayProperty** ppProp);

	/**
	* IuWFNgt@Ngǉ܂B
	*/
	HRESULT AddObjectFactory(const char* name, IARGObjectFactory* pFactory);

	/**
	* IuWFNgt@Ng̈ꗗ擾܂B
	*/
	HRESULT GetObjectFactories(vector<pair<string, IARGObjectFactory*>>& factories);

	/**
	* OpenGL̓ǂݍݗpReLXgJn܂B
	*/
	HRESULT BeginGLLoading();

	/**
	* OpenGL̓ǂݍݗpReLXgI܂B
	*/
	HRESULT EndGLLoading();

	/**
	* p^[o퐔擾܂B
	*/
	HRESULT GetPatternDetectorCount(int* pCount);

	/**
	* p^[o擾܂B
	*/
	HRESULT GetPatternDetector(int index, IPatternDetector** ppDetector);

private:

	/**
	* 摜[_܂B
	*/
	HRESULT InitImageLoaders();

	/**
	* Ԃ񍐂܂B
	*/
	HRESULT ReportStatus(ostream* pOut);

	/**
	* p^[ێ܂B
	*/
	vector<IPatternDetector*> m_detectors;

	/**
	* JavaScriptIuWFNgێ܂B
	*/
	ARJSWindow* m_pJSObject;

	/**
	* JE^ێ܂B
	*/
	int m_iCount;

	/**
	* EBhEێ܂B
	*/
	IWindow* m_pWindow;

	/**
	* Lv`foCXێ܂B
	*/
    ICaptureDevice* m_pCaptureDevice;

	/**
	* ReLXgێ܂B
	*/
	IJSContext* m_pContext;

	/**
	* Xe[^Xێ܂B
	*/
	string m_status;

	/**
	* 摜[_ێ܂B
	*/
	map<string, IImageLoader*> m_imageLoaders;

	/**
	* IuWFNgt@Ngێ܂B
	*/
	vector<pair<string, IARGObjectFactory*>> m_objectFactories;

	/**
	* vpeBNXB
	*/
	class PropertyImpl : public AbstractNamedProperty
	{
	public:

		/**
		* \z܂B
		*/
		PropertyImpl(ARWindow* pOwner, HRESULT* phr);

		/**
		* j܂B
		*/
		virtual ~PropertyImpl();

		/**
		* j܂B
		*/
		void Destroy();

	protected:
		
		/**
		* int̃vpeB擾܂B
		*/
		virtual HRESULT GetIntProperty(const char* name, int* pValue);
		
		/**
		* string̃vpeB擾܂B
		*/
		virtual HRESULT GetStringProperty(const char* name, string& value);
		
		/**
		* string̃vpeBݒ肵܂B
		*/
		virtual HRESULT SetStringProperty(const char* name, const char* value);

	private:

		/**
		* ܂B
		*/
		HRESULT Init();

		/**
		* I[i[ێ܂B
		*/
		ARWindow* m_pOwner;
	};

	/**
	* [gƂȂOtBbNNXB
	*/
	class ARRootGraphics : public ARGraphics
	{
	public:

		/**
		* \z܂B
		*/
		ARRootGraphics();

		/**
		* j܂B
		*/
		virtual ~ARRootGraphics();

		/**
		* j܂B
		*/
		void Destroy();

		/**
		* `悵܂B
		*/
		HRESULT Draw();

		/**
		* IJSObject擾܂B
		*/
		IJSObject* GetJSObject();
	};

	/**
	* [gƂȂOtBbNXێ܂B
	*/
	ARRootGraphics m_rootGraphics;

	/**
	* vpeBێ܂B
	*/
	PropertyImpl* m_pProp;
};
