#include "StdAfx.h"
#include "ARJSImage.h"
#include <GL/gl.h>
#include <GL/glut.h>
#include <AR/gsub.h>
#include <AR/video.h>
#include <AR/param.h>
#include <AR/ar.h>
#include "../../js/JSValueWrapper.h"
#include "../../common/ImageFactory.h"

#define CALLBACK_STATUS_SUCCESS 0
#define CALLBACK_STATUS_FAILED	1

GLfloat ARJSImage::m_matBase[4] = {1.0, 1.0, 1.0, 1.0};
GLfloat ARJSImage::m_matFlashShiny[1] = {50.0};

// \z܂B
ARJSImage::ARJSImage(ARWindow* pOwner, IImageLoader* pImageLoader)
: ARJSGraphicsModel(pOwner)
{
	m_pImageLoader = pImageLoader;
	m_visible = true;
	m_texture = 0;
	m_pImage = NULL;
	m_textureWidth = 0.0;
	m_textureHeight = 0.0;
	m_dScale = 1.0;

	pImageLoader->SetDefaultSize(512,512);
}

// j܂B
ARJSImage::~ARJSImage(void)
{
	if(m_texture != 0) {
		glDeleteTextures(1, & m_texture);
		m_texture = 0;
	}
	if(m_pImage != NULL) {
		m_pImage->Destroy();
		m_pImage = NULL;
	}

	// TODO: m_workerThread̏҂
}

// j܂B
void ARJSImage::Destroy()
{
	delete this;
}

// 摜[h܂B
HRESULT ARJSImage::Load(const char* filename, double scale, jsval callback)
{
	if(m_pJSObject == NULL) {
		return E_UNEXPECTED;
	}
	if(filename == NULL) {
		return E_POINTER;
	}
	return ((JSObjectImpl*)m_pJSObject)->Load(filename, scale, callback);
}

// `悵܂B
HRESULT ARJSImage::DrawModel()
{
	if(m_visible == false || m_pImage == NULL) {
		return S_FALSE;
	}

	glPushMatrix();
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	if(m_texture == 0) {
		glGenTextures(1, &m_texture);
	}

	glBindTexture(GL_TEXTURE_2D, m_texture);

	bool dirty = false;
	HRESULT hr = m_pImage->IsDirty(& dirty);
	if(dirty) {
		hr = RefreshTexture();
		if(hr == S_OK) {
			hr = m_pImage->SetDirty(false);
		}
	}

	if(hr == S_OK) {
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

		glMaterialfv(GL_FRONT, GL_SHININESS, m_matFlashShiny);	

		GLfloat matValues[4];
		CopyMemory(matValues, m_matBase, sizeof(m_matBase));
		matValues[3] *= m_pImage->GetMaterialAlpha();
		glMaterialfv(GL_FRONT, GL_SPECULAR, matValues);
		glMaterialfv(GL_FRONT, GL_AMBIENT, matValues);
		glMaterialfv(GL_FRONT, GL_DIFFUSE, matValues);
		glMaterialfv(GL_FRONT, GL_EMISSION, matValues);

		glBegin(GL_POLYGON);
			glColor4f(1.0, 1.0, 1.0, 1.0);

			double w = m_dScale;
			double h = (m_dScale * m_textureHeight / m_textureWidth) * -1.0;

			glTexCoord2f(0 , 0); 
			glVertex2f(-1.0 * w, -1.0 * h);
			glTexCoord2f(0, m_textureHeight); 
			glVertex2f(-1.0 * w, 1.0 * h);
			glTexCoord2f(m_textureWidth, m_textureHeight); 
			glVertex2f(1.0 * w, 1.0 * h);
			glTexCoord2f(m_textureWidth, 0); 
			glVertex2f(1.0 * w, -1.0 * h);
		glEnd();
		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_2D);

		hr = DrawChildren();
	}else{
		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_2D);
	}

	glPopMatrix();

	return hr;
}

// IJSObject𐶐܂B
AbstractARJSObject* ARJSImage::CreateJSObject()
{
	INamedProperty* pProp = NULL;
	HRESULT hr = GetProperty(& pProp);
	if(FAILED(hr)) {
		return NULL;
	}
	return new JSObjectImpl(this, pProp);
}

// fݒ肵܂B
HRESULT ARJSImage::SetImage(IImage* pImage)
{
	if(pImage == NULL) {
		return E_POINTER;
	}
	m_pImage = pImage;
	return S_OK;
}

// eNX`XV܂B
HRESULT ARJSImage::RefreshTexture()
{
	if(m_pImage == NULL) {
		return E_UNEXPECTED;
	}
	HRESULT hr = m_pImage->Refresh();
	if(FAILED(hr)) {
		return hr;
	}
	int size = 64;
	const int MAX_SIZE = 2048;
	while(size <= MAX_SIZE) {
		if(size >= m_pImage->GetWidth() && size >= m_pImage->GetHeight()) {
			break;
		}
		size = size * 2;
	}

	if(size > MAX_SIZE) {
		return E_FAIL;
	}

	//ImageFactory imageFactory;
	//imageFactory.DumpBitmap("loaderTest.bmp", pImage);

	const int SIZE_PIXEL = 4;
	unsigned char* pBuf = new unsigned char[size * size * SIZE_PIXEL];
	ZeroMemory(pBuf, sizeof(unsigned char) * size * size * SIZE_PIXEL);
	for(int y = 0; y < m_pImage->GetHeight(); y ++) {
		for(int x = 0; x < m_pImage->GetWidth(); x ++) {
			COMMON_ARGB argb;
			hr = m_pImage->GetPixel(x, y, & argb);
			if(FAILED(hr)) {
				delete pBuf;
				pBuf = NULL;
				return hr;
			}

			unsigned char* pOffset = &(pBuf[(x + y * size) * SIZE_PIXEL]);
			pOffset[0] = argb.r;
			pOffset[1] = argb.g;
			pOffset[2] = argb.b;
			pOffset[3] = argb.a;
		}
	}

	m_textureWidth = ((GLfloat)m_pImage->GetWidth()) / ((GLfloat) size);
	m_textureHeight = ((GLfloat)m_pImage->GetHeight()) / ((GLfloat) size);
		
	glTexImage2D(GL_TEXTURE_2D, 0, 4, size, size, 0, GL_RGBA, GL_UNSIGNED_BYTE, pBuf);

	delete pBuf;
	pBuf = NULL;

	return S_OK;
}

// \z܂B
ARJSImage::JSObjectImpl::JSObjectImpl(ARJSImage* pOwner, INamedProperty* pProp)
: AbstractARJSObject(pOwner, pProp)
{
	m_pOwner = pOwner;
	m_pProp = NULL;
}

// j܂B
ARJSImage::JSObjectImpl::~JSObjectImpl()
{
	m_pOwner = NULL;
	if(m_pProp != NULL) {
		delete m_pProp;
		m_pProp = NULL;
	}
}

// j܂B
void ARJSImage::JSObjectImpl::Destroy()
{
	delete this;
}

// [h܂B
HRESULT ARJSImage::JSObjectImpl::Load(const char* file, double scale, jsval callback)
{
	if(file == NULL) {
		return E_POINTER;
	}
	if(m_pContext == NULL) {
		return E_UNEXPECTED;
	}

	string url;
	HRESULT hr = m_pContext->GetAbsoluteURL(file, url);
	if(FAILED(hr)) {
		return hr;
	}

	if(callback != JSVAL_NULL) {
		HRESULT hr = m_callback.Set(callback);
		if(FAILED(hr)) {
			return hr;
		}
	}

	// m_dScale͒S痼[ւ̋Ȃ̂1/2Ă
	m_pOwner->m_dScale = scale / 2.0;

	ImageLoader* pNewLoader = new ImageLoader(this, m_pContext, scale);
	return m_pOwner->m_pImageLoader->Load(url.c_str(), pNewLoader);
}

// fݒ肵܂B
HRESULT ARJSImage::JSObjectImpl::SetImage(IImage* pImage)
{
	if(pImage != NULL) {
		HRESULT hr = m_pOwner->SetImage(pImage);
		if(FAILED(hr)) {
			return hr;
		}
	}

	// vpeB
	INamedProperty* pProp = NULL;
	HRESULT hr = pImage->GetProperty(& pProp);
	if(hr == S_OK) {
		m_pProp = new JSNamedProperty(pProp);

		hr = m_pProp->Attach(m_pContext, NULL);
		if(FAILED(hr)) {
			return hr;
		}
 	}

	jsval args[2];
	args[0] = JSVAL_NULL;
	args[1] = JSVAL_NULL;
	args[0] = OBJECT_TO_JSVAL(GetCore());
	if(pImage != NULL) {
		args[1] = INT_TO_JSVAL(CALLBACK_STATUS_SUCCESS);
	}else{
		args[1] = INT_TO_JSVAL(CALLBACK_STATUS_FAILED);
	}
	jsval rval = JSVAL_NULL;	

	return m_callback.InvokeSync(args, 2, & rval);
}

// ܂B
HRESULT ARJSImage::JSObjectImpl::InitMembers()
{
	HRESULT hr = m_callback.Init(m_pContext, this);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("props", true);
	if(FAILED(hr)) {
		return hr;
	}

	return AbstractARJSObject::InitMembers();
}

// vpeB擾܂B
HRESULT ARJSImage::JSObjectImpl::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	if(name == NULL) {
		return E_POINTER;
	}
	string id = name;
	if(id == "props") {
		if(vp == NULL) {
			return S_OK;
		}
		if(m_pProp == NULL) {
			*vp = JSVAL_NULL;
		}else{
			*vp = OBJECT_TO_JSVAL(m_pProp->GetCore());
		}
		return S_OK;
	}
	return AbstractARJSObject::GetProperty(cx, obj, name, vp);
}


// \z܂B
ARJSImage::ImageLoader::ImageLoader(ARJSImage::JSObjectImpl* pOwner, IJSContext* pContext, double scale)
{
	m_pOwner = pOwner;
	m_pContext = pContext;
	m_dScale = scale;
}

// j܂B
ARJSImage::ImageLoader::~ImageLoader()
{
	m_pOwner = NULL;
	m_pContext = NULL;
}

// j܂B
void ARJSImage::ImageLoader::Destroy()
{
	delete this;
}

// ̏B
HRESULT ARJSImage::ImageLoader::NotifySuccess(IImage* pImage)
{
	ImageNotifier* pNotifier = new ImageNotifier(m_pOwner, pImage);
	return m_pContext->InvokeAsync(pNotifier);
}

// s̏B
HRESULT ARJSImage::ImageLoader::NotifyFailed(const char* message)
{
	ImageNotifier* pNotifier = new ImageNotifier(m_pOwner, NULL);
	return m_pContext->InvokeAsync(pNotifier);
}

// \z܂B
ARJSImage::ImageNotifier::ImageNotifier(ARJSImage::JSObjectImpl* pOwner, IImage* pImage)
{
	m_pOwner = pOwner;
	m_pImage = pImage;
}

// j܂B
ARJSImage::ImageNotifier::~ImageNotifier()
{
	m_pOwner = NULL;
	m_pImage = NULL;
}

// j܂B
void ARJSImage::ImageNotifier::Destroy()
{
	delete this;
}

// s܂B
HRESULT ARJSImage::ImageNotifier::Invoke()
{
	return m_pOwner->SetImage(m_pImage);
}
