#include "StdAfx.h"
#include "ARJSPattern.h"
#include <GL/gl.h>
#include <GL/glut.h>
#include <AR/gsub.h>
#include <AR/video.h>
#include <AR/param.h>
#include <AR/ar.h>
#include "../../js/JSValueWrapper.h"

// }[J[𔭌Cxg
const int ARJSPattern::EVENT_FOUND = 0;

// }[J[Cxg
const int ARJSPattern::EVENT_LOST = 1;

// }[J[ύXꂽCxg
const int ARJSPattern::EVENT_CHANGED = 2;

// \z܂B
ARJSPattern::ARJSPattern(const char* url, const char* path, HRESULT* phr)
: ARPattern()
{
	m_pJSObject = NULL;
	m_pProp = NULL;
	m_patternWidth = 80.0;
	m_patternCenter[0] = 0.0;
	m_patternCenter[1] = 0.0;
	ZeroMemory(& m_trans, sizeof(m_trans));
	m_pOldMarkerInfo = NULL;
	m_url = url;
	*phr = S_OK;

	HRESULT hr = LoadPattern(path);
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
ARJSPattern::~ARJSPattern(void)
{
	if(m_pJSObject != NULL) {
		// ARJSPatternJSObjectImpl͓Ct^C
		m_pJSObject->Destroy();
		m_pJSObject = NULL;
	}
	if(m_pOldMarkerInfo != NULL) {
		delete m_pOldMarkerInfo;
		m_pOldMarkerInfo = NULL;
	}
	if(m_pProp != NULL) {
		delete m_pProp;
		m_pProp = NULL;
	}
}

// j܂B
void ARJSPattern::Destroy()
{
	delete this;
}

// JavaScriptGWݒ肵܂B
HRESULT ARJSPattern::Attach(JSEngineContext* pContext)
{
	if(pContext == NULL) {
		return E_POINTER;
	}

	HRESULT hr = S_OK;
	m_pProp = new PropertyImpl(this, & hr);
	if(m_pProp == NULL) {
		return E_OUTOFMEMORY;
	}
	if(FAILED(hr)) {
		return hr;
	}

	m_pJSObject = new JSObjectImpl(this, m_pProp);
	if(m_pJSObject == NULL) {
		return E_OUTOFMEMORY;
	}

	return m_pJSObject->Attach(pContext, NULL);
}

// IJSObject擾܂B
IJSObject* ARJSPattern::GetJSObject()
{
	return m_pJSObject;
}

// oԂʒm܂B
HRESULT ARJSPattern::Detect(ARMarkerInfoContainer* pDetectedC)
{
	if(m_pJSObject == NULL) {
		return S_OK;
	}
	ARMarkerInfo* pDetected = (ARMarkerInfo*) pDetectedC->pCore;
	int type = ARJSPattern::EVENT_FOUND;
	if(pDetected == NULL) {
		if(m_pOldMarkerInfo == NULL) {
			return S_OK;
		}
		delete m_pOldMarkerInfo;
		m_pOldMarkerInfo = NULL;
		type = ARJSPattern::EVENT_LOST;
	}else{
		if(m_pOldMarkerInfo == NULL) {
			arGetTransMat(pDetected, m_patternCenter, m_patternWidth, m_trans);

			type = ARJSPattern::EVENT_FOUND;
			m_pOldMarkerInfo = new ARMarkerInfo;
			if(m_pOldMarkerInfo == NULL) {
				return E_OUTOFMEMORY;
			}
			CopyMemory(m_pOldMarkerInfo, pDetected, sizeof(ARMarkerInfo));
			type = ARJSPattern::EVENT_FOUND;
		}else{
			double oldTrans[3][4];
			CopyMemory(oldTrans, m_trans, sizeof(oldTrans));
			arGetTransMatCont(pDetected, oldTrans, m_patternCenter, m_patternWidth,
							  m_trans);
			CopyMemory(m_pOldMarkerInfo, pDetected, sizeof(ARMarkerInfo));
			type = ARJSPattern::EVENT_CHANGED;
		}
	}

	return m_pJSObject->CallHandler(type, pDetected);
}

// \z܂B
ARJSPattern::PropertyImpl::PropertyImpl(ARJSPattern* pOwner, HRESULT* phr)
{
	m_pOwner = pOwner;
	*phr = S_OK;

	HRESULT hr = Init();
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
ARJSPattern::PropertyImpl::~PropertyImpl()
{
	m_pOwner = NULL;
}

// j܂B
void ARJSPattern::PropertyImpl::Destroy()
{
	delete this;
}

// ̃vpeB擾܂B
HRESULT ARJSPattern::PropertyImpl::GetDoubleProperty(const char* name, double* pValue)
{
	if(name == NULL || pValue == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == "width") {
		*pValue = m_pOwner->m_patternWidth;
		return S_OK;
	}else if(sname == "center.x") {
		*pValue = (m_pOwner->m_patternCenter[0]);
		return S_OK;
	}else if(sname == "center.y") {
		*pValue = (m_pOwner->m_patternCenter[1]);
		return S_OK;
	}else{
		return AbstractNamedProperty::GetDoubleProperty(name, pValue);
	}
}

// ̃vpeBݒ肵܂B
HRESULT ARJSPattern::PropertyImpl::SetDoubleProperty(const char* name, double value)
{
	if(name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == "width") {
		m_pOwner->m_patternWidth = value;
		return S_OK;
	}else if(sname == "center.x") {
		m_pOwner->m_patternCenter[0] = value;
		return S_OK;
	}else if(sname == "center.y") {
		m_pOwner->m_patternCenter[1] = value;
		return S_OK;
	}else{
		return AbstractNamedProperty::SetDoubleProperty(name, value);
	}
}

// int̃vpeB擾܂B
HRESULT ARJSPattern::PropertyImpl::GetIntProperty(const char* name, int* pValue)
{
	if(name == NULL || pValue == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == "id") {
		*pValue = m_pOwner->m_patternId;
		return S_OK;
	}else{
		return AbstractNamedProperty::GetIntProperty(name, pValue);
	}
}

// string̃vpeB擾܂B
HRESULT ARJSPattern::PropertyImpl::GetStringProperty(const char* name, string& value)
{
	if(name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == "href") {
		value = m_pOwner->m_url;
		return S_OK;
	}else{
		return AbstractNamedProperty::GetStringProperty(name, value);
	}
}

// ܂B
HRESULT ARJSPattern::PropertyImpl::Init()
{
	HRESULT hr = S_OK;

	hr = DefineProperty("id", PROP_INT);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("href", PROP_STRING);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("width", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("center.x", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("center.y", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	return S_OK;
}

// \z܂B
ARJSPattern::JSObjectImpl::JSObjectImpl(ARJSPattern* pOwner, INamedProperty* pProp)
: JSNamedProperty(pProp)
{
	m_pOwner = pOwner;
	m_pJSEvent = NULL;
}

// j܂B
ARJSPattern::JSObjectImpl::~JSObjectImpl()
{
	m_pOwner = NULL;
	if(m_pJSEvent != NULL){
		m_pJSEvent->Destroy();
		m_pJSEvent = NULL;
	}
}

// j܂B
void ARJSPattern::JSObjectImpl::Destroy()
{
	delete this;
}

// nhĂяo܂B
HRESULT ARJSPattern::JSObjectImpl::CallHandler(int type, ARMarkerInfo* pMarkerInfo)
{
	HRESULT hr = S_OK;
	JSObject* pEvent = NULL;
	hr = CreatePatternEvent(pMarkerInfo, & pEvent);
	if(FAILED(hr)) {
		return hr;
	}

	FunctionRef* pFunc = NULL;
	if(type == ARJSPattern::EVENT_FOUND) {
		pFunc = & m_onFound;
	}else if(type == ARJSPattern::EVENT_LOST) {
		pFunc = & m_onLost;
	}else if(type == ARJSPattern::EVENT_CHANGED) {
		pFunc = & m_onChanged;
	}
	if(pFunc == NULL) {
		return S_OK;
	}

	jsval eventVal = OBJECT_TO_JSVAL(pEvent);
	jsval rval = JSVAL_FALSE;
	hr = pFunc->InvokeSync(& eventVal, 1, & rval);
	if(FAILED(hr)) {
		return hr;
	}
	return S_OK;
}

// ܂B
HRESULT ARJSPattern::JSObjectImpl::InitMembers()
{
	HRESULT hr = S_OK;
	hr = DefineProperty("onfound", false);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("onlost", false);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("onchanged", false);
	if(FAILED(hr)) {
		return hr;
	}

	hr = m_onFound.Init(m_pContext, this);
	if(FAILED(hr)) {
		return hr;
	}
	hr = m_onLost.Init(m_pContext, this);
	if(FAILED(hr)) {
		return hr;
	}
	hr = m_onChanged.Init(m_pContext, this);
	if(FAILED(hr)) {
		return hr;
	}

	return JSNamedProperty::InitMembers();
}

// vpeBݒ肵܂B
HRESULT ARJSPattern::JSObjectImpl::SetProperty(JSContext *cx, JSObject *obj, const char* name, jsval vp)
{
	string sname = name;

	if(sname == "onfound") {
		return m_onFound.Set(vp);
	}else if(sname == "onlost") {
		return m_onLost.Set(vp);
	}else if(sname == "onchanged") {
		return m_onChanged.Set(vp);
	}else{
		return JSNamedProperty::SetProperty(cx, obj, name, vp);
	}
}

// vpeB擾܂B
HRESULT ARJSPattern::JSObjectImpl::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	string sname = name;

	if(sname == "onfound") {
		return m_onFound.Get(vp);
	}else if(sname == "onlost") {
		return m_onLost.Get(vp);
	}else if(sname == "onchanged") {
		return m_onChanged.Get(vp);
	}else{
		return JSNamedProperty::GetProperty(cx, obj, name, vp);
	}
}

// p^[Cxg𐶐܂B
HRESULT ARJSPattern::JSObjectImpl::CreatePatternEvent(const ARMarkerInfo* pMarkerInfo, JSObject** ppEvent)
{
	if(ppEvent == NULL) {
		return E_POINTER;
	}

	if(m_pJSEvent == NULL) {
		HRESULT hr = S_OK;
		m_pJSEvent = new JSEventImpl(m_pOwner, pMarkerInfo, & hr);
		if(m_pJSEvent == NULL) {
			return E_OUTOFMEMORY;
		}
		if(FAILED(hr)) {
			return hr;
		}
		hr = m_pJSEvent->Attach(m_pContext, NULL);
		if(FAILED(hr)) {
			return hr;
		}
	}else{
		HRESULT hr = m_pJSEvent->Set(pMarkerInfo);
		if(FAILED(hr)) {
			return hr;
		}
	}

	*ppEvent = m_pJSEvent->GetCore();
	return S_OK;
}

// \z܂B
ARJSPattern::JSEventImpl::JSEventImpl(ARJSPattern* pOwner, const ARMarkerInfo* pMarkerInfo, HRESULT* phr)
: JSMemberCollection("ARJSPattern::JSEventImpl")
{
	m_pOwner = pOwner;
	m_markerDefined = false;
	ZeroMemory(& m_markerInfo, sizeof(ARMarkerInfo));
	*phr = S_OK;

	HRESULT hr = Set(pMarkerInfo);
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
ARJSPattern::JSEventImpl::~JSEventImpl()
{
	m_pOwner = NULL;
}

// j܂B
void ARJSPattern::JSEventImpl::Destroy()
{
	delete this;
}

// ARMarkerInfoݒ肵܂B
HRESULT ARJSPattern::JSEventImpl::Set(const ARMarkerInfo* pMarkerInfo)
{
	if(pMarkerInfo == NULL) {
		m_markerDefined = false;
		ZeroMemory(& m_markerInfo, sizeof(ARMarkerInfo));
	}else{
		m_markerDefined = true;
		CopyMemory(& m_markerInfo, pMarkerInfo, sizeof(ARMarkerInfo));
	}
	return S_OK;
}

// ܂B
HRESULT ARJSPattern::JSEventImpl::InitMembers()
{
	HRESULT hr = S_OK;

	hr = DefineProperty("target", true);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("found", true);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("vertex", true);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("translate", true);
	if(FAILED(hr)) {
		return hr;
	}
	return S_OK;
}

// vpeB擾܂B
HRESULT ARJSPattern::JSEventImpl::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	if(vp == NULL) {
		return S_OK;
	}
	string sname = name;
	if(sname == "target") {
		jsval targetValue = OBJECT_TO_JSVAL(m_pOwner->m_pJSObject->GetCore());
		*vp = targetValue;
		return S_OK;
	}else if(sname == "found") {
		JSValueWrapper valuew(cx, JSVAL_NULL);

		HRESULT hr = valuew.SetBool(m_markerDefined);
		if(FAILED(hr)) {
			return hr;
		}
		return valuew.Get(vp);
	}else if(sname == "vertex") {
		if(m_markerDefined == false) {
			*vp = JSVAL_NULL;
			return S_OK;
		}
		jsval vertexValues[4];
		for(int i = 0; i < 4; i ++) {
			JSObject* pArray = NULL;
			HRESULT hr = CreateDoubleArray(m_markerInfo.vertex[i], 2, & pArray);
			if(FAILED(hr)) {
				return hr;
			}
			vertexValues[i] = OBJECT_TO_JSVAL(pArray);
		}
		JSObject* pVertexObj = JS_NewArrayObject(cx, 4, vertexValues);
		if(pVertexObj == NULL) {
			return E_FAIL;
		}
		*vp = OBJECT_TO_JSVAL(pVertexObj);
		return S_OK;
	}else if(sname == "translate") {
		if(m_markerDefined == false) {
			*vp = JSVAL_NULL;
			return S_OK;
		}
		jsval transValues[3];
		for(int i = 0; i < 3; i ++) {
			JSObject* pArray = NULL;
			HRESULT hr = CreateDoubleArray(m_pOwner->m_trans[i], 4, & pArray);
			if(FAILED(hr)) {
				return hr;
			}
			transValues[i] = OBJECT_TO_JSVAL(pArray);
		}
		JSObject* pTransObj = JS_NewArrayObject(cx, 3, transValues);
		if(pTransObj == NULL) {
			return E_FAIL;
		}
		*vp = OBJECT_TO_JSVAL(pTransObj);
		return S_OK;
	}else{
		return E_NOTIMPL;
	}
}
