#include "StdAfx.h"
#include "ARJSPatternFactory.h"
#include "../ARPattern.h"
#include <AR/ar.h>
#include "ARJSPattern.h"
#include "../../common/URLPath.h"
#include "../../js/JSValueWrapper.h"

// \z܂B
ARJSPatternFactory::ARJSPatternFactory(JSEngineContext* pContext, HRESULT* phr)
{
	m_pContext = pContext;
	if(pContext == NULL) {
		*phr = E_POINTER;
		return;
	}
}

// j܂B
ARJSPatternFactory::~ARJSPatternFactory()
{
	DestroyPatterns();
}

// j܂B
void ARJSPatternFactory::Destroy()
{
	delete this;
}

// ̃p[^擾܂B
HRESULT ARJSPatternFactory::GetCreateParamCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = 1;
	return S_OK;
}

// p^[擾܂B
HRESULT ARJSPatternFactory::GetCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = (int) m_patterns.size();
	return S_OK;
}

// p^[擾܂B
HRESULT ARJSPatternFactory::Get(int index, IJSObject** ppObject)
{
	if(ppObject == NULL) {
		return E_POINTER;
	}
	*ppObject = m_patterns[index]->GetJSObject();
	return S_OK;
}

// p^[ǉ܂B
HRESULT ARJSPatternFactory::Create(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if(argc < 1) {
		return E_INVALIDARG;
	}

	string path;
	JSValueWrapper jspath(cx, argv[0]);
	HRESULT hr = jspath.GetString(path);
	if(FAILED(hr)) {
		return hr;
	}
	string url;
	hr = m_pContext->GetAbsoluteURL(path.c_str(), url);
	if(FAILED(hr)) {
		return hr;
	}

	URLPath urlPath(url.c_str(), & hr);
	if(FAILED(hr)) {
		return hr;
	}
	string realPath;
	hr = urlPath.GetRealPath(realPath);
	if(FAILED(hr)) {
		return hr;
	}

	ARJSPattern* pNewPattern = new ARJSPattern(url.c_str(), realPath.c_str(), & hr);
	if(pNewPattern == NULL) {
		return E_OUTOFMEMORY;
	}
	if(FAILED(hr)) {
		return hr;
	}

	m_patterns.push_back(pNewPattern);

	hr = pNewPattern->Attach(m_pContext);
	if(FAILED(hr)) {
		return hr;
	}

	if(rval == NULL) {
		return S_OK;
	}
	*rval = OBJECT_TO_JSVAL(pNewPattern->GetJSObject()->GetCore());
	return S_OK;
}

// p^[̃Xg擾܂B
HRESULT ARJSPatternFactory::GetPatterns(vector<ARPattern*>& pattern)
{
	for(vector<ARPattern*>::iterator it = m_patterns.begin();
		it != m_patterns.end();
		it ++) {
			pattern.push_back(*it);
	}
	return S_OK;
}

// osȂ܂B
HRESULT ARJSPatternFactory::Process(const CAPTURE_IMAGE* pImage)
{
	if(pImage == NULL) {
		return E_POINTER;
	}
	ARUint8* pBuf = pImage->pBuffer;

    /* detect the markers in the video frame */
    ARMarkerInfo    *marker_info = NULL;
    int             marker_num = 0;
	int             thresh = 100;

    if( arDetectMarker(pBuf, thresh, &marker_info, &marker_num) < 0 ) {
		return E_FAIL;
    }

	ARMarkerInfoContainer markerInfoC;
	markerInfoC.pCore = marker_info;
	for(vector<ARPattern*>::iterator it = m_patterns.begin(); it != m_patterns.end(); it ++) {
		ARPattern* pPattern = *it;
		// p^[̏
		HRESULT hr = pPattern->Process(& markerInfoC, marker_num);

		if(FAILED(hr)) {
			return hr;
		}
	}
	return S_OK;
}

// p^[ׂč폜܂B
void ARJSPatternFactory::DestroyPatterns()
{
	for(vector<ARPattern*>::iterator it = m_patterns.begin(); it != m_patterns.end(); it ++) {
		ARPattern* pPattern = *it;
		delete pPattern;
	}
	m_patterns.clear();
}

