#include "StdAfx.h"
#include "ARJSPluginObject.h"
#include <GL/gl.h>
#include <GL/glut.h>
#include <AR/gsub.h>
#include <AR/video.h>
#include <AR/param.h>
#include <AR/ar.h>
#include "../ARWindow.h"
#include "../../js/JSValueWrapper.h"
#include "../../js/JSDispatch.h"

#define CALLBACK_STATUS_SUCCESS 0
#define CALLBACK_STATUS_FAILED	1

// \z܂B
ARJSPluginObject::ARJSPluginObject(ARWindow* pOwner, IARGObjectFactory* pFactory, HRESULT* phr)
: ARJSGraphicsModel(pOwner)
{
	m_visible = true;
	m_pObject = NULL;
	m_pFactory = NULL;
	if(pFactory == NULL) {
		*phr = E_POINTER;
		return;
	}
	
	m_pFactory = pFactory;
	pFactory->AddRef();
}

// j܂B
ARJSPluginObject::~ARJSPluginObject(void)
{
	if(m_pObject != NULL) {
		m_pObject->Release();
		m_pObject = NULL;
	}
	if(m_pFactory != NULL) {
		m_pFactory->Release();
		m_pFactory = NULL;
	}

	// TODO: m_workerThread̏҂
}

// j܂B
void ARJSPluginObject::Destroy()
{
	delete this;
}

// `悵܂B
HRESULT ARJSPluginObject::DrawModel()
{
	if(m_visible == false || m_pObject == NULL) {
		return S_FALSE;
	}

	glPushMatrix();

	HRESULT hr = m_pObject->Draw();

	if(hr == S_OK) {
		hr = DrawChildren();
	}

	glPopMatrix();

	return hr;
}

// ACh̏sȂ܂B
HRESULT ARJSPluginObject::OnIdle()
{
	if(m_pObject == NULL) {
		return S_FALSE;
	}

	HRESULT hr = m_pObject->OnIdle();
	if(FAILED(hr)) {
		return hr;
	}

	return ARGraphics::OnIdle();
}

// 3Df[h܂B
HRESULT ARJSPluginObject::Load(VARIANT* pVars, int argc, jsval callback)
{
	if(m_pJSObject == NULL) {
		return E_UNEXPECTED;
	}
	if(pVars == NULL) {
		return E_POINTER;
	}
	return ((JSObjectImpl*)m_pJSObject)->Load(pVars, argc, callback);
}

// IJSObject𐶐܂B
AbstractARJSObject* ARJSPluginObject::CreateJSObject()
{
	INamedProperty* pProp = NULL;
	HRESULT hr = GetProperty(& pProp);
	if(FAILED(hr)) {
		return NULL;
	}
	return new JSObjectImpl(this, pProp);
}

// fݒ肵܂B
HRESULT ARJSPluginObject::SetObject(IARGObject* pObject)
{
	m_pObject = pObject;
	if(pObject != NULL) {
		pObject->AddRef();
	}
	return S_OK;
}

// \z܂B
ARJSPluginObject::JSObjectImpl::JSObjectImpl(ARJSPluginObject* pOwner, INamedProperty* pProp)
: AbstractARJSObject(pOwner, pProp)
{
	m_pOwner = pOwner;
	m_pParam = NULL;
}

// j܂B
ARJSPluginObject::JSObjectImpl::~JSObjectImpl()
{
	m_pOwner = NULL;
	if(m_pParam != NULL) {
		m_pParam->Destroy();
		m_pParam = NULL;
	}
}

// j܂B
void ARJSPluginObject::JSObjectImpl::Destroy()
{
	delete this;
}

// [h܂B
HRESULT ARJSPluginObject::JSObjectImpl::Load(VARIANT* pVars, int argc, jsval callback)
{
	if(pVars == NULL) {
		return E_POINTER;
	}
	if(m_pContext == NULL) {
		return E_UNEXPECTED;
	}

	HRESULT hr = S_OK;
	if(callback != JSVAL_NULL) {
		hr = m_callback.Set(callback);
		if(FAILED(hr)) {
			return hr;
		}
	}

	ARWindow* pWindow = NULL;
	hr = m_pOwner->GetOwner(& pWindow);
	if(FAILED(hr)) {
		return hr;
	}

	ObjectLoader* pNewLoader = new ObjectLoader(this, pWindow, m_pContext, m_pOwner->m_pFactory, pVars, argc);
	return m_pOwner->m_workerThread.Start(pNewLoader);
}

// fݒ肵܂B
HRESULT ARJSPluginObject::JSObjectImpl::SetObject(HRESULT lhr, IARGObject* pObject)
{
	HRESULT hr = m_pOwner->SetObject(pObject);
	if(FAILED(hr)) {
		return hr;
	}

	if(pObject != NULL) {
		// Cxg
		int evCount = 0;
		hr = pObject->GetEventCount(& evCount);
		if(FAILED(hr)) {
			return hr;
		}

		for(int i = 0; i < evCount; i ++) {
			int maxLen = 0;
			hr = pObject->GetEvent(i, NULL, 0, &maxLen);
			if(FAILED(hr)) {
				return hr;
			}

			char* pEventName = new char[maxLen + 1];
			ZeroMemory(pEventName, sizeof(char) * (maxLen + 1));

			hr = pObject->GetEvent(i, pEventName, maxLen + 1, & maxLen);
			if(FAILED(hr)) {
				delete[] pEventName;
				pEventName = NULL;
				return hr;
			}
			pEventName[maxLen] = '\0';
			string evName = pEventName;
			delete[] pEventName;
			pEventName = NULL;

			m_eventNames.insert(evName);

			string handlerName = "on";
			handlerName += evName;
			hr = DefineProperty(handlerName.c_str(), false);
			if(FAILED(hr)) {
				return hr;
			}
		}
	}

	jsval args[2];
	args[0] = JSVAL_NULL;
	args[1] = JSVAL_NULL;
	args[0] = OBJECT_TO_JSVAL(GetCore());
	if(pObject != NULL) {
		args[1] = INT_TO_JSVAL(CALLBACK_STATUS_SUCCESS);
	}else{
		args[1] = INT_TO_JSVAL(lhr);
	}
	jsval rval = JSVAL_NULL;	

	return m_callback.InvokeSync(args, 2, & rval);
}

// ܂B
HRESULT ARJSPluginObject::JSObjectImpl::InitMembers()
{
	HRESULT hr = m_callback.Init(m_pContext, this);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("params", true);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("clock", false);
	if(FAILED(hr)) {
		return hr;
	}

	return AbstractARJSObject::InitMembers();
}

// vpeB擾܂B
HRESULT ARJSPluginObject::JSObjectImpl::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	if(cx == NULL || obj == NULL || name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == "params") {
		if(m_pParam == NULL) {
			IDispatch* pDisp = NULL;
			if(m_pOwner->m_pObject == NULL) {
				*vp = JSVAL_NULL;
				return S_OK;
			}
			HRESULT hr = m_pOwner->m_pObject->QueryInterface(& pDisp);
			if(hr == E_NOINTERFACE) {
				*vp = JSVAL_NULL;
				return S_OK;
			}
			if(FAILED(hr)) {
				return hr;
			}

			hr = S_OK;
			m_pParam = new JSDispatch(pDisp, & hr);
			if(FAILED(hr)) {
				delete m_pParam;
				m_pParam = NULL;
				return hr;
			}

			hr = m_pParam->Attach(m_pContext, NULL);
			if(FAILED(hr)) {
				delete m_pParam;
				m_pParam = NULL;
				return hr;
			}
			if(pDisp != NULL) {
				pDisp->Release();
				pDisp = NULL;
			}
		}

		*vp = OBJECT_TO_JSVAL(m_pParam->GetCore());
		return S_OK;
	}else if(sname == "clock") {
		IARGSyncClock* pClock = NULL;
		if(m_pOwner->m_pObject == NULL) {
			*vp = JSVAL_NULL;
			return S_OK;
		}
		HRESULT hr = m_pOwner->m_pObject->GetSyncClock(& pClock);
		if(hr == S_FALSE) {
			*vp = JSVAL_NULL;
			return S_OK;
		}
		IDispatch* pDisp = NULL;
		hr = pClock->QueryInterface(& pDisp);
		if(hr == E_NOINTERFACE) {
			*vp = JSVAL_NULL;
			return S_OK;
		}
		if(FAILED(hr)) {
			return hr;
		}
		pClock->Release();
		pClock = NULL;

		hr = S_OK;
		JSDispatch* pJSClock = new JSDispatch(pDisp, & hr);
		if(FAILED(hr)) {
			delete pJSClock;
			pJSClock = NULL;
			return hr;
		}

		hr = pJSClock->SetAutoDelete(true);
		if(FAILED(hr)) {
			delete pJSClock;
			pJSClock = NULL;
			return hr;
		}

		hr = pJSClock->Attach(m_pContext, NULL);
		if(FAILED(hr)) {
			delete pJSClock;
			pJSClock = NULL;
			return hr;
		}
		if(pDisp != NULL) {
			pDisp->Release();
			pDisp = NULL;
		}

		*vp = OBJECT_TO_JSVAL(pJSClock->GetCore());
		return S_OK;
	}
	
	return AbstractARJSObject::GetProperty(cx, obj, name, vp);
}

// vpeBݒ肵܂B
HRESULT ARJSPluginObject::JSObjectImpl::SetProperty(JSContext *cx, JSObject *obj, const char* name, jsval vp)
{
	if(cx == NULL || obj == NULL || name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == "clock") {
		if(m_pOwner->m_pObject == NULL) {
			return E_UNEXPECTED;
		}
		if(JSVAL_IS_NULL(vp) || JSVAL_IS_VOID(vp)) {
			return m_pOwner->m_pObject->SetSyncClock(NULL);
		}
		if(!(JSVAL_IS_OBJECT(vp))) {
			return E_INVALIDARG;
		}
		JSObject* pObj = JSVAL_TO_OBJECT(vp);
		JSObjectWrapper* pObjWrapper = (JSMemberCollection*) JS_GetPrivate(cx, pObj);
		if(pObjWrapper == NULL) {
			return E_UNEXPECTED;
		}

		IDispatch* pDisp = NULL;
		HRESULT hr = pObjWrapper->GetData(JSDISPATCH_IDISPATCH, (void**)& pDisp);
		if(FAILED(hr)) {
			return hr;
		}
		if(pDisp == NULL) {
			return E_UNEXPECTED;
		}
		IARGSyncClock* pSyncClock = NULL;
		hr = pDisp->QueryInterface(IID_IARGSyncClock, (void**)& pSyncClock);
		pDisp->Release();
		pDisp = NULL;
		if(FAILED(hr)) {
			return hr;
		}

		hr = m_pOwner->m_pObject->SetSyncClock(pSyncClock);
		pSyncClock->Release();
		pSyncClock = NULL;

		return hr;
	}else if(sname.find("on") == 0) {
		sname = sname.substr(2);
		if(m_eventNames.find(sname) != m_eventNames.end()) {
			// Cxg
			HRESULT hr = m_pOwner->m_pObject->RemoveEventDispatcher(sname.c_str(), NULL);
			if(FAILED(hr)) {
				return hr;
			}
			if(JSVAL_IS_NULL(vp) || JSVAL_IS_VOID(vp)) {
				return S_OK;
			}
			hr = S_OK;
			EventDispatcherImpl* pEventDisp = new EventDispatcherImpl(m_pContext, this, vp, & hr);
			if(FAILED(hr)) {
				delete pEventDisp;
				pEventDisp = NULL;
				return hr;
			}

			hr = m_pOwner->m_pObject->AddEventDispatcher(sname.c_str(), pEventDisp);
			if(FAILED(hr)) {
				delete pEventDisp;
				pEventDisp = NULL;
				return hr;
			}
			pEventDisp->Release();
			pEventDisp = NULL;

			return S_OK;
		}
	}

	return AbstractARJSObject::SetProperty(cx, obj, name, vp);
}

// \z܂B
ARJSPluginObject::JSObjectImpl::EventDispatcherImpl::EventDispatcherImpl(JSEngineContext* pContext, IJSObject* pOwner, jsval jsFunc, HRESULT* phr)
{
	m_pContext = pContext;
	HRESULT hr = m_jsFunc.Init(pContext, pOwner);
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}

	hr = m_jsFunc.Set(jsFunc);
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
ARJSPluginObject::JSObjectImpl::EventDispatcherImpl::~EventDispatcherImpl()
{
	m_pContext = NULL;
}

// C^tF[X擾܂B
STDMETHODIMP ARJSPluginObject::JSObjectImpl::EventDispatcherImpl::QueryInterface(const IID& riid, LPVOID* ppv)
{
	if(ppv == NULL) {
		return E_POINTER;
	}

	if(riid == IID_IARGEventDispatcher) {
		*ppv = static_cast<IARGEventDispatcher*>(this);
		AddRef();
		return S_OK;
	}

	return ARGBaseUnknown::QueryInterface(riid, ppv);
}

// Cxg̏sȂ܂B
STDMETHODIMP ARJSPluginObject::JSObjectImpl::EventDispatcherImpl::DispatchEvent(const VARIANT* pVars, int nVars)
{
	HRESULT hr = S_OK;
	Dispatcher* pDisp = new Dispatcher(this, pVars, nVars, & hr);
	if(FAILED(hr)) {
		delete pDisp;
		pDisp = NULL;
		return hr;
	}

	hr = m_pContext->InvokeAsync(pDisp);
	if(FAILED(hr)) {
		delete pDisp;
		pDisp = NULL;
		return hr;
	}
	return S_OK;
}

// \z܂B
ARJSPluginObject::JSObjectImpl::EventDispatcherImpl::Dispatcher::Dispatcher(EventDispatcherImpl* pOwner, const VARIANT* pVars, int nVars, HRESULT* phr)
{
	m_pVars = NULL;
	m_nVars = 0;
	m_pOwner = pOwner;
	pOwner->AddRef();

	HRESULT hr = Init(pVars, nVars);
	if(FAILED(hr)) {
		*phr = E_INVALIDARG;
		return;
	}
}

// j܂B
ARJSPluginObject::JSObjectImpl::EventDispatcherImpl::Dispatcher::~Dispatcher()
{
	if(m_pOwner != NULL) {
		m_pOwner->Release();
		m_pOwner = NULL;
	}
	if(m_pVars != NULL) {
		for(int i = 0; i < m_nVars; i ++) {
			VariantClear(m_pVars + i);
		}
		delete[] m_pVars;
		m_pVars = NULL;
		m_nVars = 0;
	}
}

// j܂B
void ARJSPluginObject::JSObjectImpl::EventDispatcherImpl::Dispatcher::Destroy()
{
	delete this;
}

// s܂B
HRESULT ARJSPluginObject::JSObjectImpl::EventDispatcherImpl::Dispatcher::Invoke()
{
	jsval* pArg = NULL;
	if(m_pVars == NULL) {
		pArg = new jsval[1];
		pArg[0] = JSVAL_NULL;
	}else{
		pArg = new jsval[m_nVars];
		for(int i = 0; i < m_nVars; i ++) {
			pArg[i] = JSVAL_NULL;
			HRESULT hr = JSDispatch::ConvertVARIANTtoJSVAL(m_pOwner->m_pContext, m_pVars + i, pArg + i);
			if(FAILED(hr)) {
				delete[] pArg;
				pArg = NULL;
				return hr;
			}
		}
	}

	jsval rval = JSVAL_NULL;
	HRESULT hr = m_pOwner->m_jsFunc.InvokeSync(pArg, m_nVars, & rval);

	delete[] pArg;
	pArg = NULL;
	return hr;
}

// ܂B
HRESULT ARJSPluginObject::JSObjectImpl::EventDispatcherImpl::Dispatcher::Init(const VARIANT* pVars, int nVars)
{
	if(pVars == NULL || nVars < 0) {
		return E_POINTER;
	}
	if(nVars == 0) {
		m_nVars = 0;
		m_pVars = NULL;
		return S_OK;
	}

	m_pVars = new VARIANT[nVars];
	m_nVars = nVars;
	for(int i = 0; i < nVars; i ++) {
		VARIANT* pVar = m_pVars + i;
		VariantInit(pVar);
		VariantCopy(pVar, pVars + i);
	}
	return S_OK;
}

// \z܂B
ARJSPluginObject::ObjectLoader::ObjectLoader(ARJSPluginObject::JSObjectImpl* pOwner, ARWindow* pWindow, IJSContext* pContext, IARGObjectFactory* pFactory, VARIANT* pVars, int argc)
{
	m_pOwner = pOwner;
	m_pWindow = pWindow;
	m_pContext = pContext;
	m_pVars = pVars;
	m_iArgc = argc;
	m_pFactory = pFactory;
	pFactory->AddRef();
}

// j܂B
ARJSPluginObject::ObjectLoader::~ObjectLoader()
{
	m_pOwner = NULL;
	m_pContext = NULL;
	if(m_pVars != NULL) {
		for(int j = 0; j < m_iArgc; j ++) {
			VariantClear(m_pVars + j);
		}
		delete[] m_pVars;
		m_pVars = NULL;
	}
	m_iArgc = 0;
	if(m_pFactory != NULL) {
		m_pFactory->Release();
		m_pFactory = NULL;
	}
	m_pWindow = NULL;
}

// j܂B
void ARJSPluginObject::ObjectLoader::Destroy()
{
	delete this;
}

// s܂B
HRESULT ARJSPluginObject::ObjectLoader::Invoke()
{
	if(m_pFactory == NULL || m_pWindow == NULL) {
		return E_UNEXPECTED;
	}

	bool useGL = false;
	IARGObject* pObject = NULL;
	HRESULT hr = m_pFactory->UseGL(& useGL);

	if(hr == S_OK && useGL) {
		hr = m_pWindow->BeginGLLoading();
	}

	if(hr == S_OK) {
		hr = m_pFactory->Create(m_pVars, m_iArgc, & pObject);
	}

	if(useGL) {
		HRESULT glhr = m_pWindow->EndGLLoading();
		if(!(FAILED(hr))) {
			hr = glhr;
		}
	}

	// ʒm
	ObjectNotifier* pNotifier = new ObjectNotifier(m_pOwner, hr, pObject);
	if(pObject != NULL) {
		pObject->Release();
		pObject = NULL;
	}
	return m_pContext->InvokeAsync(pNotifier);
}

// \z܂B
ARJSPluginObject::ObjectNotifier::ObjectNotifier(ARJSPluginObject::JSObjectImpl* pOwner, HRESULT hr, IARGObject* pObject)
{
	m_pOwner = pOwner;
	m_hResult = hr;
	m_pObject = pObject;
	if(pObject != NULL) {
		pObject->AddRef();
	}
}

// j܂B
ARJSPluginObject::ObjectNotifier::~ObjectNotifier()
{
	m_pOwner = NULL;
	if(m_pObject != NULL) {
		m_pObject->Release();
		m_pObject = NULL;
	}
}

// j܂B
void ARJSPluginObject::ObjectNotifier::Destroy()
{
	delete this;
}

// s܂B
HRESULT ARJSPluginObject::ObjectNotifier::Invoke()
{
	return m_pOwner->SetObject(m_hResult, m_pObject);
}
