#include "StdAfx.h"
#include "ARJSSolidCube.h"
#include <GL/gl.h>
#include <GL/glut.h>
#include <AR/gsub.h>
#include <AR/video.h>
#include <AR/param.h>
#include <AR/ar.h>
#include "../../js/JSValueWrapper.h"
#include "AbstractARJSObject.h"

// \z܂B
ARJSSolidCube::ARJSSolidCube(ARWindow* pOwner)
: ARJSGraphicsModel(pOwner)
{
	m_visible = true;
	m_size = 50.0;
	m_translate.z = m_size / 2.0;
	m_matAmbient[0] = 0.0;
	m_matAmbient[1] = 0.0;
	m_matAmbient[2] = 1.0;
	m_matAmbient[3] = 0.5;
	m_matFlash[0] = 0.0;
	m_matFlash[1] = 0.0;
	m_matFlash[2] = 1.0;
	m_matFlash[3] = 0.5;
	m_matFlashShiny[0] = 50.0;
}

// j܂B
ARJSSolidCube::~ARJSSolidCube(void)
{
	;
}

// j܂B
void ARJSSolidCube::Destroy()
{
	delete this;
}

// `悵܂B
HRESULT ARJSSolidCube::DrawModel()
{
	if(m_visible == false) {
		return S_FALSE;
	}

	glPushMatrix();
    glMaterialfv(GL_FRONT, GL_SPECULAR, m_matFlash);
    glMaterialfv(GL_FRONT, GL_SHININESS, m_matFlashShiny);	
    glMaterialfv(GL_FRONT, GL_AMBIENT, m_matAmbient);
    glMatrixMode(GL_MODELVIEW);
    glutSolidCube(m_size);

	HRESULT hr = DrawChildren();

	glPopMatrix();

	return hr;
}

// vpeB𐶐܂B
HRESULT ARJSSolidCube::CreateProperty(INamedProperty** ppProp)
{
	if(ppProp == NULL) {
		return E_POINTER;
	}
	CubePropertyImpl* pNewProp = new CubePropertyImpl(this);
	if(pNewProp == NULL) {
		return E_OUTOFMEMORY;
	}
	HRESULT hr = pNewProp->Init();
	if(FAILED(hr)) {
		delete pNewProp;
		pNewProp = NULL;
		return hr;
	}
	*ppProp = pNewProp;
	return S_OK;
}

// \z܂B
ARJSSolidCube::CubePropertyImpl::CubePropertyImpl(ARJSSolidCube* pOwner)
: ModelPropertyImpl(pOwner)
{
	m_pOwner = pOwner;
}

// j܂B
ARJSSolidCube::CubePropertyImpl::~CubePropertyImpl()
{
	m_pOwner = NULL;
}

// j܂B
void ARJSSolidCube::CubePropertyImpl::Destroy()
{
	delete this;
}

// vpeB擾܂B
HRESULT ARJSSolidCube::CubePropertyImpl::Init()
{
	HRESULT hr = S_OK;

	hr = DefineProperty("size", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("matFlashShiny", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("matAmbient.r", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("matAmbient.g", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("matAmbient.b", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("matAmbient.a", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("matFlash.r", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("matFlash.g", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("matFlash.b", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("matFlash.a", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	return ModelPropertyImpl::Init();
}

// doublẽvpeB擾܂B
HRESULT ARJSSolidCube::CubePropertyImpl::GetDoubleProperty(const char* name, double* pValue)
{
	if(name == NULL || pValue == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == "matFlashShiny") {
		GLfloat* pRefValue = NULL;
		pRefValue = m_pOwner->m_matFlashShiny;
		*pValue = *pRefValue;
		return S_OK;
	}else if(sname == "size") {
		GLfloat* pRefValue = NULL;
		pRefValue = &(m_pOwner->m_size);
		*pValue = *pRefValue;
		return S_OK;
	}

	string categ, prop;
	HRESULT hr = ParseName(name, categ, prop);
	if(FAILED(hr)) {
		return hr;
	}
	if(categ == "matAmbient") {
		GLfloat* pRefValue = NULL;
		hr = GetMember(m_pOwner->m_matAmbient, prop.c_str(), & pRefValue);
		if(FAILED(hr)) {
			return hr;
		}

		*pValue = *pRefValue;
		return S_OK;
	}else if(categ == "matFlash") {
		GLfloat* pRefValue = NULL;
		hr = GetMember(m_pOwner->m_matFlash, prop.c_str(), & pRefValue);
		if(FAILED(hr)) {
			return hr;
		}

		*pValue = *pRefValue;
		return S_OK;
	}else{
		return ModelPropertyImpl::GetDoubleProperty(name, pValue);
	}
}

// doublẽvpeBݒ肵܂B
HRESULT ARJSSolidCube::CubePropertyImpl::SetDoubleProperty(const char* name, double value)
{
	if(name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == "matFlashShiny") {
		GLfloat* pRefValue = NULL;
		pRefValue = m_pOwner->m_matFlashShiny;
		*pRefValue = value;
		return S_OK;
	}else if(sname == "size") {
		GLfloat* pRefValue = NULL;
		pRefValue = &(m_pOwner->m_size);
		*pRefValue = value;
		return S_OK;
	}
	string categ, prop;
	HRESULT hr = ParseName(name, categ, prop);
	if(FAILED(hr)) {
		return hr;
	}
	if(categ == "matAmbient") {
		GLfloat* pRefValue = NULL;
		hr = GetMember(m_pOwner->m_matAmbient, prop.c_str(), & pRefValue);
		if(FAILED(hr)) {
			return hr;
		}

		*pRefValue = value;
		return S_OK;
	}else if(categ == "matFlash") {
		GLfloat* pRefValue = NULL;
		hr = GetMember(m_pOwner->m_matFlash, prop.c_str(), & pRefValue);
		if(FAILED(hr)) {
			return hr;
		}

		*pRefValue = value;
		return S_OK;
	}else{
		return ModelPropertyImpl::SetDoubleProperty(name, value);
	}
}

// o擾܂B
HRESULT ARJSSolidCube::CubePropertyImpl::GetMember(GLfloat* pRGBA, const char* propName, GLfloat** ppValue)
{
	if(pRGBA == NULL || propName == NULL || ppValue == NULL) {
		return E_POINTER;
	}

	string spropName = propName;
	if(spropName == "r") {
		*ppValue = &(pRGBA[0]);
		return S_OK;
	}else if(spropName == "g") {
		*ppValue = &(pRGBA[1]);
		return S_OK;
	}else if(spropName == "b") {
		*ppValue = &(pRGBA[2]);
		return S_OK;
	}else if(spropName == "a") {
		*ppValue = &(pRGBA[3]);
		return S_OK;
	}
	return E_FAIL;
}

