#include "StdAfx.h"
#include "ARJSWorld.h"
#include <GL/gl.h>
#include <GL/glut.h>
#include <AR/gsub.h>
#include <AR/video.h>
#include <AR/param.h>
#include <AR/ar.h>
#include <AR/gsub_lite.h>
#include "ARJSSolidCube.h"
#include "ARJSImage.h"
#include "../ARWindow.h"
#include "../../js/JSValueWrapper.h"

#define VIEW_SCALEFACTOR_4		1.0 //4.0			// 1.0 ARToolKit unit becomes 4.0 of my OpenGL units.

// \z܂B
ARJSWorld::ARJSWorld(ARWindow* pWindow)
: ARGraphics(pWindow)
{
	m_pWindow = pWindow;
	m_pJSObject = NULL;
	ZeroMemory(& m_trans, sizeof(m_trans));
	ZeroMemory(& m_lightPosition, sizeof(m_lightPosition));
	ZeroMemory(& m_lightAmbient, sizeof(m_lightAmbient));
	ZeroMemory(& m_lightDiffuse, sizeof(m_lightDiffuse));

    m_lightPosition[0] = 100.0;
    m_lightPosition[1] = -200.0;
    m_lightPosition[2] = 200.0;
    m_lightPosition[3] = 0.0;
    m_lightAmbient[0] = m_lightAmbient[1] = m_lightAmbient[2] = 0.1;
	m_lightAmbient[3] = 0.1;
    m_lightDiffuse[0] = m_lightDiffuse[1] = m_lightDiffuse[2] = 0.9;
	m_lightDiffuse[3] = 0.1;
}

// j܂B
ARJSWorld::~ARJSWorld(void)
{
	if(m_pJSObject != NULL) {
		// thisJSObjectImpl͓Ct^C
		m_pJSObject->Destroy();
		m_pJSObject = NULL;
	}
	m_pWindow = NULL;
}

// j܂B
void ARJSWorld::Destroy()
{
	delete this;
}

// JavaScriptGWݒ肵܂B
HRESULT ARJSWorld::Attach(JSEngineContext* pContext)
{
	if(pContext == NULL) {
		return E_POINTER;
	}

	INamedProperty* pProp = NULL;
	HRESULT hr = GetProperty(& pProp);
	if(FAILED(hr)) {
		return hr;
	}

	m_pJSObject = new JSObjectImpl(this, pProp);
	if(m_pJSObject == NULL) {
		return E_OUTOFMEMORY;
	}

	hr = m_pJSObject->Attach(pContext, NULL);
	if(FAILED(hr)) {
		return hr;
	}

	IArrayProperty* pChildNodes = NULL;
	hr = GetChildren(& pChildNodes);
	if(FAILED(hr)) {
		return hr;
	}

	hr = m_childNodes.Attach(pChildNodes, pContext, m_pJSObject);
	if(FAILED(hr)) {
		return hr;
	}

	return S_OK;
}

// IJSObject擾܂B
IJSObject* ARJSWorld::GetJSObject()
{
	return m_pJSObject;
}

// `悵܂B
HRESULT ARJSWorld::Draw()
{
	if(m_visible == false) {
		return S_FALSE;
	}
    double    gl_para[16];
    
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    
	arglCameraViewRH(m_trans, gl_para, VIEW_SCALEFACTOR_4);
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixd( gl_para );

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_POSITION, m_lightPosition);
    glLightfv(GL_LIGHT0, GL_AMBIENT, m_lightAmbient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, m_lightDiffuse);

	HRESULT hr = DrawChildren();
	
	glDisable( GL_LIGHTING );

    glDisable( GL_DEPTH_TEST );

	return hr;
}

// vpeB𐶐܂B
HRESULT ARJSWorld::CreateProperty(INamedProperty** ppProp)
{
	if(ppProp == NULL) {
		return E_POINTER;
	}
	WorldPropertyImpl* pNewProp = new WorldPropertyImpl(this);
	if(pNewProp == NULL) {
		return E_OUTOFMEMORY;
	}
	HRESULT hr = pNewProp->Init();
	if(FAILED(hr)) {
		delete pNewProp;
		pNewProp = NULL;
		return hr;
	}
	*ppProp = pNewProp;
	return S_OK;
}

// \z܂B
ARJSWorld::JSObjectImpl::JSObjectImpl(ARJSWorld* pOwner, INamedProperty* pProp)
: AbstractARJSObject(pOwner, pProp)
{
	m_pOwner = pOwner;
}

// j܂B
ARJSWorld::JSObjectImpl::~JSObjectImpl()
{
	m_pOwner = NULL;
}

// j܂B
void ARJSWorld::JSObjectImpl::Destroy()
{
	delete this;
}

// o܂B
HRESULT ARJSWorld::JSObjectImpl::InitMembers()
{
	HRESULT hr = DefineProperty("translate", false);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("childNodes", true);
	if(FAILED(hr)) {
		return hr;
	}

	return AbstractARJSObject::InitMembers();
}

// vpeBݒ肵܂B
HRESULT ARJSWorld::JSObjectImpl::SetProperty(JSContext *cx, JSObject *obj, const char* name, jsval vp)
{
	string sname = name;

	if(sname == "translate") {
		if(vp == NULL) {
			return E_POINTER;
		}

		JSObject* pObj = NULL;
		JSBool bret = JS_ValueToObject(cx, vp , & pObj);
		if(bret == JS_FALSE) {
			return E_INVALIDARG;
		}
		if(JS_IsArrayObject(cx, pObj) == JS_FALSE) {
			return E_INVALIDARG;
		}

		jsuint len = 0;
		if(JS_GetArrayLength(cx, pObj, & len) == JS_FALSE) {
			return E_INVALIDARG;
		}
		if(len != 3) {
			return E_INVALIDARG;
		}

		for(jsuint i = 0; i < len; i ++) {
			jsval elemv = JSVAL_NULL;
			if(JS_GetElement(cx, pObj, i, & elemv) == JS_FALSE) {
				return E_INVALIDARG;
			}

			HRESULT hr = SetTransElement(cx, i, elemv);
			if(FAILED(hr)) {
				return hr;
			}
		}

		return S_OK;
	}else{
		return AbstractARJSObject::SetProperty(cx, obj, name, vp);
	}
}

// vpeB擾܂B
HRESULT ARJSWorld::JSObjectImpl::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	string sname = name;

	if(sname == "translate") {
		if(vp == NULL) {
			return S_OK;
		}
		jsval transValues[3];
		for(int i = 0; i < 3; i ++) {
			JSObject* pArray = NULL;
			HRESULT hr = CreateDoubleArray(m_pOwner->m_trans[i], 4, & pArray);
			if(FAILED(hr)) {
				return hr;
			}
			transValues[i] = OBJECT_TO_JSVAL(pArray);
		}
		JSObject* pTransObj = JS_NewArrayObject(cx, 3, transValues);
		if(pTransObj == NULL) {
			return E_FAIL;
		}
		*vp = OBJECT_TO_JSVAL(pTransObj);
		return S_OK;
	}else if(sname == "childNodes") {
		if(vp == NULL) {
			return S_OK;
		}
		IJSObject* pObj = NULL;
		HRESULT hr = m_pOwner->m_childNodes.GetJSObject(& pObj);
		if(FAILED(hr)) {
			return hr;
		}

		*vp = OBJECT_TO_JSVAL(pObj->GetCore());
		return S_OK;
	}else{
		return AbstractARJSObject::GetProperty(cx, obj, name, vp);
	}
}

// IDݒ肵܂B
HRESULT ARJSWorld::JSObjectImpl::SetTransElement(JSContext *cx, jsuint index, jsval arrv)
{
	JSObject* pObj = NULL;
	JSBool bret = JS_ValueToObject(cx, arrv, & pObj);
	if(bret == JS_FALSE) {
		return E_INVALIDARG;
	}
	if(JS_IsArrayObject(cx, pObj) == JS_FALSE) {
		return E_INVALIDARG;
	}

	jsuint len = 0;
	if(JS_GetArrayLength(cx, pObj, & len) == JS_FALSE) {
		return E_INVALIDARG;
	}
	if(len != 4) {
		return E_INVALIDARG;
	}

	for(jsuint i = 0; i < len; i ++) {
		jsval elemv = JSVAL_NULL;
		if(JS_GetElement(cx, pObj, i, & elemv) == JS_FALSE) {
			return E_INVALIDARG;
		}

		jsdouble v = 0.0;
		bret = JS_ValueToNumber(cx, elemv, & v);
		if(bret == JS_FALSE) {
			return E_INVALIDARG;
		}
		m_pOwner->m_trans[index][i] = v;
	}

	return S_OK;
}

// \z܂B
ARJSWorld::WorldPropertyImpl::WorldPropertyImpl(ARJSWorld* pOwner)
: GraphicsPropertyImpl(pOwner)
{
	m_pOwner = pOwner;
}

// j܂B
ARJSWorld::WorldPropertyImpl::~WorldPropertyImpl()
{
	m_pOwner = NULL;
}

// j܂B
void ARJSWorld::WorldPropertyImpl::Destroy()
{
	delete this;
}
		
// ܂B
HRESULT ARJSWorld::WorldPropertyImpl::Init()
{
	HRESULT hr = S_OK;

	hr = DefineProperty("lightPosition.x", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightPosition.y", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightPosition.z", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightPosition.w", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightAmbient.r", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightAmbient.g", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightAmbient.b", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightAmbient.a", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightDiffuse.r", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightDiffuse.g", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightDiffuse.b", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("lightDiffuse.a", PROP_DOUBLE);
	if(FAILED(hr)) {
		return hr;
	}

	return GraphicsPropertyImpl::Init();
}

// doublẽvpeB擾܂B
HRESULT ARJSWorld::WorldPropertyImpl::GetDoubleProperty(const char* name, double* pValue)
{
	if(name == NULL || pValue == NULL) {
		return E_POINTER;
	}
	string categ, prop;
	HRESULT hr = ParseName(name, categ, prop);
	if(FAILED(hr)) {
		return hr;
	}
	if(categ == "lightPosition") {
		if(prop == "x") {
			*pValue = m_pOwner->m_lightPosition[0];
		}else if(prop == "y") {
			*pValue = m_pOwner->m_lightPosition[1];
		}else if(prop == "z") {
			*pValue = m_pOwner->m_lightPosition[2];
		}else if(prop == "w") {
			*pValue = m_pOwner->m_lightPosition[3];
		}else{
			return E_FAIL;
		}
		return S_OK;
	}else if(categ == "lightAmbient") {
		GLfloat* pRefValue = NULL;
		hr = GetMember(m_pOwner->m_lightAmbient, prop.c_str(), & pRefValue);
		if(FAILED(hr)) {
			return hr;
		}

		*pValue = *pRefValue;
		return S_OK;
	}else if(categ == "lightDiffuse") {
		GLfloat* pRefValue = NULL;
		hr = GetMember(m_pOwner->m_lightDiffuse, prop.c_str(), & pRefValue);
		if(FAILED(hr)) {
			return hr;
		}

		*pValue = *pRefValue;
		return S_OK;
	}else{
		return GraphicsPropertyImpl::GetDoubleProperty(name, pValue);
	}
}

// doublẽvpeBݒ肵܂B
HRESULT ARJSWorld::WorldPropertyImpl::SetDoubleProperty(const char* name, double value)
{
	if(name == NULL) {
		return E_POINTER;
	}
	string categ, prop;
	HRESULT hr = ParseName(name, categ, prop);
	if(FAILED(hr)) {
		return hr;
	}
	if(categ == "lightPosition") {
		if(prop == "x") {
			m_pOwner->m_lightPosition[0] = value;
		}else if(prop == "y") {
			m_pOwner->m_lightPosition[1] = value;
		}else if(prop == "z") {
			m_pOwner->m_lightPosition[2] = value;
		}else if(prop == "w") {
			m_pOwner->m_lightPosition[3] = value;
		}else{
			return E_FAIL;
		}
		return S_OK;
	}else if(categ == "lightAmbient") {
		GLfloat* pRefValue = NULL;
		hr = GetMember(m_pOwner->m_lightAmbient, prop.c_str(), & pRefValue);
		if(FAILED(hr)) {
			return hr;
		}

		*pRefValue = value;
		return S_OK;
	}else if(categ == "lightDiffuse") {
		GLfloat* pRefValue = NULL;
		hr = GetMember(m_pOwner->m_lightDiffuse, prop.c_str(), & pRefValue);
		if(FAILED(hr)) {
			return hr;
		}

		*pRefValue = value;
		return S_OK;
	}else{
		return GraphicsPropertyImpl::SetDoubleProperty(name, value);
	}
}

// O͂܂B
HRESULT ARJSWorld::WorldPropertyImpl::ParseName(const char* name, string& categoryName, string& propName)
{
	if(name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	string::size_type dotPos = sname.find(".");
	if(dotPos == string::npos) {
		categoryName = "";
		propName = sname;
		return S_OK;
	}

	categoryName = sname.substr(0, dotPos);
	propName = sname.substr(dotPos + 1);
	return S_OK;
}

// o擾܂B
HRESULT ARJSWorld::WorldPropertyImpl::GetMember(float* pColor, const char* propName, GLfloat** ppValue)
{
	if(pColor == NULL || propName == NULL || ppValue == NULL) {
		return E_POINTER;
	}

	string spropName = propName;
	if(spropName == "r") {
		*ppValue = pColor + 0;
		return S_OK;
	}else if(spropName == "g") {
		*ppValue = pColor + 1;
		return S_OK;
	}else if(spropName == "b") {
		*ppValue = pColor + 2;
		return S_OK;
	}else if(spropName == "a") {
		*ppValue = pColor + 3;
		return S_OK;
	}
	return E_FAIL;
}
