#include "StdAfx.h"
#include "AbstractProperty.h"

// \z܂B
AbstractNamedProperty::AbstractNamedProperty()
{
	;
}

// j܂B
AbstractNamedProperty::~AbstractNamedProperty()
{
	;
}

// vpeBl擾܂B
HRESULT AbstractNamedProperty::GetPropertyValue(const char* name, IPropertyValue** ppValue)
{
	if(name == NULL || ppValue == NULL) {
		return E_POINTER;
	}
	*ppValue = new ValueImpl(this, name);
	return S_OK;
}

// vpeB擾܂B
HRESULT AbstractNamedProperty::GetPropertyCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = (int) m_names.size();
	return S_OK;
}

// vpeB擾܂B
HRESULT AbstractNamedProperty::GetPropertyName(int index, string& name)
{
	if(index >= (int)m_names.size()) {
		return E_INVALIDARG;
	}

	name = m_names[index];
	return S_OK;
}

// vpeB擾܂B
HRESULT AbstractNamedProperty::GetPropertyDesc(const char* name, PROPERTY_DESC* pDesc)
{
	if(m_descs.count(name) == 0) {
		return E_INVALIDARG;
	}

	PROPERTY_DESC desc = m_descs[name];
	memcpy_s(pDesc, sizeof(PROPERTY_DESC), & desc, sizeof(desc));
	return S_OK;
}

// ̃vpeB擾܂B
HRESULT AbstractNamedProperty::GetDoubleProperty(const char* name, double* pValue)
{
	return E_NOTIMPL;
}

// ̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::SetDoubleProperty(const char* name, double value)
{
	return E_NOTIMPL;
}
	
// bool̃vpeB擾܂B
HRESULT AbstractNamedProperty::GetBoolProperty(const char* name, bool* pValue)
{
	return E_NOTIMPL;
}
	
// bool̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::SetBoolProperty(const char* name, bool value)
{
	return E_NOTIMPL;
}

// int̃vpeB擾܂B
HRESULT AbstractNamedProperty::GetIntProperty(const char* name, int* pValue)
{
	return E_NOTIMPL;
}
	
// int̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::SetIntProperty(const char* name, int value)
{
	return E_NOTIMPL;
}

// string̃vpeB擾܂B
HRESULT AbstractNamedProperty::GetStringProperty(const char* name, string& value)
{
	return E_NOTIMPL;
}
	
// string̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::SetStringProperty(const char* name, const char* value)
{
	return E_NOTIMPL;
}

// IJSObject̃vpeB擾܂B
HRESULT AbstractNamedProperty::GetJSObject(const char* name, IJSObject** ppValue)
{
	return E_NOTIMPL;
}

// IJSObject̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::SetJSObject(const char* name, IJSObject* pValue)
{
	return E_NOTIMPL;
}

// vpeB`܂B
HRESULT AbstractNamedProperty::DefineProperty(const char* name, const PROPERTY_DESC* pDesc)
{
	PROPERTY_DESC desc;
	memcpy_s(& desc, sizeof(desc), pDesc, sizeof(PROPERTY_DESC));

	m_descs[name] = desc;
	m_names.push_back(name);

	return S_OK;
}

// vpeB`܂B
HRESULT AbstractNamedProperty::DefineProperty(const char* name, PROPERTY_TYPE type)
{
	PROPERTY_DESC desc;
	ZeroMemory(& desc, sizeof(desc));

	desc.type = type;

	return DefineProperty(name, & desc);
}

// \z܂B
AbstractNamedProperty::ValueImpl::ValueImpl(AbstractNamedProperty* pOwner, const char* name)
{
	m_pOwner = pOwner;
	m_name = name;
}

// j܂B
AbstractNamedProperty::ValueImpl::~ValueImpl()
{
	m_pOwner = NULL;
}

// j܂B
void AbstractNamedProperty::ValueImpl::Release()
{
	delete this;
}

// vpeB擾܂B
HRESULT AbstractNamedProperty::ValueImpl::GetPropertyDesc(PROPERTY_DESC* pDesc)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetPropertyDesc(m_name.c_str(), pDesc);
}

// ̃vpeB擾܂B
HRESULT AbstractNamedProperty::ValueImpl::GetDoubleProperty(double* pValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetDoubleProperty(m_name.c_str(), pValue);
}

// ̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::ValueImpl::SetDoubleProperty(double value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetDoubleProperty(m_name.c_str(), value);
}

// bool̃vpeB擾܂B
HRESULT AbstractNamedProperty::ValueImpl::GetBoolProperty(bool* pValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetBoolProperty(m_name.c_str(), pValue);
}

// bool̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::ValueImpl::SetBoolProperty(bool value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetBoolProperty(m_name.c_str(), value);
}

// int̃vpeB擾܂B
HRESULT AbstractNamedProperty::ValueImpl::GetIntProperty(int* pValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetIntProperty(m_name.c_str(), pValue);
}

// int̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::ValueImpl::SetIntProperty(int value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetIntProperty(m_name.c_str(), value);
}

// string̃vpeB擾܂B
HRESULT AbstractNamedProperty::ValueImpl::GetStringProperty(string& value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetStringProperty(m_name.c_str(), value);
}

// string̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::ValueImpl::SetStringProperty(const char* value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetStringProperty(m_name.c_str(), value);
}

// IJSObject̃vpeB擾܂B
HRESULT AbstractNamedProperty::ValueImpl::GetJSObject(IJSObject** ppValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetJSObject(m_name.c_str(), ppValue);
}

// IJSObject̃vpeBݒ肵܂B
HRESULT AbstractNamedProperty::ValueImpl::SetJSObject(IJSObject* pValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetJSObject(m_name.c_str(), pValue);
}


// \z܂B
AbstractArrayProperty::AbstractArrayProperty()
{
	m_pNotify = NULL;
}

// j܂B
AbstractArrayProperty::~AbstractArrayProperty()
{
	m_pNotify = NULL;
}

// ʒmIuWFNgݒ肵܂B
HRESULT AbstractArrayProperty::SetNotify(IArrayPropertyNotify* pNotify)
{
	m_pNotify = pNotify;
	return S_OK;
}

// vpeBl擾܂B
HRESULT AbstractArrayProperty::GetPropertyValue(int index, IPropertyValue** ppValue)
{
	if(ppValue == NULL) {
		return E_POINTER;
	}
	*ppValue = new ValueImpl(this, index);
	return S_OK;
}
	
// vpeB擾܂B
HRESULT AbstractArrayProperty::GetPropertyDesc(int index, PROPERTY_DESC* pDesc)
{
	return E_NOTIMPL;
}

// ̃vpeB擾܂B
HRESULT AbstractArrayProperty::GetDoubleProperty(int index, double* pValue)
{
	return E_NOTIMPL;
}

// ̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::SetDoubleProperty(int index, double value)
{
	return E_NOTIMPL;
}
	
// bool̃vpeB擾܂B
HRESULT AbstractArrayProperty::GetBoolProperty(int index, bool* pValue)
{
	return E_NOTIMPL;
}
	
// bool̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::SetBoolProperty(int index, bool value)
{
	return E_NOTIMPL;
}

// int̃vpeB擾܂B
HRESULT AbstractArrayProperty::GetIntProperty(int index, int* pValue)
{
	return E_NOTIMPL;
}
	
// int̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::SetIntProperty(int index, int value)
{
	return E_NOTIMPL;
}

// string̃vpeB擾܂B
HRESULT AbstractArrayProperty::GetStringProperty(int index, string& value)
{
	return E_NOTIMPL;
}
	
// string̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::SetStringProperty(int index, const char* value)
{
	return E_NOTIMPL;
}

// IJSObject̃vpeB擾܂B
HRESULT AbstractArrayProperty::GetJSObject(int index, IJSObject** ppValue)
{
	return E_NOTIMPL;
}

// IJSObject̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::SetJSObject(int index, IJSObject* pValue)
{
	return E_NOTIMPL;
}

// ύXʒm܂B
HRESULT AbstractArrayProperty::NotifyChanged()
{
	if(m_pNotify == NULL) {
		return S_OK;
	}
	return m_pNotify->NotifyChanged(this);
}

// \z܂B
AbstractArrayProperty::ValueImpl::ValueImpl(AbstractArrayProperty* pOwner, int index)
{
	m_pOwner = pOwner;
	m_iIndex = index;
}

// j܂B
AbstractArrayProperty::ValueImpl::~ValueImpl()
{
	m_pOwner = NULL;
}

// j܂B
void AbstractArrayProperty::ValueImpl::Release()
{
	delete this;
}

// vpeB擾܂B
HRESULT AbstractArrayProperty::ValueImpl::GetPropertyDesc(PROPERTY_DESC* pDesc)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetPropertyDesc(m_iIndex, pDesc);
}

// ̃vpeB擾܂B
HRESULT AbstractArrayProperty::ValueImpl::GetDoubleProperty(double* pValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetDoubleProperty(m_iIndex, pValue);
}

// ̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::ValueImpl::SetDoubleProperty(double value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetDoubleProperty(m_iIndex, value);
}

// bool̃vpeB擾܂B
HRESULT AbstractArrayProperty::ValueImpl::GetBoolProperty(bool* pValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetBoolProperty(m_iIndex, pValue);
}

// bool̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::ValueImpl::SetBoolProperty(bool value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetBoolProperty(m_iIndex, value);
}

// int̃vpeB擾܂B
HRESULT AbstractArrayProperty::ValueImpl::GetIntProperty(int* pValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetIntProperty(m_iIndex, pValue);
}

// int̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::ValueImpl::SetIntProperty(int value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetIntProperty(m_iIndex, value);
}

// string̃vpeB擾܂B
HRESULT AbstractArrayProperty::ValueImpl::GetStringProperty(string& value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetStringProperty(m_iIndex, value);
}

// string̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::ValueImpl::SetStringProperty(const char* value)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetStringProperty(m_iIndex, value);
}

// IJSObject̃vpeB擾܂B
HRESULT AbstractArrayProperty::ValueImpl::GetJSObject(IJSObject** ppValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->GetJSObject(m_iIndex, ppValue);
}

// IJSObject̃vpeBݒ肵܂B
HRESULT AbstractArrayProperty::ValueImpl::SetJSObject(IJSObject* pValue)
{
	if(m_pOwner == NULL) {
		return E_UNEXPECTED;
	}
	return m_pOwner->SetJSObject(m_iIndex, pValue);
}

