#include "StdAfx.h"
#include "DiscPath.h"
#include "URLPath.h"

// \z܂B
DiscPath::DiscPath(const char* path, HRESULT* phr)
{
	*phr = S_OK;

	HRESULT hr = Parse(path);
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// \z܂B
DiscPath::DiscPath(vector<string>& elems, HRESULT* phr)
{
	*phr = S_OK;

	for(vector<string>::iterator it = elems.begin(); it != elems.end(); it++) {
		m_elements.push_back(*it);
	}
}

// j܂B
DiscPath::~DiscPath()
{
	;
}

// pXvf̐擾܂B
HRESULT DiscPath::GetPathElementCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = (int) m_elements.size();
	return S_OK;
}

// pXvf擾܂B
HRESULT DiscPath::GetPathElement(int index, string& pathElem)
{
	if(index >= (int)m_elements.size()) {
		return E_INVALIDARG;
	}
	pathElem = m_elements[index];
	return S_OK;
}

// ΃pXǂ擾܂B
HRESULT DiscPath::IsAbsolute(bool* pAbsolute)
{
	if(pAbsolute == NULL) {
		return E_POINTER;
	}

	*pAbsolute = false;
	if(m_elements.size() < 1) {
		return S_OK;
	}
	string firstElem = m_elements[0];
	if(firstElem.length() == 2) {
		char ch1 = firstElem[0];
		char ch2 = firstElem[1];

		if((('A' <= ch1 && ch1 <= 'Z') || ('a' <= ch1 && ch1 <= 'z')) && ch2 == ':') {
			*pAbsolute = true;
		}
	}
	return S_OK;
}

// pX폜܂B
HRESULT DiscPath::RemoveLast()
{
	if(m_elements.size() == 0) {
		return S_FALSE;
	}
	vector<string>::iterator it = m_elements.begin();
	int count = m_elements.size() - 1;
	m_elements.erase(it + count);
	return S_OK;
}

// pXɒǉ܂B
HRESULT DiscPath::AddLast(const char* element)
{
	if(element == NULL) {
		return E_POINTER;
	}
	m_elements.push_back(element);
	return S_OK;
}

// 񉻂܂B
HRESULT DiscPath::ToString(string& path)
{
	int count = 0;

	string buf = "";
	for(vector<string>::iterator it = m_elements.begin();
		it != m_elements.end();
		it ++) {
		if(count > 0) {
			buf += "\\";
		}
		string elem = (*it);
		HRESULT hr = URLPath::DecodeURLPath(elem.c_str(), elem);
		if(FAILED(hr)) {
			return hr;
		}
		buf += elem;
		count ++;
	}

	path = buf;
	return S_OK;
}

// URL擾܂B
HRESULT DiscPath::GetURL(string& url)
{
	HRESULT hr = S_OK;
	bool bAbsolute = false;

	hr = IsAbsolute(& bAbsolute);
	if(FAILED(hr)) {
		return hr;
	}

	if(bAbsolute == true) {
		vector<string> eelems;
		int count = 0;
		for(vector<string>::iterator it = m_elements.begin(); it != m_elements.end(); it ++, count ++) {
			string elem = *it;
			if(count > 0) {
				hr = URLPath::EncodeURLPath(elem.c_str(), elem);
				if(FAILED(hr)) {
					return hr;
				}
			}
			eelems.push_back(elem);
		}
		URLPath urlpath("file", NULL, 0, eelems, & hr);
		if(FAILED(hr)) {
			return hr;
		}
		hr = urlpath.AddFirstPathElement("");
		if(FAILED(hr)) {
			return hr;
		}
		return urlpath.ToString(url);
	}else{
		vector<string> eelems;
		for(vector<string>::iterator it = m_elements.begin(); it != m_elements.end(); it ++) {
			string elem = *it;
			hr = URLPath::EncodeURLPath(elem.c_str(), elem);
			if(FAILED(hr)) {
				return hr;
			}
			eelems.push_back(elem);
		}
		URLPath urlpath(NULL, NULL, 0, eelems, & hr);
		if(FAILED(hr)) {
			return hr;
		}
		return urlpath.ToString(url);
	}
}

// ͂܂B
HRESULT DiscPath::Parse(const char* path)
{
	if(path == NULL) {
		return E_POINTER;
	}
	string buf = "";
	string spath = path;

	for(string::iterator it = spath.begin(); it != spath.end(); it ++) {
		char ch = *it;

		if(ch == '\\' || ch == '/') {
			m_elements.push_back(buf);
			buf = "";
		}else{
			buf += ch;
		}
	}

	m_elements.push_back(buf);
	return S_OK;
}
