#include "StdAfx.h"
#include "DispatchWrapper.h"

// \z܂B
DispatchWrapper::DispatchWrapper(IDispatch* pDispatch, HRESULT* phr)
{
	m_pDispatch = pDispatch;
	*phr = S_OK;

	if(pDispatch == NULL) {
		*phr = E_POINTER;
		return;
	}

	m_pDispatch = pDispatch;
	m_pDispatch->AddRef();

	HRESULT hr = EnumMembers();
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
DispatchWrapper::~DispatchWrapper(void)
{
	for(vector<Member*>::iterator it = m_members.begin();
		it != m_members.end();
		it ++) {
		Member* pMember = *it;
		delete pMember;
		pMember = NULL;
	}
	m_members.clear();

	if(m_pDispatch != NULL) {
		m_pDispatch->Release();
		m_pDispatch = NULL;
	}
}

// IDispatch擾܂B
HRESULT DispatchWrapper::GetDispatch(IDispatch** ppDisp)
{
	if(ppDisp == NULL) {
		return E_POINTER;
	}
	*ppDisp = m_pDispatch;
	m_pDispatch->AddRef();
	return S_OK;
}

// o擾܂B
HRESULT DispatchWrapper::GetMemberCount(UINT* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = m_members.size();
	return S_OK;
}

// o擾܂B
HRESULT DispatchWrapper::GetMemberName(UINT index, string& name)
{
	if(index >= m_members.size()) {
		return E_INVALIDARG;
	}
	Member* pMember = m_members[index];
	return pMember->GetName(name);
}

// o^Cv擾܂B
HRESULT DispatchWrapper::GetMemberWFlags(UINT index, WORD* pWFlags)
{
	if(index >= m_members.size()) {
		return E_INVALIDARG;
	}
	Member* pMember = m_members[index];
	return pMember->GetWFlags(pWFlags);
}

// p[^擾܂B
HRESULT DispatchWrapper::GetMemberParamCount(UINT memberIndex, UINT* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	if(memberIndex >= m_members.size()) {
		return E_INVALIDARG;
	}
	Member* pMember = m_members[memberIndex];
	return pMember->GetParamCount(pCount);
}

// p[^擾܂B
HRESULT DispatchWrapper::GetMemberParamType(UINT memberIndex, UINT paramIndex, VARTYPE* pVarType)
{
	if(pVarType == NULL) {
		return E_POINTER;
	}
	if(memberIndex >= m_members.size()) {
		return E_INVALIDARG;
	}
	Member* pMember = m_members[memberIndex];
	return pMember->GetParamType(paramIndex, pVarType);
}

// ߂l擾܂B
HRESULT DispatchWrapper::GetMemberReturnType(UINT memberIndex, VARTYPE* pVarType)
{
	if(pVarType == NULL) {
		return E_POINTER;
	}
	if(memberIndex >= m_members.size()) {
		return E_INVALIDARG;
	}
	Member* pMember = m_members[memberIndex];
	return pMember->GetReturnType(pVarType);
}

// os܂B
HRESULT DispatchWrapper::Invoke(UINT memberIndex, DISPPARAMS* pParams, VARIANT* pResult)
{
	if(memberIndex >= m_members.size()) {
		return E_INVALIDARG;
	}
	Member* pMember = m_members[memberIndex];
	return pMember->Invoke(pParams, pResult);
}

// ܂B
HRESULT DispatchWrapper::FindMember(const char* name, WORD wFlags, UINT* pIndex)
{
	if(name == NULL || pIndex == NULL) {
		return E_POINTER;
	}
	for(UINT i = 0; i < m_members.size(); i ++) {
		Member* pMember = m_members[i];
		string memName;
		HRESULT hr = pMember->GetName(memName);
		if(FAILED(hr)) {
			return hr;
		}

		WORD memWFlags = 0;
		hr = pMember->GetWFlags(& memWFlags);
		if(FAILED(hr)) {
			return hr;
		}

		string sname = name;
		if(memName == sname && memWFlags == wFlags) {
			*pIndex = i;
			return S_OK;
		}
	}
	return E_INVALIDARG;
}

// BSTRstringIuWFNg擾܂B
HRESULT DispatchWrapper::GetStringFromBSTR(string& value, BSTR bstr)
{
	if(bstr == NULL) {
		return E_POINTER;
	}

	UINT len = SysStringLen(bstr);
	if(len == 0) {
		value = "";
		return S_OK;
	}

	int bufSize = len * 2 + 1;
	char* pBuf = new char[bufSize];
	if(pBuf == NULL) {
		return E_OUTOFMEMORY;
	}

	int nret = WideCharToMultiByte(CP_ACP, 0, bstr, len, pBuf, bufSize, NULL, NULL);
	pBuf[nret] = '\0';

	value = pBuf;

	delete[] pBuf;
	pBuf = NULL;

	return S_OK;
}

// stringBSTRIuWFNg擾܂B
HRESULT DispatchWrapper::GetBSTRFromString(BSTR* pBSTR, const char* value)
{
	if(pBSTR == NULL || value == NULL) {
		return E_POINTER;
	}
	int destLen = (int)MultiByteToWideChar(CP_ACP, 0, value, strlen(value), NULL, 0);

	BSTR bstr = SysAllocStringLen(NULL, destLen);

	MultiByteToWideChar(CP_ACP, 0, value, strlen(value), bstr, destLen);

	*pBSTR = bstr;
	return S_OK;
}

// o񋓂܂B
HRESULT DispatchWrapper::EnumMembers()
{
	UINT count = 0;
	HRESULT hr = m_pDispatch->GetTypeInfoCount(& count);
	if(FAILED(hr)) {
		return hr;
	}

	for(UINT i = 0; i < count; i ++) {
		ITypeInfo* pTypeInfo = NULL;
		hr = m_pDispatch->GetTypeInfo(i, GetUserDefaultLCID(), & pTypeInfo);
		if(FAILED(hr)) {
			return hr;
		}
		TYPEATTR* pTypeAttr = NULL;
		hr = pTypeInfo->GetTypeAttr(& pTypeAttr);
		if(FAILED(hr)) {
			return hr;
		}

		GUID guid = pTypeAttr->guid;
		pTypeInfo->ReleaseTypeAttr(pTypeAttr);
		pTypeAttr = NULL;

		if(guid == IID_IUnknown || guid == IID_IDispatch) {
			pTypeInfo->Release();
			pTypeInfo = NULL;
			continue;
		}

		UINT index = 0;
		hr = S_OK;
		while(hr == S_OK) {
			FUNCDESC* pFuncDesc = NULL;
			hr = pTypeInfo->GetFuncDesc(index, & pFuncDesc);
			if(hr == TYPE_E_ELEMENTNOTFOUND || hr == S_FALSE) {
				break; 
			}
			if(FAILED(hr)) {
				pTypeInfo->Release();
				pTypeInfo = NULL;
				return hr;
			}
			
			MEMBERID memid = pFuncDesc->memid;
			INVOKEKIND kind = pFuncDesc->invkind;
			pTypeInfo->ReleaseFuncDesc(pFuncDesc);
			pFuncDesc = NULL;

			if(kind == INVOKE_FUNC || kind == INVOKE_PROPERTYGET || kind == INVOKE_PROPERTYPUT) {
				Member* pNewMember = new Member(this, pTypeInfo, index, memid, kind, & hr);
				if(FAILED(hr)) {
					pTypeInfo->Release();
					pTypeInfo = NULL;
					return hr;
				}
				m_members.push_back(pNewMember);
			}

			index ++;
		}

		pTypeInfo->Release();
		pTypeInfo = NULL;
	}

	return S_OK;
}

// \z܂B
DispatchWrapper::Member::Member(DispatchWrapper* pOwner, ITypeInfo* pTypeInfo, UINT index, MEMBERID memid, INVOKEKIND kind, HRESULT* phr)
{
	m_pOwner = pOwner;
	m_name = "";
	m_iIndex = index;
	m_pTypeInfo = pTypeInfo;
	m_pTypeInfo->AddRef();
	m_memId = memid;
	m_returnType = VT_NULL;
	m_kind = kind;
	*phr = S_OK;

	HRESULT hr = Init();
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
DispatchWrapper::Member::~Member()
{
	m_pOwner = NULL;

	if(m_pTypeInfo != NULL) {
		m_pTypeInfo->Release();
		m_pTypeInfo = NULL;
	}
}

// O擾܂B
HRESULT DispatchWrapper::Member::GetName(string& name)
{
	name = m_name;
	return S_OK;
}

// p[^擾܂B
HRESULT DispatchWrapper::Member::GetParamCount(UINT* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = m_params.size();
	return S_OK;
}

// p[^擾܂B
HRESULT DispatchWrapper::Member::GetParamType(UINT index, VARTYPE* pVarType)
{
	if(pVarType == NULL) {
		return E_POINTER;
	}
	if(index >= m_params.size()) {
		return E_INVALIDARG;
	}
	*pVarType = m_params[index];
	return S_OK;
}

// ߂l擾܂B
HRESULT DispatchWrapper::Member::GetReturnType(VARTYPE* pVarType)
{
	if(pVarType == NULL) {
		return E_POINTER;
	}
	*pVarType = m_returnType;
	return S_OK;
}

// \bhs܂B
HRESULT DispatchWrapper::Member::Invoke(DISPPARAMS *pParams, VARIANT *pResult)
{
	IDispatch* pDispatch = m_pOwner->m_pDispatch;

	BSTR bstrName = NULL;
	HRESULT hr = m_pOwner->GetBSTRFromString(& bstrName, m_name.c_str());
	if(FAILED(hr)) {
		return hr;
	}

	DISPID dispId = 0;
	hr = pDispatch->GetIDsOfNames(IID_NULL, & bstrName, 1, GetUserDefaultLCID(), & dispId);
	SysFreeString(bstrName);
	if(FAILED(hr)) {
		return hr;
	}

	EXCEPINFO excepInfo;
	ZeroMemory(& excepInfo, sizeof(excepInfo));
	UINT uArgErr = 0;
	WORD wFlags = DISPATCH_METHOD;

	if(m_kind == INVOKE_FUNC) {
		wFlags = DISPATCH_METHOD;
	}else if(m_kind == INVOKE_PROPERTYGET) {
		wFlags = DISPATCH_PROPERTYGET;
	}else if(m_kind == INVOKE_PROPERTYPUT) {
		wFlags = DISPATCH_PROPERTYPUT;
	}

	hr = pDispatch->Invoke(dispId, IID_NULL, GetUserDefaultLCID(), wFlags, pParams, pResult, & excepInfo, & uArgErr);
	if(FAILED(hr)) {
		return hr;
	}

	return S_OK;
}

// Ăяo^Cv擾܂B
HRESULT DispatchWrapper::Member::GetWFlags(WORD* pWFlags)
{
	if(pWFlags == NULL) {
		return E_POINTER;
	}
	WORD wFlags = DISPATCH_METHOD;
	if(m_kind == INVOKE_FUNC) {
		wFlags = DISPATCH_METHOD;
	}else if(m_kind == INVOKE_PROPERTYGET) {
		if(m_params.size() > 0) {
			wFlags = DISPATCH_METHOD;
		}else{
			wFlags = DISPATCH_PROPERTYGET;
		}
	}else if(m_kind == INVOKE_PROPERTYPUT) {
		wFlags = DISPATCH_PROPERTYPUT;
	}
	*pWFlags = wFlags;
	return S_OK;
}

// ܂B
HRESULT DispatchWrapper::Member::Init()
{
	HRESULT hr = InitName();
	if(FAILED(hr)) {
		return hr;
	}

	return InitParams();
}

// O܂B
HRESULT DispatchWrapper::Member::InitName()
{
	const UINT MAX_NAMES = 256;
	BSTR* pNames = new BSTR[MAX_NAMES];
	ZeroMemory(pNames, sizeof(pNames));
	UINT nameCount = 0;

	HRESULT hr = m_pTypeInfo->GetNames(m_memId, pNames, MAX_NAMES, & nameCount);
	if(FAILED(hr)) {
		return hr;
	}
	if(nameCount == 0) {
		delete[] pNames;
		pNames = NULL;
		return S_FALSE;
	}

	string sname;
	hr = m_pOwner->GetStringFromBSTR(sname,pNames[0]);
	if(hr == S_OK) {
		m_name = sname;
	}

	for(UINT k = 0; k < nameCount; k ++) {
		BSTR pName = pNames[k];
		SysFreeString(pName);
		pNames[k] = NULL;
	}
	delete[] pNames;
	pNames = NULL;

	return hr;
}

// p[^܂B
HRESULT DispatchWrapper::Member::InitParams()
{
	FUNCDESC* pFuncDesc = NULL;
	HRESULT hr = m_pTypeInfo->GetFuncDesc(m_iIndex, & pFuncDesc);
	if(FAILED(hr)) {
		return hr;
	}

	for(int i = 0; i < pFuncDesc->cParams; i ++) {
		ELEMDESC* pElemDesc = &(pFuncDesc->lprgelemdescParam[i]);
		TYPEDESC typedesc = pElemDesc->tdesc;
		m_params.push_back(typedesc.vt);
	}

	m_returnType = pFuncDesc->elemdescFunc.tdesc.vt;
	
	m_pTypeInfo->ReleaseFuncDesc(pFuncDesc);
	pFuncDesc = NULL;

	return S_OK;
}
