#include "StdAfx.h"
#include "GLImageDump.h"
#include "../common/CritSection.h"
#include <gl/glut.h>
#include <mmsystem.h>

// Xbh̑ҋ@^CAEgłB
#define PROC_TIMEOUT 1000

// \z܂B
GLImageDump::GLImageDump(IWindow* pWindow, int bufferSize, HRESULT* phr)
{
	if(pWindow == NULL || bufferSize <= 0) {
		*phr = E_POINTER;
		return;
	}
	m_pWindow = pWindow;
	m_hEvBuffer = NULL;
	m_pFreeBuf = NULL;
	m_pPostBuf = NULL;
	m_pWorker = NULL;
	m_hStopEvent = NULL;

	HRESULT hr = Init(bufferSize);
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
GLImageDump::~GLImageDump()
{
	Stop();

	if(m_pFreeBuf != NULL) {
		delete m_pFreeBuf;
		m_pFreeBuf = NULL;
	}
	if(m_pPostBuf != NULL) {
		delete m_pPostBuf;
		m_pPostBuf = NULL;
	}
	if(m_hEvBuffer != NULL) {
		CloseHandle(m_hEvBuffer);
		m_hEvBuffer = NULL;
	}
}

// j܂B
void GLImageDump::Destroy()
{
	delete this;
}

// `̑OB
HRESULT GLImageDump::OnPreDraw()
{
	return S_OK;
}

// `̌㏈B
HRESULT GLImageDump::OnPostDraw()
{
	Buffer* pBuffer = NULL;
	HRESULT hr = PopForDraw(& pBuffer);
	if(FAILED(hr)) {
		return hr;
	}
	if(hr == S_FALSE) {
		return S_OK;
	}

	LONGLONG t = 0;
	int w = 0, h = 0;
	unsigned char* data = NULL;

	hr = pBuffer->GetBuffer(& t, & w, & h, & data);
	if(hr == S_OK) {
		int align = 4;
		glGetIntegerv(GL_PACK_ALIGNMENT, & align);

		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glReadBuffer(GL_BACK);

		// obt@̓e
		// bmpIuWFNg̃sNZf[^i[Ă̈ɒڃRs[B
		glReadPixels(
					  0,                 //ǂݎ̈̍xW
					  0,                 //ǂݎ̈̍yW
					  w,             //ǂݎ̈̕
					  h,            //ǂݎ̈̍
					  GL_RGB,           //擾F̌`
					  GL_UNSIGNED_BYTE,  //ǂݎf[^ۑž^
					  data      //rbg}bṽsNZf[^iۂɂ̓oCgzjւ̃|C^
					);
		
		// Ôߌɖ߂
		glPixelStorei(GL_PACK_ALIGNMENT, align);
	}

	return PushForProcess(pBuffer);
}

// Jn܂B
HRESULT GLImageDump::Start()
{
	if(m_pWorker != NULL || m_hStopEvent != NULL) {
		return E_UNEXPECTED;
	}

	m_hStopEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if(m_hStopEvent == NULL) {
		return E_UNEXPECTED;
	}
	m_pWorker = new ProcessWorker(this);

	return m_workerThread.Start(m_pWorker);
}

// ~܂B
HRESULT GLImageDump::Stop()
{
	if(m_pWorker == NULL || m_hStopEvent == NULL) {
		return E_UNEXPECTED;
	}

	m_pWorker->Stop();

	WaitForSingleObject(m_hStopEvent, INFINITE);
	CloseHandle(m_hStopEvent);
	m_hStopEvent = NULL;
	m_pWorker = NULL; // Worker͎Iɏ

	return S_OK;
}

// ܂B
HRESULT GLImageDump::Init(int bufferSize)
{
	m_hEvBuffer = CreateEvent(NULL, FALSE, FALSE, NULL);
	if(m_hEvBuffer == NULL) {
		return E_UNEXPECTED;
	}

	int x = 0, y = 0, w = 0, h = 0;
	HRESULT hr = m_pWindow->GetRect(& x, & y, & w, & h);
	if(FAILED(hr)) {
		return hr;
	}

	Buffer* pLastBuf = NULL;
	for(int i = 0; i < bufferSize; i ++) {
		Buffer* pBuf = new Buffer(w, h);
		if(pLastBuf == NULL) {
			m_pFreeBuf = pBuf;
		}else{
			hr = pLastBuf->SetNext(pBuf);
			if(FAILED(hr)) {
				return hr;
			}
		}

		pLastBuf = pBuf;
	}
	return S_OK;
}

// \z܂B
GLImageDump::Buffer::Buffer(int width, int height)
{
	m_iWidth = width;
	m_iHeight = height;
	m_pBuffer = new unsigned char[width * height * 4];
	m_pNext = NULL;
	m_llLastUpdated = 0;
}

// j܂B
GLImageDump::Buffer::~Buffer()
{
	if(m_pBuffer != NULL) {
		delete[] m_pBuffer;
		m_pBuffer = NULL;
	}
	if(m_pNext != NULL) {
		delete m_pNext;
		m_pNext = NULL;
	}
}

// ̃obt@ݒ肵܂B
HRESULT GLImageDump::Buffer::SetNext(Buffer* pNext)
{
	m_pNext = pNext;
	return S_OK;
}

// ̃obt@擾܂B
HRESULT GLImageDump::Buffer::GetNext(Buffer** ppNext)
{
	if(ppNext == NULL) {
		return E_POINTER;
	}
	*ppNext = m_pNext;
	return S_OK;
}

// obt@擾܂B
HRESULT GLImageDump::Buffer::GetBuffer(LONGLONG* pLastUpdated, int* pWidth, int* pHeight, unsigned char** ppBuffer)
{
	if(pLastUpdated == NULL || pWidth == NULL || pHeight == NULL || ppBuffer == NULL) {
		return E_POINTER;
	}
	*pLastUpdated = m_llLastUpdated;
	*pWidth = m_iWidth;
	*pHeight = m_iHeight;
	*ppBuffer = m_pBuffer;
	return S_OK;
}

// eXV܂B
HRESULT GLImageDump::Buffer::Update()
{
	m_llLastUpdated = timeGetTime();
	return S_OK;
}

// \z܂B
GLImageDump::ProcessWorker::ProcessWorker(GLImageDump* pOwner)
{
	m_pOwner = pOwner;
	m_bContinue = true;
}

// j܂B
GLImageDump::ProcessWorker::~ProcessWorker()
{
	;
}

// j܂B
void GLImageDump::ProcessWorker::Destroy()
{
	delete this;
}

// s܂B
HRESULT GLImageDump::ProcessWorker::Invoke()
{
	while(IsContinue()) {
		Buffer* pBuffer = NULL;
		HRESULT hr = m_pOwner->PopForProcess(& pBuffer);
		if(FAILED(hr)) {
			return hr;
		}
		if(hr == S_FALSE) {
			WaitForSingleObject(m_pOwner->m_hEvBuffer, PROC_TIMEOUT);
			continue;
		}

		LONGLONG t = 0;
		int w = 0, h = 0;
		unsigned char* pBuf = NULL;
		hr = pBuffer->GetBuffer(& t, & w, & h, & pBuf);
		if(FAILED(hr)) {
			return hr;
		}

		hr = m_pOwner->Process(t, w, h, pBuf);
		if(FAILED(hr)) {
			return hr;
		}

		hr = m_pOwner->PushForDraw(pBuffer);
		if(FAILED(hr)) {
			return hr;
		}
	}
	SetEvent(m_pOwner->m_hStopEvent);
	return S_OK;
}

// ~܂B
void GLImageDump::ProcessWorker::Stop()
{
	CritSection cs(& m_continueMutex);
	m_bContinue = false;
}

// p邩ǂێ܂B
bool GLImageDump::ProcessWorker::IsContinue()
{
	CritSection cs(& m_continueMutex);
	return m_bContinue;
}

// `pɃobt@擾܂B
HRESULT GLImageDump::PopForDraw(Buffer** ppBuffer)
{
	if(ppBuffer == NULL) {
		return E_POINTER;
	}

	CritSection cs(& m_bufferMutex);
	*ppBuffer = m_pFreeBuf;
	if(*ppBuffer == NULL) {
		return S_FALSE;
	}

	Buffer* pNextBuf = NULL;
	HRESULT hr = (*ppBuffer)->GetNext(& pNextBuf);
	m_pFreeBuf = pNextBuf;
	if(FAILED(hr)) {
		return hr;
	}

	return (*ppBuffer)->SetNext(NULL);
}

// pɃobt@擾܂B
HRESULT GLImageDump::PopForProcess(Buffer** ppBuffer)
{
	if(ppBuffer == NULL) {
		return E_POINTER;
	}
	CritSection cs(& m_bufferMutex);
	*ppBuffer = m_pPostBuf;
	if(*ppBuffer == NULL) {
		return S_FALSE;
	}

	Buffer* pNextBuf = NULL;
	HRESULT hr = (*ppBuffer)->GetNext(& pNextBuf);
	m_pPostBuf = pNextBuf;
	if(FAILED(hr)) {
		return hr;
	}

	return (*ppBuffer)->SetNext(NULL);
}

// `pɃobt@ǉ܂B
HRESULT GLImageDump::PushForDraw(Buffer* pBuffer)
{
	if(pBuffer == NULL) {
		return E_POINTER;
	}

	// 擪ɐڑ
	CritSection cs(& m_bufferMutex);
	Buffer* pNextBuffer = m_pFreeBuf;
	HRESULT hr = pBuffer->SetNext(pNextBuffer);
	m_pFreeBuf = pBuffer;

	return hr;
}

// pɃobt@ǉ܂B
HRESULT GLImageDump::PushForProcess(Buffer* pBuffer)
{
	if(pBuffer == NULL) {
		return E_POINTER;
	}

	// ɐڑ
	CritSection cs(& m_bufferMutex);
	Buffer* pNextBuffer = m_pPostBuf;
	while(pNextBuffer != NULL) {
		Buffer* pBuf = NULL;
		HRESULT hr = pNextBuffer->GetNext(& pBuf);
		if(FAILED(hr)) {
			return hr;
		}
		if(pBuf == NULL) {
			break;
		}
		pNextBuffer = pBuf;
	}
	if(pNextBuffer == NULL) {
		m_pPostBuf = pBuffer;
		SetEvent(m_hEvBuffer);
		return pBuffer->Update();
	}
	SetEvent(m_hEvBuffer);

	HRESULT hr = pNextBuffer->SetNext(pBuffer);
	if(FAILED(hr)) {
		return hr;
	}
	return pBuffer->Update();
}
