#pragma once

#include "../window/IWindow.h"
#include "../common/WorkerThread.h"
#include "../common/Mutex.h"

/**
* OpenGL̉摜_v邽߂̒ۃNXłB
*/
class GLImageDump : public IWindowGLHook
{
public:

	/**
	* \z܂B
	*/
	GLImageDump(IWindow* pWindow, int bufferSize, HRESULT* phr);

	/**
	* j܂B
	*/
	virtual ~GLImageDump();

	/**
	* j܂B
	*/
	virtual void Destroy();

	/**
	* `̑OB
	*/
	virtual HRESULT OnPreDraw();

	/**
	* `̌㏈B
	*/
	virtual HRESULT OnPostDraw();

	/**
	* Jn܂B
	*/
	virtual HRESULT Start();

	/**
	* ~܂B
	*/
	virtual HRESULT Stop();

protected:

	/**
	* obt@̓e܂B
	*/
	virtual HRESULT Process(LONGLONG time, int width, int height, unsigned char* data) = 0;

private:

	/**
	* ܂B
	*/
	HRESULT Init(int bufferSize);

	/**
	* obt@ێ܂B
	*/
	class Buffer
	{
	public:

		/**
		* \z܂B
		*/
		Buffer(int width, int height);

		/**
		* j܂B
		*/
		virtual ~Buffer();

		/**
		* ̃obt@ݒ肵܂B
		*/
		HRESULT SetNext(Buffer* pNext);

		/**
		* ̃obt@擾܂B
		*/
		HRESULT GetNext(Buffer** ppNext);

		/**
		* obt@擾܂B
		*/
		HRESULT GetBuffer(LONGLONG* pLastUpdated, int* pWidth, int* pHeight, unsigned char** ppBuffer);

		/**
		* eXV܂B
		*/
		HRESULT Update();

	private:

		/**
		* ێ܂B
		*/
		int m_iWidth;

		/**
		* ێ܂B
		*/
		int m_iHeight;

		/**
		* \z܂B
		*/
		unsigned char* m_pBuffer;

		/**
		* ێ܂B
		*/
		Buffer* m_pNext;

		/**
		* ŏIXVێ܂B
		*/
		LONGLONG m_llLastUpdated;

	};

	/**
	* p̃[J[łB
	*/
	class ProcessWorker : public IWorker
	{
	public:

		/**
		* \z܂B
		*/
		ProcessWorker(GLImageDump* pOwner);

		/**
		* j܂B
		*/
		virtual ~ProcessWorker();

		/**
		* j܂B
		*/
		virtual void Destroy();
		
		/**
		* s܂B
		* G[̓XbĥȂ΂Ȃ܂B
		*/
		virtual HRESULT Invoke();

		/**
		* ~܂B
		*/
		void Stop();

	private:

		/**
		* p邩ǂێ܂B
		*/
		bool IsContinue();

		/**
		* I[i[ێ܂B
		*/
		GLImageDump* m_pOwner;

		/**
		* p邩ǂێ܂B
		*/
		bool m_bContinue;

		/**
		* ptOpMutexێ܂B
		*/
		Mutex m_continueMutex;

	};

	/**
	* `pɃobt@擾܂B
	*/
	HRESULT PopForDraw(Buffer** ppBuffer);

	/**
	* pɃobt@擾܂B
	*/
	HRESULT PopForProcess(Buffer** ppBuffer);

	/**
	* `pɃobt@ǉ܂B
	*/
	HRESULT PushForDraw(Buffer* pBuffer);

	/**
	* pɃobt@ǉ܂B
	*/
	HRESULT PushForProcess(Buffer* pBuffer);

	/**
	* EBhEێ܂B
	*/
	IWindow* m_pWindow;

	/**
	* Cxgobt@̒ʒmCxgێ܂B
	*/
	HANDLE m_hEvBuffer;

	/**
	* obt@pMutexێ܂B
	*/
	Mutex m_bufferMutex;

	/**
	* gpĂȂobt@ێ܂B
	*/
	Buffer* m_pFreeBuf;

	/**
	* `ς݂̃obt@ێ܂B
	*/
	Buffer* m_pPostBuf;

	/**
	* [J[ێ܂B
	*/
	ProcessWorker* m_pWorker;

	/**
	* ~ʒmp̃Cxgێ܂B
	*/
	HANDLE m_hStopEvent;

	/**
	* Xbhێ܂B
	*/
	WorkerThread m_workerThread;

};
