#include "StdAfx.h"
#include "JpgImageLoader.h"
#include "JpegCore.h"
#include "../common/URLPath.h"

// \z܂B
JpgImageLoader::JpgImageLoader()
{
	;
}

// j܂B
JpgImageLoader::~JpgImageLoader()
{
}

// j܂B
void JpgImageLoader::Destroy()
{
	delete this;
}

// ftHgTCYݒ肵܂B
HRESULT JpgImageLoader::SetDefaultSize(int width, int height)
{
	// ɉȂ
	return S_OK;
}

// 摜ǂݍ݂܂B
HRESULT JpgImageLoader::Load(const char* path, IImageLoaderNotify* pNotify)
{
	if(path == NULL || pNotify == NULL) {
		return E_POINTER;
	}

	HRESULT hr = S_OK;
	URLPath urlPath(path, & hr);
	if(FAILED(hr)) {
		return hr;
	}

	string realPath;
	hr = urlPath.GetRealPath(realPath);
	if(FAILED(hr)) {
		return hr;
	}

	WorkerImpl* pWorkerImpl = new WorkerImpl(realPath.c_str(), pNotify);
	return m_workerThread.Start(pWorkerImpl);
}

// 摜ǂݍ݂܂B
HRESULT JpgImageLoader::Load(ISource* pSource, IImageLoaderNotify* pNotify)
{
	return E_NOTIMPL;
}

// \z܂B
JpgImageLoader::WorkerImpl::WorkerImpl(const char* path, IImageLoaderNotify* pNotify)
{
	m_path = path;
	m_pNotify = pNotify;
	m_pBuffer = NULL;
	m_iWidth = 0;
	m_iHeight = 0;
}

// j܂B
JpgImageLoader::WorkerImpl::~WorkerImpl()
{
	if(m_pNotify != NULL) {
		m_pNotify->Destroy();
		m_pNotify = NULL;
	}
	m_pBuffer = NULL;
}

// j܂B
void JpgImageLoader::WorkerImpl::Destroy()
{
	delete this;
}

// s܂B
HRESULT JpgImageLoader::WorkerImpl::Invoke()
{
	IImage* pImage = NULL;
	HRESULT hr = Read(& pImage);
	if(FAILED(hr)) {
		// TODO:
		m_pNotify->NotifyFailed("Failed to render image");
	}else{
		m_pNotify->NotifySuccess(pImage);
	}

	return S_OK;
}

// ǂݍ݂܂B
HRESULT JpgImageLoader::WorkerImpl::Read(IImage** ppImage)
{
	if(ppImage == NULL) {
		return E_POINTER;
	}

	FILE* pFile = fopen(m_path.c_str(), "rb");
	if(pFile == NULL) {
		return E_FAIL;
	}

	READ_INFO readInfo;
	ZeroMemory(& readInfo, sizeof(readInfo));
	readInfo.pPrivData = this;
	readInfo.pReadScanLine = WorkerImpl::SReadScanLine;
	
	HRESULT hr = jpegRead(pFile, & readInfo);
	if(FAILED(hr)) {
		if(m_pBuffer != NULL) {
			delete m_pBuffer;
			m_pBuffer = NULL;
		}
		return hr;
	}

	*ppImage = new ImageImpl(m_iWidth, m_iHeight, m_pBuffer);
	return S_OK;
}

// obt@Rs[܂B
HRESULT JpgImageLoader::WorkerImpl::CreateBuffer(const struct jpeg_decompress_struct* pInfo)
{
	if(pInfo == NULL) {
		return E_POINTER;
	}

	int size = pInfo->output_width * pInfo->output_height * 3;
	unsigned char* pNewBuf = new unsigned char[size];
	m_pBuffer = pNewBuf;

	m_iWidth = pInfo->output_width;
	m_iHeight = pInfo->output_height;

	return S_OK;
}

// obt@Rs[܂B
HRESULT JpgImageLoader::WorkerImpl::ReadScanLine(const struct jpeg_decompress_struct* pInfo, const JSAMPARRAY pSrc, int y)
{
	if(pInfo == NULL || pSrc == NULL) {
		return E_POINTER;
	}
	if(m_pBuffer == NULL) {
		HRESULT hr = CreateBuffer(pInfo);
		if(FAILED(hr)) {
			return hr;
		}
	}

	for(unsigned int x = 0; x < pInfo->output_width; x ++) {
		int srcPos = pInfo->output_components * x;
		int destPos = (x + (y * pInfo->output_width)) * 3;
		
		unsigned char* pRelDest = &(m_pBuffer[destPos]);
		const unsigned char* pRelSrc = (const unsigned char*) &(pSrc[0][srcPos]);

		if(pInfo->output_components == 3) {
			pRelDest[0] = pRelSrc[0];
			pRelDest[1] = pRelSrc[1];
			pRelDest[2] = pRelSrc[2];
		}else{
			return E_UNEXPECTED;
		}
	}

	return S_OK;
}

// obt@Rs[܂B
HRESULT JpgImageLoader::WorkerImpl::SReadScanLine(void* pPrivData, const struct jpeg_decompress_struct* pInfo, const JSAMPARRAY pSrc, int y)
{
	if(pPrivData == NULL) {
		return E_POINTER;
	}
	JpgImageLoader::WorkerImpl* pWorker = (JpgImageLoader::WorkerImpl*) pPrivData;
	return pWorker->ReadScanLine(pInfo, pSrc, y);
}

// \z܂B
JpgImageLoader::ImageImpl::ImageImpl(int width, int height, unsigned char* pBuf)
: AbstractImage(width, height)
{
	m_pBuf = pBuf;
}

// j܂B
JpgImageLoader::ImageImpl::~ImageImpl()
{
	if(m_pBuf != NULL) {
		delete m_pBuf;
		m_pBuf = NULL;
	}
}

// j܂B
void JpgImageLoader::ImageImpl::Destroy()
{
	delete this;
}

// sNZ擾܂B
HRESULT JpgImageLoader::ImageImpl::GetPixel(int x, int y, COMMON_ARGB* pARGB)
{
	if(m_pBuf == NULL) {
		return E_UNEXPECTED;
	}
	if(pARGB == NULL) {
		return E_POINTER;
	}

	unsigned char *pPixel = &(m_pBuf[(y * m_iWidth + x) * 3]);
	pARGB->a = (unsigned char)(0xff * m_dAlpha);
	pARGB->r = pPixel[0];
	pARGB->g = pPixel[1];
	pARGB->b = pPixel[2];
	return S_OK;
}
