#include "StdAfx.h"
#include "PngImageLoader.h"
#include "../common/SourceFactory.h"

// \z܂B
PngImageLoader::PngImageLoader()
{
	;
}

// j܂B
PngImageLoader::~PngImageLoader()
{
}

// j܂B
void PngImageLoader::Destroy()
{
	delete this;
}

// ftHgTCYݒ肵܂B
HRESULT PngImageLoader::SetDefaultSize(int width, int height)
{
	// ɉȂ
	return S_OK;
}

// 摜ǂݍ݂܂B
HRESULT PngImageLoader::Load(const char* path, IImageLoaderNotify* pNotify)
{
	if(path == NULL || pNotify == NULL) {
		return E_POINTER;
	}

	// TODO: 
	HRESULT hr = S_OK;
	SourceFactory sourceFactory("", & hr);
	if(FAILED(hr)) {
		return hr;
	}
	ISource* pSource = NULL;

	hr = sourceFactory.Load(path, & pSource);
	if(FAILED(hr)) {
		return hr;
	}

	return Load(pSource, pNotify);
}

// 摜ǂݍ݂܂B
HRESULT PngImageLoader::Load(ISource* pSource, IImageLoaderNotify* pNotify)
{
	if(pSource == NULL || pNotify == NULL) {
		return E_POINTER;
	}

	WorkerImpl* pWorkerImpl = new WorkerImpl(pSource, pNotify);
	return m_workerThread.Start(pWorkerImpl);
}

// \z܂B
PngImageLoader::WorkerImpl::WorkerImpl(ISource* pSource, IImageLoaderNotify* pNotify)
{
	m_pSource = pSource;
	m_pNotify = pNotify;
}

// j܂B
PngImageLoader::WorkerImpl::~WorkerImpl()
{
	if(m_pSource != NULL) {
		m_pSource->Destroy();
		m_pSource = NULL;
	}
	if(m_pNotify != NULL) {
		m_pNotify->Destroy();
		m_pNotify = NULL;
	}
}

// j܂B
void PngImageLoader::WorkerImpl::Destroy()
{
	delete this;
}

// s܂B
HRESULT PngImageLoader::WorkerImpl::Invoke()
{
	IImage* pImage = NULL;
	HRESULT hr = Read(& pImage);
	if(FAILED(hr)) {
		// TODO:
		m_pNotify->NotifyFailed("Failed to render image");
	}else{
		m_pNotify->NotifySuccess(pImage);
	}

	return S_OK;
}

// ǂݍ݂܂B
HRESULT PngImageLoader::WorkerImpl::Read(IImage** ppImage)
{
	if(ppImage == NULL) {
		return E_POINTER;
	}
 
	/*if(!png_check_sig(FileImage, fsize)) {
		return E_FAIL;
	}*/
	
	png_struct *pPng = NULL;
	png_info *pInfo = NULL;

	if((pPng = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) == NULL) {
		return E_FAIL;
	}
	if((pInfo = png_create_info_struct(pPng)) == NULL) {
		return E_FAIL;
	}
	
	png_set_read_fn(pPng, (png_voidp)m_pSource, (png_rw_ptr)WorkerImpl::PngReadFunc);
	png_read_info(pPng, pInfo);
	
	png_uint_32 PngWidth = 0;
	png_uint_32 PngHeight = 0;
	int bpp = 0;
	int ColorType = 0;

	png_get_IHDR(pPng, pInfo, &PngWidth, &PngHeight, &bpp, &ColorType, NULL, NULL, NULL);
	
	if(png_get_valid(pPng, pInfo, PNG_INFO_tRNS)) {
		png_set_expand(pPng);
	}
	if(ColorType == PNG_COLOR_TYPE_PALETTE) {
		png_set_expand(pPng);
	}else if(ColorType == PNG_COLOR_TYPE_GRAY && bpp < 8) {
		png_set_expand(pPng);
	}else if(bpp > 8) {
        png_set_strip_16(pPng);
	}
	if(ColorType == PNG_COLOR_TYPE_GRAY) {
        png_set_gray_to_rgb(pPng);
	}
	
	unsigned char *pBmpBuffer = new unsigned char[PngWidth * PngHeight * 4];
    unsigned char **pLines = (unsigned char **)malloc(sizeof(unsigned char *) * PngHeight);
	if(pLines == NULL) {
		return E_OUTOFMEMORY;
	}

	for(unsigned int i = 0; i < PngHeight; i++) {
		pLines[i] = pBmpBuffer + PngWidth * 4 * i;
	}

	if(!(ColorType & PNG_COLOR_MASK_ALPHA)) {
		png_set_filler(pPng, 0, 1);
	}
	
	png_set_bgr(pPng);
	png_read_image(pPng, pLines);

    free(pLines);
	pLines = NULL;

	png_destroy_info_struct(pPng, & pInfo);
	png_destroy_read_struct(&pPng, NULL, NULL);
	pPng = NULL;

	*ppImage = new ImageImpl(PngWidth, PngHeight, pBmpBuffer);
	return S_OK;
}
  
void PngImageLoader::WorkerImpl::PngReadFunc(png_struct *pPng, png_bytep buf, png_size_t size)
{
	ISource* pSource = (ISource*) png_get_io_ptr(pPng);
	for(unsigned int i = 0; i < size; i ++) {
		char dt = 0;
		HRESULT hr = pSource->Read(& dt);
		if(hr == S_FALSE) {
			return;
		}
		buf[i] = dt;
	}
}	

// \z܂B
PngImageLoader::ImageImpl::ImageImpl(int width, int height, unsigned char* pBuf)
: AbstractImage(width, height)
{
	m_pBuf = pBuf;
}

// j܂B
PngImageLoader::ImageImpl::~ImageImpl()
{
	if(m_pBuf != NULL) {
		delete m_pBuf;
		m_pBuf = NULL;
	}
}

// j܂B
void PngImageLoader::ImageImpl::Destroy()
{
	delete this;
}

// sNZ擾܂B
HRESULT PngImageLoader::ImageImpl::GetPixel(int x, int y, COMMON_ARGB* pARGB)
{
	if(m_pBuf == NULL) {
		return E_UNEXPECTED;
	}
	if(pARGB == NULL) {
		return E_POINTER;
	}

	unsigned char *pPixel = &(m_pBuf[(y * m_iWidth + x) * 4]);
	pARGB->a = (unsigned char)(/*pPixel[3]*/ 0xff * m_dAlpha);
	pARGB->r = pPixel[2];
	pARGB->g = pPixel[1];
	pARGB->b = pPixel[0];
	return S_OK;
}
