#include "StdAfx.h"
#include "JSDispatch.h"

// O[oIuWFNg̗p
JSClass g_dispatchClass = {
	"dispatch",JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,JS_PropertyStub,JS_PropertyStub,JS_PropertyStub,
	JS_EnumerateStub,JS_ResolveStub,JS_ConvertStub,JSDispatch::FinalizeOp
};

// VARIANTJSVALɕϊ܂B
HRESULT JSDispatch::ConvertVARIANTtoJSVAL(JSEngineContext* pContext, VARIANT* pVar, jsval* pJsval)
{
	if(pContext == NULL || pVar == NULL || pJsval == NULL) {
		return E_POINTER;
	}
	VARTYPE vt = pVar->vt;
	if(vt == VT_EMPTY) {
		*pJsval = JSVAL_VOID;
		return S_OK;
	}else if(vt == VT_NULL) {
		*pJsval = JSVAL_NULL;
		return S_OK;
	}else if(vt == VT_DISPATCH) {
		if(pVar->pdispVal == NULL) {
			*pJsval = JSVAL_NULL;
			return S_OK;
		}
		HRESULT hr = S_OK;
		JSDispatch* pDisp = new JSDispatch(pVar->pdispVal, & hr);
		if(FAILED(hr)) {
			pDisp->Destroy();
			pDisp = NULL;
			return hr;
		}

		hr = pDisp->SetAutoDelete(true);
		if(FAILED(hr)) {
			pDisp->Destroy();
			pDisp = NULL;
			return hr;
		}

		hr = pDisp->Attach(pContext, NULL);
		if(FAILED(hr)) {
			pDisp->Destroy();
			pDisp = NULL;
			return hr;
		}

		*pJsval = OBJECT_TO_JSVAL(pDisp->GetCore());
		return S_OK;
	}
	if(vt == VT_BOOL) {
		if(pVar->boolVal) {
			*pJsval = JSVAL_TRUE;
		}else{
			*pJsval = JSVAL_FALSE;
		}
		return S_OK;
	}
	if(vt == VT_INT || vt == VT_I1 || vt == VT_I2 || vt == VT_I4 || vt == VT_I8) {
		*pJsval = INT_TO_JSVAL(pVar->intVal);
		return S_OK;
	}
	if(vt == VT_UINT || vt == VT_UI1 || vt == VT_UI2 || vt == VT_UI4 || vt == VT_UI8) {
		*pJsval = INT_TO_JSVAL(pVar->uintVal);
		return S_OK;
	}
	if(vt == VT_R8) {
		jsdouble dv = pVar->dblVal;
		JSBool br = JS_NewNumberValue(pContext->GetCore(), dv, pJsval);
		if(br == JS_FALSE) {
			return E_FAIL;
		}
		return S_OK;
	}
	if(vt == VT_BSTR) {
		if(pVar->bstrVal == NULL) {
			*pJsval = JSVAL_NULL;
			return S_OK;
		}
		string value;
		HRESULT hr = DispatchWrapper::GetStringFromBSTR(value, pVar->bstrVal);
		if(FAILED(hr)) {
			return hr;
		}

		JSString* pStr = JS_NewStringCopyN(pContext->GetCore(), value.c_str(), value.size());
		if(pStr == NULL) {
			return E_FAIL;
		}
		*pJsval = STRING_TO_JSVAL(pStr);
		return S_OK;
	}
	return E_NOTIMPL;
}

// JSVALVARIANTɕϊ܂B
HRESULT JSDispatch::ConvertJSVALtoVARIANT(JSEngineContext* pContext, jsval val, VARIANT* pVar, VARTYPE vt)
{
	if(pContext == NULL || pVar == NULL) {
		return E_POINTER;
	}
	if(val == JSVAL_NULL) {
		pVar->vt = VT_NULL;
		return S_OK;
	}else if(val == JSVAL_VOID) {
		pVar->vt = VT_EMPTY;
		return S_OK;
	}else if(vt == VT_INT || vt == VT_I1 || vt == VT_I2 || vt == VT_I4 || vt == VT_I8) {
		pVar->vt = vt;
		pVar->intVal = JSVAL_TO_INT(val);
		return S_OK;
	}else if(vt == VT_UINT || vt == VT_UI1 || vt == VT_UI2 || vt == VT_UI4 || vt == VT_UI8) {
		pVar->vt = vt;
		pVar->uintVal = JSVAL_TO_INT(val);
		return S_OK;
	}else if(vt == VT_BOOL) {
		pVar->vt = vt;
		if(val == JSVAL_TRUE) {
			pVar->boolVal = true;
		}else{
			pVar->boolVal = false;
		}
		return S_OK;
	}else if(vt == VT_R8) {
		pVar->vt = vt;
		if(JSVAL_IS_INT(val)) {
			pVar->dblVal = JSVAL_TO_INT(val);
		}else if(JSVAL_IS_DOUBLE(val)) {
			jsdouble* pDouble = JSVAL_TO_DOUBLE(val);
			if(pDouble == NULL) {
				return E_FAIL;
			}else{
				pVar->dblVal = *pDouble;
			}
		}else{
			return E_FAIL;
		}
		return S_OK;
	}else if(vt == VT_BSTR) {
		string str;
		HRESULT hr = JSObjectWrapper::GetValueToString(pContext, val, str);
		if(FAILED(hr)) {
			return hr;
		}

		pVar->vt = VT_BSTR;
		pVar->bstrVal = NULL;
		return DispatchWrapper::GetBSTRFromString(&(pVar->bstrVal), str.c_str());
	}
	return E_NOTIMPL;
}

// \z܂B
JSDispatch::JSDispatch(IDispatch* pDispatch, HRESULT* phr)
: JSMemberCollection("JSDispatch")
{
	m_pDispatch = NULL;
	m_bAutoDelete = false;
	*phr = S_OK;

	HRESULT hr = S_OK;
	m_pDispatch = new DispatchWrapper(pDispatch, phr);
	if(FAILED(& hr)) {
		delete m_pDispatch;
		m_pDispatch = NULL;
		*phr = hr;
		return;
	}
}

// j܂B
JSDispatch::~JSDispatch()
{
	JSObject* pCore = GetCore();
	if(pCore != NULL) {
		JS_SetPrivate(m_pContext->GetCore(), pCore, NULL);
		pCore = NULL;
	}
	if(m_pDispatch != NULL) {
		delete m_pDispatch;
		m_pDispatch = NULL;
	}
}

// j܂B
void JSDispatch::Destroy()
{
	delete this;
}

// JavaScriptGWƃIuWFNgݒ肵܂B
HRESULT JSDispatch::Attach(JSEngineContext* pContext, JSObject* pObject)
{
	if(pContext == NULL) {
		return E_POINTER;
	}

	if(pObject == NULL) {
		JSObject* pNewObj = JS_NewObject(pContext->GetCore(), & g_dispatchClass, NULL, NULL);
		if(pNewObj == NULL) {
			return E_FAIL;
		}
		pObject = pNewObj;
	}

	return JSMemberCollection::Attach(pContext, pObject, ! m_bAutoDelete);
}

// j̏ݒ肵܂B
void JSDispatch::FinalizeOp(JSContext* cx, JSObject* obj)
{
	JSMemberCollection* pColl = JSOW_GET_PRIVATE(cx, obj, JSMemberCollection);
	if(pColl == NULL) {
		return;
	}
	JSDispatch* pDisp = (JSDispatch*) pColl;
	pDisp->Detach();
	if(pDisp->m_bAutoDelete) {
		pDisp->Destroy();
	}
}

// Iɍ폜邩ǂݒ肵܂B
HRESULT JSDispatch::SetAutoDelete(bool autoDelete)
{
	m_bAutoDelete = autoDelete;
	return S_OK;
}

// IuWFNgɊ֘AtĂf[^擾܂B
HRESULT JSDispatch::GetData(const char* name, void** ppv)
{
	if(name == NULL || ppv == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == JSDISPATCH_IDISPATCH) {
		IDispatch* pDisp = NULL;
		HRESULT hr = m_pDispatch->GetDispatch(& pDisp);
		if(FAILED(hr)) {
			return hr;
		}
		*ppv = pDisp;
		return S_OK;
	}
	return JSMemberCollection::GetData(name, ppv);
}

// ܂B
HRESULT JSDispatch::InitMembers()
{
	if(m_pDispatch == NULL) {
		return E_UNEXPECTED;
	}

	set<string> members;
	UINT memCount = 0;
	HRESULT hr = m_pDispatch->GetMemberCount(& memCount);
	if(FAILED(hr)) {
		return hr;
	}

	for(UINT i = 0; i < memCount; i ++) {
		string name;
		hr = m_pDispatch->GetMemberName(i, name);
		if(FAILED(hr)) {
			return hr;
		}

		if(members.find(name) != members.end()) {
			// o^ς
			continue;
		}

		members.insert(name);
		WORD wFlags = 0;
		hr = m_pDispatch->GetMemberWFlags(i, & wFlags);
		if(FAILED(hr)) {
			return hr;
		}

		UINT paramCount = 0;
		hr = m_pDispatch->GetMemberParamCount(i, & paramCount);
		if(FAILED(hr)) {
			return hr;
		}

		if(wFlags == DISPATCH_METHOD) {
			hr = DefineFunction(name.c_str(), paramCount);
			if(FAILED(hr)) {
				return hr;
			}
		}else{
			UINT getIndex = 0;
			UINT setIndex = 0;

			HRESULT getHr = m_pDispatch->FindMember(name.c_str(), DISPATCH_PROPERTYGET, & getIndex);
			HRESULT setHr = m_pDispatch->FindMember(name.c_str(), DISPATCH_PROPERTYPUT, & setIndex);

			if(FAILED(getHr)) {
				return E_FAIL;
			}

			bool readonly = true;
			if(setHr == S_OK) {
				readonly = false;
			}

			hr = DefineProperty(name.c_str(), readonly);
			if(FAILED(hr)) {
				return hr;
			}
		}
	}

	return S_OK;
}

// vpeBݒ肵܂B
HRESULT JSDispatch::SetProperty(JSContext *cx, JSObject *obj, const char* name, jsval vp)
{
	if(m_pDispatch == NULL) {
		return E_UNEXPECTED;
	}

	UINT memIndex = 0;
	HRESULT hr = m_pDispatch->FindMember(name, DISPATCH_PROPERTYPUT, & memIndex);
	if(FAILED(hr)) {
		return hr;
	}

	VARTYPE vt = VT_NULL;
	hr = m_pDispatch->GetMemberReturnType(memIndex, & vt);
	if(FAILED(hr)) {
		return hr;
	}

	DISPPARAMS dispParams;
	ZeroMemory(& dispParams, sizeof(dispParams));
	VARIANT var;
	VARIANT rvar;
	VariantInit(& var);
	VariantInit(& rvar);

	hr = ConvertJSVALtoVARIANT(m_pContext, vp, & var, vt);
	if(FAILED(hr)) {
		VariantClear(& var);
		VariantClear(& rvar);
		return hr;
	}

	dispParams.cArgs = 1;
	dispParams.rgvarg = &var;

	hr = m_pDispatch->Invoke(memIndex, & dispParams, & rvar);

	VariantClear(& var);
	VariantClear(& rvar);
	return hr;
}

// vpeB擾܂B
HRESULT JSDispatch::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	if(m_pDispatch == NULL) {
		return E_UNEXPECTED;
	}

	UINT memIndex = 0;
	HRESULT hr = m_pDispatch->FindMember(name, DISPATCH_PROPERTYGET, & memIndex);
	if(FAILED(hr)) {
		return hr;
	}

	VARTYPE vt = VT_NULL;
	hr = m_pDispatch->GetMemberReturnType(memIndex, & vt);
	if(FAILED(hr)) {
		return hr;
	}

	DISPPARAMS dispParams;
	ZeroMemory(& dispParams, sizeof(dispParams));
	VARIANT var;
	VariantInit(& var);
	hr = m_pDispatch->Invoke(memIndex, & dispParams, & var);
	if(FAILED(hr)) {
		VariantClear(& var);
		return hr;
	}

	hr = ConvertVARIANTtoJSVAL(m_pContext, & var, vp);
	VariantClear(& var);
	return hr;
}

// ֐s܂B
HRESULT JSDispatch::InvokeFunction(JSContext *cx, JSObject *obj, const char* name, uintN argc, jsval *argv, jsval *rval)
{
	if(m_pDispatch == NULL) {
		return E_UNEXPECTED;
	}

	UINT memIndex = 0;
	HRESULT hr = m_pDispatch->FindMember(name, DISPATCH_METHOD, & memIndex);
	if(FAILED(hr)) {
		return hr;
	}

	UINT params = 0;
	hr = m_pDispatch->GetMemberParamCount(memIndex, & params);
	if(FAILED(hr)) {
		return hr;
	}

	DISPPARAMS dispParams;
	ZeroMemory(& dispParams, sizeof(dispParams));
	VARIANT* pVarArgs = NULL;
	if(params > 0) {
		pVarArgs = new VARIANT[params];
		ZeroMemory(pVarArgs, sizeof(VARIANT) * params);
		for(UINT i = 0; i < params; i ++) {
			VARIANT* pVarArg = &(pVarArgs[i]);
			VariantInit(pVarArg);
			if(argc > i) {
				VARTYPE argVt = VT_NULL;
				hr = m_pDispatch->GetMemberParamType(memIndex, i, & argVt);
				if(FAILED(hr)) {
					for(UINT j = 0; j < i; j ++) {
						VariantClear(&(pVarArgs[j]));
					}
					delete pVarArgs;
					pVarArgs = NULL;
					return hr;
				}

				hr = ConvertJSVALtoVARIANT(m_pContext, argv[i], pVarArg, argVt);
				if(FAILED(hr)) {
					for(UINT j = 0; j < i; j ++) {
						VariantClear(&(pVarArgs[j]));
					}
					delete pVarArgs;
					pVarArgs = NULL;
					return hr;
				}
			}
		}
	}
	dispParams.cArgs = params;
	dispParams.rgvarg = pVarArgs;
	VARIANT rvar;
	VariantInit(& rvar);

	hr = m_pDispatch->Invoke(memIndex, & dispParams, & rvar);
	if(pVarArgs != NULL) {
		for(UINT j = 0; j < params; j ++) {
			VariantClear(&(pVarArgs[j]));
		}
		delete pVarArgs;
		pVarArgs = NULL;
	}

	if(FAILED(hr)) {
		return hr;
	}

	VARTYPE retVt = VT_NULL;
	hr = m_pDispatch->GetMemberReturnType(memIndex, & retVt);
	if(FAILED(hr)) {
		return hr;
	}

	return ConvertVARIANTtoJSVAL(m_pContext, & rvar, rval);
}
