#pragma once

#include <jsapi.h>
#include "IJSContext.h"
#include "../common/Mutex.h"

class JSEngineContext;
class JSGlobalObject;
class IJSObject;
class URLPath;

#define GET_JS_ENGINE_CONTEXT(pContext) \
	((JSEngineContext*) JS_GetContextPrivate(pContext))

/**
* JavaScript̃ReLXgłB
*/
class JSEngineContext : public IJSContext
{
public:

	/**
	* \z܂B
	*/
	JSEngineContext(JSRuntime* pRuntime, size_t stackChunkSize, HRESULT* phr);

	/**
	* j܂B
	*/
	virtual ~JSEngineContext(void);

	/**
	* ̃IuWFNgj܂B
	*/
	virtual void Destroy();

	/**
	* ʒmIuWFNgݒ肵܂B
	*/
	virtual HRESULT SetNotify(IJSContextNotify* pNotify);

	/**
	* R[hs܂B
	*/
	virtual HRESULT Evaluate(IJSCode* pCode);

	/**
	* VIuWFNg𐶐܂B
	*/
	HRESULT CreateNewObject(JSObject** ppObject);

	/**
	* SpiderMonkeỹRAIuWFNg擾܂B
	*/
	JSContext* GetCore();

	/**
	* O[ovpeB`܂B
	*/
	HRESULT DefineGlobalProperty(const char* name, IJSObject* pObject);

	/**
	* ֐𓯊s܂B
	*/
	HRESULT InvokeSync(jsval function, IJSObject* pThis, jsval* argv, int argc, jsval* rval);

	/**
	* O[oIuWFNg擾܂B
	*/
	HRESULT GetGlobalObject(IJSObject** ppObject);

	/**
	* \񂵂܂B
	*/
	HRESULT InvokeAsync(IJSFunction* pFunction);

	/**
	* \񂳂Ă鏈s܂B
	*/
	HRESULT ProcessAsync();

	/**
	* URL擾܂B
	*/
	HRESULT GetAbsoluteURL(const char* path, string& url);

	/**
	* {ƂȂURLݒ肵܂B
	*/
	HRESULT SetBaseURL(const char* baseURL);

	/**
	* {ƂȂURL擾܂B
	*/
	HRESULT GetBaseURL(string& url);

private:

	/**
	* ܂B
	*/
	HRESULT Init(JSRuntime* pRuntime, size_t stackChunkSize);

	/**
	* G[ʒm܂B
	*/
	void NotifyError(const char* message, JSErrorReport* pReport);

	/**
	* L[o܂B
	*/
	HRESULT PopQueue(queue<IJSFunction*>& queue);

	/**
	* L[Ɋ֐ǉ܂B
	*/
	HRESULT PushQueue(IJSFunction* pFunction);

	/**
	* L[̃IuWFNgׂč폜܂B
	*/
	HRESULT DestroyQueue(queue<IJSFunction*>& queue);

	/**
	* G[ʒm܂B
	*/
	static void ErrorReporter(JSContext* pContext, const char* message, JSErrorReport* pReport);

	/**
	* ReLXg̎̂ێ܂B
	*/
	JSContext* m_pContext;

	/**
	* O[oIuWFNgێ܂B
	*/
	JSGlobalObject* m_pGlobal;

	/**
	* vpeBꗗێ܂B
	*/
	vector<IJSObject*> m_props;

	/**
	* 񓯊s̃L[łB
	*/
	queue<IJSFunction*> m_asyncQueue;

	/**
	* rpMutexB
	*/
	Mutex m_mutex;

	/**
	* {ƂȂURLB
	*/
	URLPath* m_pBaseURL;
};
