#include "StdAfx.h"
#include "JSGlobalObject.h"

#define MAX_TIMER_ID	100000L

// \z܂B
JSGlobalObject::JSGlobalObject()
{
	m_pNotify = NULL;
	m_iErrorReceived = 0;
}

// j܂B
JSGlobalObject::~JSGlobalObject()
{
	for(map<string,IJSObject*>::iterator it = m_props.begin(); it != m_props.end(); it ++) {
		IJSObject* pProp = it->second;
		pProp->Destroy();
	}
	m_props.clear();
	m_pNotify = NULL;
}

// j܂B
void JSGlobalObject::Destroy()
{
	delete this;
}

// ʒmIuWFNgݒ肵܂B
HRESULT JSGlobalObject::SetNotify(IJSContextNotify* pNotify)
{
	m_pNotify = pNotify;
	return S_OK;
}

// ܂B
HRESULT JSGlobalObject::Init(JSContext* pContext, JSObject* pObject)
{
	JSFunction* pFunction = JS_DefineFunction(pContext, pObject, "alert", JSOW_FUNCTION_REF(JSGlobalObject, SAlert), 1, 0);
	if(pFunction == NULL) {
		return E_FAIL;
	}
	pFunction = JS_DefineFunction(pContext, pObject, "printLine", JSOW_FUNCTION_REF(JSGlobalObject, SPrintLine), 1, 0);
	if(pFunction == NULL) {
		return E_FAIL;
	}
	pFunction = JS_DefineFunction(pContext, pObject, "setTimeout", JSOW_FUNCTION_REF(JSGlobalObject, SSetTimeout), 2, 0);
	if(pFunction == NULL) {
		return E_FAIL;
	}
	pFunction = JS_DefineFunction(pContext, pObject, "clearTimeout", JSOW_FUNCTION_REF(JSGlobalObject, SClearTimeout), 1, 0);
	if(pFunction == NULL) {
		return E_FAIL;
	}
	for(map<string,IJSObject*>::iterator it = m_props.begin();
		it != m_props.end();
		it ++) {
		string name = it->first;
		JSBool bret = JS_DefineProperty(pContext, pObject, name.c_str(), JSVAL_NULL, JSOW_PROPERTY_REF(JSGlobalObject,SGetter), JSOW_PROPERTY_REF(JSGlobalObject,SSetter), JSPROP_READONLY);
		if(bret == JS_FALSE) {
			return E_FAIL;
		}
	}
	return S_OK;
}

// vpeB`܂B
HRESULT JSGlobalObject::DefineProperty(const char* name, IJSObject* pObject)
{
	if(name == NULL || pObject == NULL) {
		return E_POINTER;
	}
	if(m_props[name] != NULL) {
		return E_FAIL;
	}

	m_props[name] = pObject;
	JSObject* pCore = GetCore();
	if(pCore == NULL) {
		return S_OK;
	}
	JSBool bret = JS_DefineProperty(m_pContext->GetCore(), pCore, name, JSVAL_NULL, JSOW_PROPERTY_REF(JSGlobalObject,SGetter), JSOW_PROPERTY_REF(JSGlobalObject,SSetter), JSPROP_READONLY);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}
	return S_OK;
}

// G[ʒm܂B
HRESULT JSGlobalObject::NotifyError(const char* message, const char* filename, unsigned int lineno, const char* token)
{
	m_iErrorReceived ++;
	if(m_pNotify != NULL) {
		return m_pNotify->OnError(message, filename, lineno, token);
	}

	printf("ERROR: %s\n", message);
	return S_OK;
}

// G[󂯎ǂ𔻒肵܂B
HRESULT JSGlobalObject::IsErrorReceived(bool* pReceived)
{
	if(pReceived == NULL) {
		return E_POINTER;
	}
	if(m_iErrorReceived > 0) {
		*pReceived = true;
	}else{
		*pReceived = false;
	}
	return S_OK;
}

// G[󂯎ǂNA܂B
HRESULT JSGlobalObject::ClearErrorReceived()
{
	m_iErrorReceived = 0;
	return S_OK;
}

// ^C}[\񂵂܂B
HRESULT JSGlobalObject::ReserveTimer(long* pId)
{
	if(pId == NULL) {
		return E_POINTER;
	}

	long id = 1L;
	bool reserved = false;
	while(id < MAX_TIMER_ID) {
		if(m_timers.count(id) == 0) {
			reserved = true;
			break;
		}
		id ++;
	}
	if(reserved == false) {
		return E_UNEXPECTED;
	}

	m_timers[id] = true;
	*pId = id;
	return S_OK;
}

// ^C}[s܂B
HRESULT JSGlobalObject::InvokeTimer(long id, bool* pEnabled)
{
	if(pEnabled == NULL) {
		return E_POINTER;
	}
	map<long, bool>::iterator it = m_timers.find(id);
	if(it == m_timers.end()) {
		return S_FALSE;
	}

	bool br = m_timers[id];
	m_timers.erase(it);

	*pEnabled = br;
	return S_OK;
}

// Alert֐̒`
JSOW_FUNCTION_DEF(JSGlobalObject, SAlert, AlertImpl);

// Alert֐̒`
HRESULT JSGlobalObject::AlertImpl(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	string message;
	HRESULT hr = GetValueToString(argv[0], message);
	if(FAILED(hr)) {
		return hr;
	}

	if(m_pNotify != NULL) {
		return m_pNotify->OnAlert(message.c_str());
	}

	// TODO:
	printf("ALERT: %s\n", message.c_str());

	return S_OK;
}

// PrintLine֐̒`
JSOW_FUNCTION_DEF(JSGlobalObject, SPrintLine, PrintLineImpl);

// Alert֐̒`
HRESULT JSGlobalObject::PrintLineImpl(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	string message;
	HRESULT hr = GetValueToString(argv[0], message);
	if(FAILED(hr)) {
		return hr;
	}

	if(m_pNotify != NULL) {
		return m_pNotify->OnPrintLine(message.c_str());
	}

	// TODO:
	printf("PRINTLINE: %s\n", message.c_str());

	return S_OK;
}

// SetTimeout֐̒`
JSOW_FUNCTION_DEF(JSGlobalObject, SSetTimeout, SetTimeoutImpl);

// SetTimeout֐̒`
HRESULT JSGlobalObject::SetTimeoutImpl(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if(cx == NULL || obj == NULL || argv == NULL) {
		return E_POINTER;
	}
	if(argc < 2) {
		return E_INVALIDARG;
	}

	jsval func = argv[0];
	jsval msec = argv[1];
	long lmsec = 0L;

	JSBool bret = JS_ValueToInt32(cx, msec, & lmsec);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	FunctionRef* pFunc = new FunctionRef();
	HRESULT hr = pFunc->Init(m_pContext, this);
	if(FAILED(hr)) {
		delete pFunc;
		pFunc = NULL;
		return hr;
	}

	hr = pFunc->Set(func);
	if(FAILED(hr)) {
		delete pFunc;
		pFunc = NULL;
		return hr;
	}
 
	long id = 0L;
	hr = ReserveTimer(& id);
	if(FAILED(hr)) {
		delete pFunc;
		pFunc = NULL;
		return hr;
	}

	TimerWait* pWait = new TimerWait(this, id, lmsec, pFunc);
	hr = m_workerThread.Start(pWait);
	if(FAILED(hr)) {
		delete pFunc;
		pFunc = NULL;
		delete pWait;
		pWait = NULL;
		return hr;
	}
	
	if(rval != NULL) {
		// TODO: ̊֐OKH
		*rval = INT_TO_JSVAL(id);
	}
	return S_OK;
}

// ClearTimeout֐̒`
JSOW_FUNCTION_DEF(JSGlobalObject, SClearTimeout, ClearTimeoutImpl);

// Alert֐̒`
HRESULT JSGlobalObject::ClearTimeoutImpl(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if(cx == NULL || obj == NULL || argv == NULL) {
		return E_POINTER;
	}
	if(argc < 1) {
		return E_INVALIDARG;
	}

	jsval id = argv[0];
	long lid = 0L;

	JSBool bret = JS_ValueToInt32(cx, id, & lid);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	map<long, bool>::iterator it = m_timers.find(lid);
	if(it == m_timers.end()) {
		return S_OK;
	}

	m_timers[lid] = false;
	return S_OK;
}

// vpeB̒`
JSOW_PROPERTY_DEF(JSGlobalObject, SSetter, SetProperty, SGetter, GetProperty);

// vpeBݒ肵܂B
HRESULT JSGlobalObject::SetProperty(JSContext* cx, JSObject* obj, jsval idval, jsval* vp)
{
	return E_NOTIMPL;
}

// vpeBݒ肵܂B
HRESULT JSGlobalObject::GetProperty(JSContext* cx, JSObject* obj, jsval idval, jsval* vp)
{
	if(vp == NULL) {
		return S_FALSE;
	}
	string name;
	HRESULT hr = GetValueToString(idval, name);
	if(FAILED(hr)) {
		return hr;
	}

	IJSObject* pObject = m_props[name];
	if(pObject == NULL) {
		return E_UNEXPECTED;
	}
	
	*vp = OBJECT_TO_JSVAL(pObject->GetCore());
	return S_OK;
}

// \z܂B
JSGlobalObject::TimerWait::TimerWait(JSGlobalObject* pOwner, long id, long timeout, FunctionRef* pFunc)
{
	m_pOwner = pOwner;
	m_lId = id;
	m_pFunc = pFunc;
	m_lTimeout = timeout;
}

// j܂B
JSGlobalObject::TimerWait::~TimerWait()
{
	m_pOwner = NULL;
	m_pFunc = NULL;
}

// j܂B
void JSGlobalObject::TimerWait::Destroy()
{
	delete this;
}

// s܂B
HRESULT JSGlobalObject::TimerWait::Invoke()
{
	Sleep(m_lTimeout);

	TimerNotify* pNotify = new TimerNotify(m_pOwner, m_lId, m_pFunc);
	return m_pOwner->m_pContext->InvokeAsync(pNotify);
}

// \z܂B
JSGlobalObject::TimerNotify::TimerNotify(JSGlobalObject* pOwner, long id, FunctionRef* pFunc)
{
	m_pOwner = pOwner;
	m_lId = id;
	m_pFunc = pFunc;
}

// j܂B
JSGlobalObject::TimerNotify::~TimerNotify()
{
	if(m_pFunc != NULL) {
		delete m_pFunc;
		m_pFunc = NULL;
	}
	m_pOwner = NULL;
}

// j܂B
void JSGlobalObject::TimerNotify::Destroy()
{
	delete this;
}

// s܂B
HRESULT JSGlobalObject::TimerNotify::Invoke()
{
	if(m_pFunc == NULL) {
		return E_UNEXPECTED;
	}
	bool enabled = false;
	HRESULT hr = m_pOwner->InvokeTimer(m_lId, & enabled);
	if(FAILED(hr)) {
		return hr;
	}

	if(enabled == false) {
		return S_OK;
	}
	jsval rval = JSVAL_NULL;
	jsval argv = JSVAL_NULL;
	return m_pFunc->InvokeSync(& argv, 0, & rval);
}
