#include "StdAfx.h"
#include "JSObjectWrapper.h"

// jsval當o܂B
HRESULT JSObjectWrapper::GetValueToString(JSEngineContext* pContext, jsval val, string& str)
{
	if(pContext == NULL) {
		return E_POINTER;
	}
	JSContext* cx = pContext->GetCore();
	JSString* pStr = JS_ValueToString(cx, val);
	if(pStr == NULL) {
		return E_FAIL;
	}

	char* pMBS = JS_EncodeString(cx, pStr);
	if(pMBS == NULL) {
		return E_FAIL;
	}
	str = pMBS;
	JS_free(cx, pMBS);
	pMBS = NULL;

	return S_OK;
}

// G[񍐂܂B
void JSObjectWrapper::ReportError(JSContext* cx, const char* location, HRESULT hr)
{
	string loc = "Unknown";
	if(location != NULL) {
		loc = location;
	}

	string msg = "";
	LPSTR s = NULL;
	if(::FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM, NULL, hr, 0, (LPSTR)&s, 0, NULL) == 0) {
		msg = "Unknown Error";
	}else{
		msg = s;
		::LocalFree(s);
	}
	JS_ReportError(cx, "%s: %s(%08x)", loc.c_str(), msg.c_str(), hr);
}

// \z܂B
JSObjectWrapper::JSObjectWrapper()
{
	m_pContext = NULL;
	m_pObject = NULL;
}

// j܂B
JSObjectWrapper::~JSObjectWrapper()
{
	Detach();
}

// JavaScriptGWƃIuWFNgݒ肵܂B
HRESULT JSObjectWrapper::Attach(JSEngineContext* pContext, JSObject* pObject)
{
	return Attach(pContext, pObject, true);
}

// JavaScriptGWƃIuWFNgݒ肵܂B
HRESULT JSObjectWrapper::Attach(JSEngineContext* pContext, JSObject* pObject, bool addRoot)
{
	if(pContext == NULL) {
		return E_POINTER;
	}

	HRESULT hr = S_OK;
	m_pContext = pContext;
	if(pObject == NULL) {
		JSObject* pNewObj = NULL;
		hr = m_pContext->CreateNewObject(& pNewObj);
		if(FAILED(hr)) {
			return hr;
		}
		pObject = pNewObj;
	}

	JSContext* cx = m_pContext->GetCore();
	JSBool bret = JS_SetPrivate(cx, pObject, GetPrivate());
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	m_pObject = pObject;

	if(addRoot) {
		bret = JS_AddRoot(cx, & m_pObject);
		if(bret == JS_FALSE) {
			return E_FAIL;
		}
	}
	return Init(cx, m_pObject);
}

// IuWFNg̎̂擾܂B
JSObject* JSObjectWrapper::GetCore()
{
	return m_pObject;
}

// ꏊ擾܂B
void JSObjectWrapper::GetLocation(JSContext* cx, const char* className, const char* implName, jsval idval, string& location)
{
	if(className == NULL) {
		location += "(unknown object)";
	}else{
		location += className;
	}
	location += "::";
	if(implName == NULL) {
		location += "(unknown object)";
	}else{
		location += implName;
	}
	if(idval != JSVAL_NULL) {
		string sid;
		HRESULT hr = GetValueToString(idval, sid);
		if(hr == S_OK) {
			location += "[id=";
			location += sid;
			location += "]";
		}
	}
}

// IuWFNgɊ֘AtĂf[^擾܂B
HRESULT JSObjectWrapper::GetData(const char* name, void** ppv)
{
	if(name == NULL || ppv == NULL) {
		return E_POINTER;
	}
	return S_FALSE;
}

// jsval當o܂B
HRESULT JSObjectWrapper::GetValueToString(jsval val, string& str)
{
	return JSObjectWrapper::GetValueToString(m_pContext, val, str);
}

// l̔z𐶐܂B
HRESULT JSObjectWrapper::CreateDoubleArray(double* pValues, int size, JSObject** ppArray)
{
	if(pValues == NULL || ppArray == NULL) {
		return E_POINTER;
	}
	jsval* pJSVals = new jsval[size];
	if(pJSVals == NULL) {
		return E_OUTOFMEMORY;
	}

	for(int i = 0; i < size; i ++) {
		JSBool bret = JS_NewNumberValue(m_pContext->GetCore(), pValues[i], &(pJSVals[i]));
		if(bret == JS_FALSE) {
			return E_FAIL;
		}
	}

	JSObject* pNewArray = JS_NewArrayObject(m_pContext->GetCore(), size, pJSVals);

	delete pJSVals;
	pJSVals = NULL;

	if(pNewArray == NULL) {
		return E_FAIL;
	}

	*ppArray = pNewArray;
	return S_OK;
}

// JSObjectƂ̃IuWFNg̊֌W܂B
void JSObjectWrapper::Detach()
{
	if(m_pObject != NULL && m_pContext != NULL) {
		JS_RemoveRoot(m_pContext->GetCore(), & m_pObject);
	}
	m_pObject = NULL;
	m_pContext = NULL;
}

// \z܂B
JSObjectWrapper::FunctionRef::FunctionRef()
{
	m_pContext = NULL;
	m_pFunction = NULL;
	m_pOwner = NULL;
}

// j܂B
JSObjectWrapper::FunctionRef::~FunctionRef()
{
	if(m_pFunction != NULL && m_pContext != NULL) {
		JS_RemoveRoot(m_pContext->GetCore(), m_pFunction);
		delete m_pFunction;
		m_pFunction = NULL;
	}
	m_pOwner = NULL;
	m_pContext = NULL;
}

// ܂B
HRESULT JSObjectWrapper::FunctionRef::Init(JSEngineContext* pContext, IJSObject* pOwner)
{
	if(pContext == NULL || pOwner == NULL) {
		return E_POINTER;
	}
	m_pContext = pContext;
	m_pOwner = pOwner;
	return S_OK;
}

// s܂B
HRESULT JSObjectWrapper::FunctionRef::InvokeSync(jsval* argv, int argc, jsval* rval)
{
	if(m_pFunction == NULL) {
		return S_OK;
	}

	if(m_pContext == NULL) {
		return E_UNEXPECTED;
	}
	return m_pContext->InvokeSync((*m_pFunction), m_pOwner, argv, argc, rval);
}

// ֐ݒ肵܂B
HRESULT JSObjectWrapper::FunctionRef::Set(jsval func)
{
	if(m_pContext == NULL) {
		return E_UNEXPECTED;
	}
	if(!(JSVAL_IS_OBJECT(func))) {
		return E_INVALIDARG;
	}

	JSContext* cx = m_pContext->GetCore();
	if(m_pFunction == NULL) {
		m_pFunction = new jsval;
	}else{
		JSBool bret = JS_RemoveRoot(cx, m_pFunction);
		if(bret == JS_FALSE) {
			return E_FAIL;
		}
	}
	(*m_pFunction) = func;
	JSBool bret = JS_AddRoot(cx, m_pFunction);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}
	return S_OK;
}

// ֐擾܂B
HRESULT JSObjectWrapper::FunctionRef::Get(jsval* pFunc)
{
	if(pFunc == NULL) {
		return E_POINTER;
	}
	if(m_pFunction == NULL) {
		(*pFunc) = JSVAL_NULL;
		return S_FALSE;
	}

	(*pFunc) = (*m_pFunction);
	return S_OK;
}
