#include "StdAfx.h"
#include "JSValueWrapper.h"

// \z܂B
JSValueWrapper::JSValueWrapper(JSContext* pContext, jsval value)
{
	m_pContext = pContext;
	m_value = value;
}

// j܂B
JSValueWrapper::~JSValueWrapper()
{
	m_pContext = NULL;
	m_value = JSVAL_NULL;
}

// lݒ肵܂B
HRESULT JSValueWrapper::Set(jsval value)
{
	m_value = value;
	return S_OK;
}

// l擾܂B
HRESULT JSValueWrapper::Get(jsval* pValue)
{
	if(pValue == NULL) {
		return E_POINTER;
	}
	*pValue = m_value;
	return S_OK;
}

// booll擾܂B
HRESULT JSValueWrapper::GetBool(bool* pValue)
{
	if(pValue == NULL) {
		return E_POINTER;
	}
	JSBool jsb = JS_FALSE;
	JSBool bret = JS_ValueToBoolean(m_pContext, m_value, & jsb);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	if(jsb == JS_FALSE) {
		*pValue = false;
	}else{
		*pValue = true;
	}
	return S_OK;
}

// intl擾܂B
HRESULT JSValueWrapper::GetInt(int* pValue)
{
	if(pValue == NULL) {
		return E_POINTER;
	}
	int32 value = 0;
	JSBool bret = JS_ValueToInt32(m_pContext, m_value, & value);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	*pValue = value;
	return S_OK;
}

// GLfloatl擾܂B
HRESULT JSValueWrapper::GetGLfloat(GLfloat* pFloat)
{
	if(pFloat == NULL) {
		return E_POINTER;
	}
	if(JSVAL_IS_NULL(m_value) || JSVAL_IS_VOID(m_value)) {
		return E_FAIL;
	}

	jsdouble val = 0.0;
	JSBool bret = JS_ValueToNumber(m_pContext, m_value, &val);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	*pFloat = (GLfloat)val;
	return S_OK;
}

// doublel擾܂B
HRESULT JSValueWrapper::GetDouble(double* pDouble)
{
	if(pDouble == NULL) {
		return E_POINTER;
	}
	if(JSVAL_IS_NULL(m_value) || JSVAL_IS_VOID(m_value)) {
		return E_FAIL;
	}

	jsdouble val = 0.0;
	JSBool bret = JS_ValueToNumber(m_pContext, m_value, &val);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	*pDouble = (double)val;
	return S_OK;
}

// GLfloatl擾܂B
HRESULT JSValueWrapper::GetGLfloat(const char* name, GLfloat* pFloat)
{
	if(name == NULL || pFloat == NULL) {
		return E_POINTER;
	}

	JSValueWrapper elem(m_pContext, JSVAL_ZERO);
	HRESULT hr = GetJSVAL(name, & elem);
	if(FAILED(hr)) {
		return hr;
	}

	return elem.GetGLfloat(pFloat);
}

// IuWFNgWƉ߂Ēl擾܂B
HRESULT JSValueWrapper::GetXYZ(GLfloat* pX, GLfloat* pY, GLfloat* pZ)
{
	HRESULT hr = S_OK;

	hr = GetGLfloat("x", pX);
	if(FAILED(hr)) {
		return hr;
	}

	hr = GetGLfloat("y", pY);
	if(FAILED(hr)) {
		return hr;
	}

	hr = GetGLfloat("z", pZ);
	if(FAILED(hr)) {
		return hr;
	}

	return S_OK;
}

// IuWFNgFƉ߂Ēl擾܂B
HRESULT JSValueWrapper::GetRGBA(GLfloat* pFloats)
{
	HRESULT hr = S_OK;

	hr = GetGLfloat("r", &(pFloats[0]));
	if(FAILED(hr)) {
		return hr;
	}

	hr = GetGLfloat("g", &(pFloats[1]));
	if(FAILED(hr)) {
		return hr;
	}

	hr = GetGLfloat("b", &(pFloats[2]));
	if(FAILED(hr)) {
		return hr;
	}

	hr = GetGLfloat("a", &(pFloats[3]));
	if(FAILED(hr)) {
		return hr;
	}

	return S_OK;
}

// JSVALl擾܂B
HRESULT JSValueWrapper::GetJSVAL(const char* name, JSValueWrapper* pValue)
{
	if(name == NULL || pValue == NULL) {
		return E_POINTER;
	}
	JSObject* pObject = NULL;
	JSBool bret = JS_ValueToObject(m_pContext, m_value, & pObject);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	jsval value = JSVAL_NULL;
	bret = JS_GetProperty(m_pContext, pObject, name, & value);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	return pValue->Set(value);
}

// boollݒ肵܂B
HRESULT JSValueWrapper::SetBool(bool value)
{
	JSBool jsb = JS_FALSE;
	if(value) {
		jsb = JS_TRUE;
	}
	m_value = BOOLEAN_TO_JSVAL(jsb);
	return S_OK;
}

// intlݒ肵܂B
HRESULT JSValueWrapper::SetInt(int value)
{
	m_value = INT_TO_JSVAL(value);
	return S_OK;
}

// GLfloatlݒ肵܂B
HRESULT JSValueWrapper::SetGLfloat(GLfloat value)
{
	jsdouble dval = value;

	JSBool bret = JS_NewNumberValue(m_pContext, dval, & m_value);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}
	return S_OK;
}

// doublelݒ肵܂B
HRESULT JSValueWrapper::SetDouble(double value)
{
	jsdouble dval = value;

	JSBool bret = JS_NewNumberValue(m_pContext, dval, & m_value);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}
	return S_OK;
}

// GLfloatlݒ肵܂B
HRESULT JSValueWrapper::SetGLfloat(const char* name, GLfloat value)
{
	JSValueWrapper elem(m_pContext, JSVAL_ZERO);
	HRESULT hr = elem.SetGLfloat(value);
	if(FAILED(hr)) {
		return hr;
	}

	return SetJSVAL(name, & elem);
}

// WIuWFNgɕϊ܂B
HRESULT JSValueWrapper::SetXYZ(GLfloat x, GLfloat y, GLfloat z)
{
	HRESULT hr = S_OK;

	hr = SetGLfloat("x", x);
	if(FAILED(hr)) {
		return hr;
	}

	hr = SetGLfloat("y", y);
	if(FAILED(hr)) {
		return hr;
	}

	hr = SetGLfloat("z", z);
	if(FAILED(hr)) {
		return hr;
	}

	return S_OK;
}

// FIuWFNgɕϊ܂B
HRESULT JSValueWrapper::SetRGBA(const GLfloat* pFloats)
{
	HRESULT hr = S_OK;

	hr = SetGLfloat("r", pFloats[0]);
	if(FAILED(hr)) {
		return hr;
	}

	hr = SetGLfloat("g", pFloats[1]);
	if(FAILED(hr)) {
		return hr;
	}

	hr = SetGLfloat("b", pFloats[2]);
	if(FAILED(hr)) {
		return hr;
	}

	hr = SetGLfloat("a", pFloats[3]);
	if(FAILED(hr)) {
		return hr;
	}

	return S_OK;
}

// JSVALlݒ肵܂B
HRESULT JSValueWrapper::SetJSVAL(const char* name, JSValueWrapper* pValue)
{
	if(name == NULL || pValue == NULL) {
		return E_POINTER;
	}
	JSObject* pObject = NULL;
	if((!(JSVAL_IS_OBJECT(m_value))) || JSVAL_IS_NULL(m_value)) {
		pObject = JS_NewObject(m_pContext, NULL, NULL, NULL);
		if(pObject == NULL) {
			return E_FAIL;
		}
		m_value = OBJECT_TO_JSVAL(pObject);
	}
	JSBool bret = JS_ValueToObject(m_pContext, m_value, & pObject);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	jsval value = JSVAL_NULL;
	HRESULT hr = pValue->Get(& value);
	if(FAILED(hr)) {
		return hr;
	}
	bret = JS_SetProperty(m_pContext, pObject, name, & value);
	if(bret == JS_FALSE) {
		return E_FAIL;
	}

	return S_OK;
}

// l擾܂B
HRESULT JSValueWrapper::GetString(string& str)
{
	JSString* pStr = JS_ValueToString(m_pContext, m_value);
	if(pStr == NULL) {
		return E_FAIL;
	}

	char* pMBS = JS_EncodeString(m_pContext, pStr);
	if(pMBS == NULL) {
		return E_FAIL;
	}
	str = pMBS;
	JS_free(m_pContext, pMBS);
	pMBS = NULL;

	return S_OK;
}

// lݒ肵܂B
HRESULT JSValueWrapper::SetString(const char* str)
{
	if(str == NULL) {
		return E_POINTER;
	}

	size_t len = strlen(str);

	JSString* pNewStr = JS_NewStringCopyN(m_pContext, str, len);
	if(pNewStr == NULL) {
		return E_FAIL;
	}

	m_value = STRING_TO_JSVAL(pNewStr);
 
	return S_OK;
}
