#pragma once

#include "../js/JSMemberCollection.h"
#include "IARGPluginEnvironment.h"
#include "IARGPlugin.h"
#include "DefaultARGPluginEnvironment.h"
#include "../ar/IPatternDetector.h"
#include "ARGBaseUnknown.h"

class IWindow;
class ARWindow;
class JSEngineContext;
class JSDispatch;

using namespace std;

/**
* vOC}l[WłB
*/
class JSPluginManager : public JSMemberCollection
{
public:

	/**
	* \z܂B
	*/
	JSPluginManager(const char* basePath, IWindow* pWindow, ARWindow* pARWindow);

	/**
	* j܂B
	*/
	virtual ~JSPluginManager(void);

	/**
	* j܂B
	*/
	void Destroy();

	/**
	* JavaScriptGWݒ肵܂B
	*/
	HRESULT SetJSEngineContext(JSEngineContext* pContext);

protected:

	/**
	* o܂B
	*/
	HRESULT InitMembers();
	
	/**
	* ֐s܂B
	*/
	HRESULT InvokeFunction(JSContext *cx, JSObject *obj, const char* name, uintN argc, jsval *argv, jsval *rval);
	
	/**
	* vpeB擾܂B
	*/
	HRESULT GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp);

private:

	/**
	* x[XpXێ܂B
	*/
	string m_basePath;

	/**
	* EBhEێ܂B
	*/
	IWindow* m_pWindow;

	/**
	* AREBhEIuWFNgێ܂B
	*/
	ARWindow* m_pARWindow;

	/**
	* JavaScriptReLXgێ܂B
	*/
	JSEngineContext* m_pJSEngineContext;

	/**
	* vOC̈ꗗێ܂B
	*/
	vector<pair<HMODULE, IARGPlugin*>> m_plugins;

	/**
	* s̎łB
	*/
	class ARGPluginEnvironmentImpl : public DefaultARGPluginEnvironment
	{
	public:

		/**
		* \z܂B
		*/
		ARGPluginEnvironmentImpl(JSPluginManager* pOwner);

		/**
		* j܂B
		*/
		virtual ~ARGPluginEnvironmentImpl();

		/**
		* pX擾܂B
		*/
		virtual STDMETHODIMP GetRealPath(const char* relativePath, char* absolutePath, int maxLen, int* pLen);

		/**
		* vpeBǉ܂B
		*/
		virtual STDMETHODIMP AddProperty(const char* name, IDispatch* pDisp);

		/**
		* IuWFNgt@Ngǉ܂B
		*/
		virtual STDMETHODIMP AddObjectFactory(const char* name, IARGObjectFactory* pObjectFactory);

		/**
		* p^[oǉ܂B
		*/
		virtual STDMETHODIMP AddPatternDetector(const char* name, IARGPatternDetector* pPatternDetector);

		/**
		* R\[ɕo͂܂B
		*/
		virtual STDMETHODIMP PrintConsole(const char* text);

	private:

		/**
		* p^[õbp[łB
		*/
		class PatternDetectorWrapper : public IPatternDetector
		{
		public:

			/**
			* \z܂B
			*/
			PatternDetectorWrapper(JSEngineContext* pContext, IARGPatternDetector* pDetector);

			/**
			* j܂B
			*/
			virtual ~PatternDetectorWrapper();

			/**
			* j܂B
			*/
			void Destroy();

			/**
			* ̃p[^擾܂B
			*/
			HRESULT GetCreateParamCount(int* pCount);

			/**
			* p^[ǉ܂B
			*/
			HRESULT Create(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval);

			/**
			* osȂ܂B
			*/
			HRESULT Process(const CAPTURE_IMAGE* pImage);

			/**
			* p^[擾܂B
			*/
			HRESULT GetCount(int* pCount);

			/**
			* p^[擾܂B
			*/
			HRESULT Get(int index, IJSObject** ppObject);

		private:

			/**
			* p^[łB
			*/
			class JSPattern : public JSMemberCollection
			{
			public:

				/**
				* \z܂B
				*/
				JSPattern(IARGPattern* pPattern, HRESULT* phr);

				/**
				* j܂B
				*/
				virtual ~JSPattern();

				/**
				* j܂B
				*/
				void Destroy();

			protected:

				/**
				* ܂B
				*/
				virtual HRESULT InitMembers();
				
				/**
				* vpeBݒ肵܂B
				*/
				virtual HRESULT SetProperty(JSContext *cx, JSObject *obj, const char* name, jsval vp);
				
				/**
				* vpeB擾܂B
				*/
				virtual HRESULT GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp);

			private:

				/**
				* ύXs܂B
				*/
				HRESULT DispatchChanged(PATTERN_STATE state);

				/**
				* p^[Cxg𐶐܂B
				*/
				HRESULT CreatePatternEvent(PATTERN_STATE state, JSObject** ppEvent);

				/**
				* ĎpIuWFNg̎łB
				*/
				class ObserverImpl : public ARGBaseUnknown, public IARGPatternObserver
				{
				public:

					/**
					* \z܂B
					*/
					ObserverImpl(JSPattern* pOwner);

					/**
					* j܂B
					*/
					virtual ~ObserverImpl();

					DECLARE_BASE_UNKNOWN;

					/**
					* C^tF[X擾܂B
					*/
					virtual STDMETHODIMP QueryInterface(const IID& riid, void** ppv);

					/**
					* Cxg̏sȂ܂B
					*/
					virtual STDMETHODIMP DispatchChanged(PATTERN_STATE state);

				private:

					/**
					* I[i[ێ܂B
					*/
					JSPattern* m_pOwner;

				};

				/**
				* CxgIuWFNg̃bp[łB
				*/
				class JSEventImpl : public JSMemberCollection
				{
				public:

					/**
					* \z܂B
					*/
					JSEventImpl(JSPattern* pOwner, PATTERN_STATE state, HRESULT* phr);

					/**
					* j܂B
					*/
					virtual ~JSEventImpl();

					/**
					* j܂B
					*/
					virtual void Destroy();

					/**
					* Ԃݒ肵܂B
					*/
					HRESULT Set(PATTERN_STATE state);

				protected:

					/**
					* ܂B
					*/
					virtual HRESULT InitMembers();
					
					/**
					* vpeB擾܂B
					*/
					virtual HRESULT GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp);

				private:

					/**
					* LIuWFNgێ܂B
					*/
					JSPattern* m_pOwner;

					/**
					* }[J[`ς݂ǂێ܂B
					*/
					PATTERN_STATE m_state;

				};

				/**
				* p^[ێ܂B
				*/
				IARGPattern* m_pPattern;

				/**
				* p[^ێ܂B
				*/
				JSDispatch* m_pParam;

				/**
				* }[J[𔭌CxgłB
				*/
				FunctionRef m_onFound;

				/**
				* }[J[CxgłB
				*/
				FunctionRef m_onLost;

				/**
				* }[J[ωCxgłB
				*/
				FunctionRef m_onChanged;

				/**
				* JavaScript EventIuWFNgێ܂B
				*/
				JSEventImpl* m_pJSEvent;

			};

			/**
			* ReLXgێ܂B
			*/
			JSEngineContext* m_pContext;

			/**
			* oێ܂B
			*/
			IARGPatternDetector* m_pDetector;

			/**
			* p^[̈ꗗێ܂B
			*/
			vector<JSPattern*> m_patterns;

		};

		/**
		* I[i[ێ܂B
		*/
		JSPluginManager* m_pOwner;
	};

	/**
	* sێ܂B
	*/
	ARGPluginEnvironmentImpl* m_pEnv;

};
