#include "StdAfx.h"
#include "ARCaptureDevice.h"

#define VIEW_DISTANCE_MIN		4.0			// Objects closer to the camera than this will not be displayed.
#define VIEW_DISTANCE_MAX		4000.0		// Objects further away from the camera than this will not be displayed.

// \z܂B
ARCaptureDevice::ARCaptureDevice()
{
	m_arglSettings = NULL;
	m_pData = NULL;
	ZeroMemory(& m_cparam, sizeof(m_cparam));
	m_cRef = 1;
}

// j܂B
ARCaptureDevice::~ARCaptureDevice(void)
{
    arVideoClose();

	m_pData = NULL;
	if(m_arglSettings != NULL) {
		arglCleanup(m_arglSettings);
		m_arglSettings = NULL;
	}
}

// ̃IuWFNgj܂B
void ARCaptureDevice::Destroy()
{
	Stop();
	Release();
}

// foCX擾܂B
HRESULT ARCaptureDevice::GetDeviceCount(int* pCount) 
{
	return E_NOTIMPL;
}

// foCX擾܂B
HRESULT ARCaptureDevice::GetDevice(int index, string& name, CAPTURE_DEVICE_ATTR* pAttr) 
{
	return E_NOTIMPL;
}

// ܂B
HRESULT ARCaptureDevice::Init(const char* config) 
{
	string sconfig = config;
	char* pconfig = new char[sconfig.size() + 1];
	ZeroMemory(pconfig, sizeof(char) * (sconfig.size() + 1));
	memcpy(pconfig, sconfig.c_str(), sizeof(char) * (sconfig.size() + 1));

	int ret = arVideoOpen(pconfig);
	delete[] pconfig;
	pconfig = NULL;

	if(ret < 0) {
		return E_FAIL;
	}

	return S_OK;
}

// J`܂B
HRESULT ARCaptureDevice::InitCamera(const char* config)
{
	int xsize = 0, ysize = 0;
	
	HRESULT hr = GetSize(& xsize, & ysize);
	if(FAILED(hr)) {
		return hr;
	}

    ARParam  wparam;
	// {ݒ[h
    if(arParamLoad(config, 1, &wparam) < 0) {
		return E_FAIL;
    }

	// TCYɂ킹
    arParamChangeSize( &wparam, xsize, ysize, &m_cparam );
    arInitCparam( &m_cparam );

	return S_OK;
}

// TCY擾܂B
HRESULT ARCaptureDevice::GetSize(int* pWidth, int* pHeight) 
{
	if(pWidth == NULL || pHeight == NULL) {
		return E_POINTER;
	}
	if(arVideoInqSize(pWidth, pHeight) < 0) {
		return E_FAIL;
	}
	return S_OK;
}

// OpenGLp̕ϊs擾܂B
HRESULT ARCaptureDevice::GetProjectionMatrix(double* pMatrix)
{
	arglCameraFrustumRH(&m_cparam, VIEW_DISTANCE_MIN, VIEW_DISTANCE_MAX, pMatrix);
	return S_OK;
}

// foCXJn܂B
HRESULT ARCaptureDevice::Start() 
{
    arVideoCapStart();

	// arglCȕ
	if ((m_arglSettings = arglSetupForCurrentContext()) == NULL) {
		return E_FAIL;
	}
	return S_OK;
}

// foCX~܂B
HRESULT ARCaptureDevice::Stop() 
{
    arVideoCapStop();
	return S_OK;
}

// C[W擾܂B
HRESULT ARCaptureDevice::GetImage(CAPTURE_IMAGE* pImage) 
{
	if(pImage == NULL) {
		return E_POINTER;
	}

	AddRef();
    if( (m_pData = arVideoGetImage()) == NULL ) {
		Release();
        return S_FALSE;
    }

	pImage->pBuffer = m_pData;
	Release();

	if(arVideoInqSize(&(pImage->width), &(pImage->height)) < 0) {
		return E_FAIL;
	}
	return S_OK;
}

// C[WJ܂B
HRESULT ARCaptureDevice::ReleaseImage(CAPTURE_IMAGE* pImage)
{
	if(pImage == NULL) {
		return E_POINTER;
	}

	return S_OK;
}

// ŌGetImagef[^`悵܂B
HRESULT ARCaptureDevice::Draw() 
{
	if(m_pData == NULL) {
		return S_FALSE;
	}
	arglDispImage(m_pData, &m_cparam, 1.0, m_arglSettings);	// zoom = 1.0.

	return S_OK;
}

// ̃t[Lv`܂B
HRESULT ARCaptureDevice::CaptureNext()
{
	arVideoCapNext();
	m_pData = NULL;

	return S_OK;
}

// ݂̏|[g܂B
HRESULT ARCaptureDevice::ReportStatus(ostream* pOut)
{
    (*pOut) << "*** Camera Parameter ***\n";
	(*pOut) << "width = " << m_cparam.xsize;
	(*pOut) << ", height = " << m_cparam.xsize << "\n";
	
	/*if (arglDrawModeGet(g_arglSettings) == AR_DRAW_BY_GL_DRAW_PIXELS) {
		fprintf(stderr, "DrawMode (C)   : GL_DRAW_PIXELS\n");
	} else if (arglTexmapModeGet(g_arglSettings) == AR_DRAW_TEXTURE_FULL_IMAGE) {
		fprintf(stderr, "DrawMode (C)   : TEXTURE MAPPING (FULL RESOLUTION)\n");
	} else {
		fprintf(stderr, "DrawMode (C)   : TEXTURE MAPPING (HALF RESOLUTION)\n");
	}*/

	return S_OK;
}

// QƉ񐔂Z܂B
LONG ARCaptureDevice::AddRef()
{
	return InterlockedIncrement(&m_cRef);
}

// QƉ񐔂Z܂B
LONG ARCaptureDevice::Release()
{
	LONG res = InterlockedDecrement(&m_cRef);
	if(0 == res) {
		delete this;
	}
	return res;
}
