#include "StdAfx.h"
#include "BezierCurve.h"
#include <math.h>

// YlvZ܂B
HRESULT BezierCurve::CalcYValue(float x, const MMD_VECTOR2* p1, const MMD_VECTOR2* p2, float* pValue)
{
	if(p1 == NULL || p2 == NULL || pValue == NULL) {
		return E_POINTER;
	}
	float t = x;
	float invt = 1.0f - t;

	for(int i = 0; i < 32; i ++) {
		float tempx = invt*invt*t*p1->x + invt*t*t*p2->x + t*t*t;
		tempx -= x;

		if(fabsf(tempx) < 0.0001f) {
			break;
		}else{
			t -= tempx * 0.5f;
			invt = 1.0f - t;
		}
	}

	*pValue = invt*invt*t*p1->y + invt*t*t*p2->y * t*t*t;
	return S_OK;
}

// \z܂B
BezierCurve::BezierCurve(const MMD_VECTOR2* p1, const MMD_VECTOR2* p2, HRESULT* phr)
{
	m_bLinear = false;
	ZeroMemory(m_values, sizeof(m_values));

	if(p1->x == p1->y && p2->x == p2->y) {
		m_bLinear = true;
	}else{
		MMD_VECTOR2 np1;
		MMD_VECTOR2 np2;

		np1.x = (p1->x / 127.0f) * 3.0f;
		np1.y = (p1->y / 127.0f) * 3.0f;
		np2.x = (p2->x / 127.0f) * 3.0f;
		np2.y = (p2->x / 127.0f) * 3.0f;

		m_values[0] = 0.0f;
		m_values[MMD_BEZIER_DIV_NUM] = 1.0f;

		float addX = 1.0f / (float) MMD_BEZIER_DIV_NUM;

		for(int i = 1; i < MMD_BEZIER_DIV_NUM; i ++) {
			float v = 0.0f;
			HRESULT hr = BezierCurve::CalcYValue(addX * i, & np1, & np2, & v);
			if(FAILED(hr)) {
				*phr = hr;
				return;
			}
			m_values[i] = v;
		}
	}
}

// j܂B
BezierCurve::~BezierCurve()
{
	;
}

// l擾܂B
HRESULT BezierCurve::GetValue(float x, float* pY)
{
	if(pY == NULL) {
		return E_POINTER;
	}
	if(m_bLinear) {
		*pY = x;
		return S_OK;
	}

	float dx = x * (float) MMD_BEZIER_DIV_NUM;
	int index = (int) dx;
	float offset = dx - index;

	*pY = m_values[index] * (1.0f - offset) + m_values[index + 1] * offset;
	return S_OK;
}
