#include "StdAfx.h"
#include "MMDModel.h"
#include "model/MMDVertexList.h"
#include "model/MMDBone.h"
#include "model/MMDMorp.h"
#include "model/MMDMaterial.h"
#include "model/MMDIK.h"
#include "IDataMutex.h"

// \z܂B
MMDModel::MMDModel()
{
	m_pVertexList = NULL;
	m_scale = 1.0;
}

// j܂B
MMDModel::~MMDModel()
{
	for(vector<MMDMaterial*>::size_type i = 0; i < m_materials.size(); i ++) {
		MMDMaterial* pTarget = m_materials[i];
		if(pTarget != NULL) {
			delete pTarget;
			pTarget = NULL;
		}
		m_materials[i] = NULL;
	}
	m_materials.clear();
	for(vector<MMDMorp*>::size_type i = 0; i < m_morps.size(); i ++) {
		MMDMorp* pTarget = m_morps[i];
		if(pTarget != NULL) {
			delete pTarget;
			pTarget = NULL;
		}
		m_morps[i] = NULL;
	}
	m_morps.clear();
	for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
		MMDBone* pTarget = m_bones[i];
		if(pTarget != NULL) {
			delete pTarget;
			pTarget = NULL;
		}
		m_bones[i] = NULL;
	}
	m_bones.clear();
	for(vector<MMDIK*>::size_type i = 0; i < m_iks.size(); i ++) {
		MMDIK* pTarget = m_iks[i];
		if(pTarget != NULL) {
			delete pTarget;
			pTarget = NULL;
		}
		m_iks[i] = NULL;
	}
	m_iks.clear();

	if(m_pVertexList != NULL) {
		delete m_pVertexList;
		m_pVertexList = NULL;
	}
}

// PMDt@CJ܂B
HRESULT MMDModel::OpenPMD(const char* path)
{
	if(path == NULL) {
		return E_POINTER;
	}
	if(m_pVertexList != NULL) {
		return E_UNEXPECTED;
	}

	clsPMDFile pmdFile;
	BOOL br = pmdFile.Open(path);
	if(br == FALSE) {
		return E_UNEXPECTED;
	}

	// f̍\z
	PMD_VERTEX_CHUNK vertexes = pmdFile.GetVertexChunk();
	PMD_INDEX_CHUNK indices = pmdFile.GetIndexChunk();

	HRESULT hr = S_OK;
	m_pVertexList = new MMDVertexList(& vertexes, & indices, & hr);
	if(FAILED(hr)) {
		return hr;
	}

	PMD_BONE_CHUNK bones = pmdFile.GetBoneChunk();
	for(PMD_BONE_CHUNK::size_type i = 0; i < bones.size(); i ++) {
		hr = S_OK;
		PMD_BONE_RECORD bone = bones[i];
		m_bones.push_back(new MMDBone(& bone, & hr));
		if(FAILED(hr)) {
			return hr;
		}
	}
	for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
		hr = m_bones[i]->Init(m_bones);
		if(FAILED(hr)) {
			return hr;
		}
	}

	PMD_MORP_CHUNK morps = pmdFile.GetMorpChunk();
	for(PMD_MORP_CHUNK::size_type i = 0; i < morps.size(); i ++) {
		hr = S_OK;
		PMD_MORP_RECORD morp = morps[i];
		if(i == 0) {
			m_morps.push_back(new MMDMorp(& morp, NULL, & hr));
			if(FAILED(hr)) {
				return hr;
			}
		}else{
			PMD_MORP_RECORD baseMorp = morps[0];
			m_morps.push_back(new MMDMorp(& morp, & baseMorp, & hr));
			if(FAILED(hr)) {
				return hr;
			}
		}
	}

	PMD_IK_CHUNK iks = pmdFile.GetIKChunk();
	for(PMD_IK_CHUNK::size_type i = 0; i < iks.size(); i ++) {
		hr = S_OK;
		PMD_IK_RECORD ik = iks[i];
		m_iks.push_back(new MMDIK(& ik, & hr));
		if(FAILED(hr)) {
			return hr;
		}
	}
	std::sort(m_iks.begin(), m_iks.end(), MMDIK::Compare);
	for(vector<MMDIK*>::size_type i = 0; i < m_iks.size(); i ++) {
		hr = m_iks[i]->Init(m_bones);
		if(FAILED(hr)) {
			return hr;
		}
	}

	PMD_MATERIAL_CHUNK materials = pmdFile.GetMaterialChunk();
	for(PMD_MATERIAL_CHUNK::size_type i = 0; i < materials.size(); i ++) {
		hr = S_OK;
		PMD_MATERIAL_RECORD material = materials[i];
		m_materials.push_back(new MMDMaterial(& material, & hr));
		if(FAILED(hr)) {
			return hr;
		}
	}
	int offset = 0;
	int nextOffset = 0;
	for(vector<MMDMaterial*>::size_type i = 0; i < m_materials.size(); i ++) {
		hr = m_materials[i]->Init(m_pVertexList, m_bones, offset, & nextOffset);
		if(FAILED(hr)) {
			return hr;
		}
		offset = nextOffset;
	}

	return S_OK;
}

// VMDt@CJ܂B
HRESULT MMDModel::OpenVMD(const char* path)
{
	if(path == NULL) {
		return E_POINTER;
	}
	clsVMDFile vmdFile;
	BOOL br = vmdFile.Open(path);
	if(br == FALSE) {
		return E_UNEXPECTED;
	}
	HRESULT hr = S_OK;

	// Bone
	VMD_MOTION_CHUNK motions = vmdFile.GetMotionChunk();
	vector<string> motionNames;
	for(VMD_MOTION_CHUNK::size_type j = 0; j < motions.size(); j ++) {
		VMD_MOTION_RECORD motion = motions[j];

		string motionName = motion.name;
		motionNames.push_back(motionName);

		bool added = false;
		for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
			MMDBone* pBone = m_bones[i];
			hr = pBone->IsTarget(& motion);
			if(FAILED(hr)) {
				return hr;
			}
			if(hr == S_OK) {
				hr = pBone->AddMotion(& motion);
				if(FAILED(hr)) {
					return hr;
				}
				added = true;
				break;
			}
		}
		if(added == false) {
			// TODO: x\H
		}
	}
	for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
		MMDBone* pBone = m_bones[i];
		hr = pBone->PrepareMotion();
		if(FAILED(hr)) {
			return hr;
		}
	}

	// Face
	VMD_MORP_CHUNK morps = vmdFile.GetMorpChunk();
	for(VMD_MORP_CHUNK::size_type j = 0; j < morps.size(); j ++) {
		VMD_MORP_RECORD morp = morps[j];
		bool added = false;
		for(vector<MMDMorp*>::size_type i = 0; i < m_morps.size(); i ++) {
			MMDMorp* pMorp = m_morps[i];
			hr = pMorp->IsTarget(& morp);
			if(FAILED(hr)) {
				return hr;
			}
			if(hr == S_OK) {
				hr = pMorp->AddMotion(& morp);
				if(FAILED(hr)) {
					return hr;
				}
				added = true;
				break;
			}
		}
		if(added == false) {
			// TODO: x\H
		}
	}
	for(vector<MMDMorp*>::size_type i = 0; i < m_morps.size(); i ++) {
		MMDMorp* pMorp = m_morps[i];
		hr = pMorp->PrepareMotion();
		if(FAILED(hr)) {
			return hr;
		}
	}
	return S_OK;
}

// eNX`܂B
HRESULT MMDModel::Prepare(IMMDTextureProvider* pTextureProvider)
{
	if(pTextureProvider == NULL) {
		return E_POINTER;
	}
	for(vector<MMDMaterial*>::size_type i = 0; i < m_materials.size(); i ++) {
		HRESULT hr = m_materials[i]->Prepare(pTextureProvider);
		if(FAILED(hr)) {
			return hr;
		}
	}
	return S_OK;
}

// XP[Oݒ肵܂B
HRESULT MMDModel::SetScale(float scale)
{
	m_scale = scale;
	return S_OK;
}

// XP[O擾܂B
HRESULT MMDModel::GetScale(float* pScale)
{
	if(pScale == NULL) {
		return E_POINTER;
	}
	*pScale = m_scale;
	return S_OK;
}

// [Vj܂B
HRESULT MMDModel::ClearMotion()
{
	for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
		HRESULT hr = m_bones[i]->ClearMotion();
		if(FAILED(hr)) {
			return hr;
		}
	}
	for(vector<MMDMorp*>::size_type i = 0; i < m_morps.size(); i ++) {
		HRESULT hr = m_morps[i]->ClearMotion();
		if(FAILED(hr)) {
			return hr;
		}
	}
	return S_OK;
}

// XVsȂ܂B
HRESULT MMDModel::UpdateAsync(IDataMutex* pMutex, float frameNo)
{
	if(m_pVertexList == NULL || pMutex == NULL) {
		return E_POINTER;
	}
	HRESULT hr = UpdateMotion(frameNo);
	if(FAILED(hr)) {
		return hr;
	}

	for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
		hr = m_bones[i]->UpdateSkinning();
		if(FAILED(hr)) {
			return hr;
		}
	}

	pMutex->Begin();
	hr = m_pVertexList->UpdateSkinning();
	pMutex->End();

	return hr;
}

// `揈sȂ܂B
HRESULT MMDModel::DrawAsync(IDataMutex* pMutex)
{
	if(pMutex == NULL) {
		return E_POINTER;
	}
	pMutex->Begin();
	HRESULT hr = UpdateVertexBuffer();
	pMutex->End();
	if(FAILED(hr)) {
		return hr;
	}

	glPushMatrix();
	glScalef(m_scale, m_scale, m_scale * -1.0f);
	GLboolean normalizeEnabled = glIsEnabled(GL_NORMALIZE);
	glEnable(GL_NORMALIZE);

	for(vector<MMDMaterial*>::size_type i = 0; i < m_materials.size(); i ++) {
		HRESULT hr = m_materials[i]->Draw();
		if(FAILED(hr)) {
			return hr;
		}
	}

	if(normalizeEnabled == GL_FALSE) {
		glDisable(GL_NORMALIZE);
	}
	glPopMatrix();
	return S_OK;
}

// \ݒ肵܂B
HRESULT MMDModel::SetFace(const char* faceName)
{
	if(faceName == NULL) {
		return E_POINTER;
	}
	string name = faceName;
	MMDMorp* pSelectedElem = NULL;

	for(vector<MMDMorp*>::size_type i = 0; i < m_morps.size(); i ++) {
		MMDMorp* pElem = m_morps[i];
		string elemName;
		HRESULT hr = pElem->GetName(elemName);
		if(FAILED(hr)) {
			return hr;
		}
		if(elemName == name) {
			pSelectedElem = pElem;
			break;
		}
	}
	if(pSelectedElem == NULL) {
		return E_UNEXPECTED;
	}
	MMD_VERTEX_DESC** ppOriginalDescs = NULL;
	HRESULT hr = m_pVertexList->GetVertexDescs(& ppOriginalDescs);
	if(FAILED(hr)) {
		return hr;
	}

	return pSelectedElem->Set(ppOriginalDescs);
}

// \擾܂B
HRESULT MMDModel::GetFaceCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = m_morps.size();
	return S_OK;
}

// \擾܂B
HRESULT MMDModel::GetFaceName(int index, string& faceName)
{
	if((vector<MMDMorp*>::size_type) index >= m_morps.size()) {
		return E_INVALIDARG;
	}
	return m_morps[index]->GetName(faceName);
}

// Bone擾܂B
HRESULT MMDModel::GetBoneCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = m_bones.size();
	return S_OK;
}

// Bone擾܂B
HRESULT MMDModel::GetBoneName(int index, string& boneName)
{
	if((vector<MMDBone*>::size_type) index >= m_bones.size()) {
		return E_INVALIDARG;
	}
	return m_bones[index]->GetName(boneName);
}

// IK擾܂B
HRESULT MMDModel::GetIKCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = m_iks.size();
	return S_OK;
}

// IK擾܂B
HRESULT MMDModel::GetIKTargetName(int index, string& ikName)
{
	if((vector<MMDIK*>::size_type) index >= m_iks.size()) {
		return E_INVALIDARG;
	}
	return m_iks[index]->GetTargetName(ikName);
}

// IKLǂ擾܂B
HRESULT MMDModel::IsIKEnabled(int index, bool* pValue)
{
	if((vector<MMDIK*>::size_type) index >= m_iks.size()) {
		return E_INVALIDARG;
	}
	return m_iks[index]->IsEnabled(pValue);
}

// IKLǂݒ肵܂B
HRESULT MMDModel::SetIKEnabled(int index, bool value)
{
	if((vector<MMDIK*>::size_type) index >= m_iks.size()) {
		return E_INVALIDARG;
	}
	return m_iks[index]->SetEnabled(value);
}

// BonẻԂݒ肵܂B
HRESULT MMDModel::SetBoneVisible(int index, bool visible)
{
	if((vector<MMDBone*>::size_type) index >= m_bones.size()) {
		return E_INVALIDARG;
	}
	MMDBone* pBone = m_bones[index];
	bool curVisible = false;
	HRESULT hr = pBone->IsVisible(& curVisible);
	if(FAILED(hr)) {
		return hr;
	}
	if(curVisible == visible) {
		return S_OK;
	}

	hr = pBone->SetVisible(visible);
	if(FAILED(hr)) {
		return hr;
	}

	for(vector<MMDMaterial*>::size_type i = 0; i < m_materials.size(); i ++) {
		MMDMaterial* pElem = m_materials[i];
		hr = pElem->UpdateVisibility();
		if(FAILED(hr)) {
			return hr;
		}
	}
	return S_OK;
}

// BonẻԂ擾܂B
HRESULT MMDModel::IsBoneVisible(int index, bool* pVisible)
{
	if(pVisible == NULL) {
		return E_POINTER;
	}
	if((vector<MMDBone*>::size_type) index >= m_bones.size()) {
		return E_INVALIDARG;
	}
	return m_bones[index]->IsVisible(pVisible);
}

// őt[擾܂B
HRESULT MMDModel::GetMaxFrame(int* pFrameNo)
{
	if(pFrameNo == NULL) {
		return E_POINTER;
	}
	int ret = 0;
	int validCount = 0;
	for(vector<MMDMorp*>::size_type i = 0; i < m_morps.size(); i ++) {
		int f = 0;
		HRESULT hr = m_morps[i]->GetMaxFrame(& f);
		if(FAILED(hr)) {
			return hr;
		}
		if(hr != S_OK) {
			continue;
		}
		validCount ++;
		if(f > ret) {
			ret = f;
		}
	}
	for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
		int f = 0;
		HRESULT hr = m_bones[i]->GetMaxFrame(& f);
		if(FAILED(hr)) {
			return hr;
		}
		if(hr != S_OK) {
			continue;
		}
		validCount ++;
		if(f > ret) {
			ret = f;
		}
	}
	if(validCount == 0) {
		return S_FALSE;
	}
	*pFrameNo = ret;
	return ret;
}

// _obt@Zbg܂B
HRESULT MMDModel::ResetVertexes()
{
	if(m_pVertexList == NULL) {
		return E_POINTER;
	}
	return m_pVertexList->ResetVertexes();
}

// XLjOsXV܂B
HRESULT MMDModel::UpdateSkinning()
{
	if(m_pVertexList == NULL) {
		return E_POINTER;
	}
	for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
		HRESULT hr = m_bones[i]->UpdateSkinning();
		if(FAILED(hr)) {
			return hr;
		}
	}

	return m_pVertexList->UpdateSkinning();
}

// _obt@XV܂B
HRESULT MMDModel::UpdateVertexBuffer()
{
	for(vector<MMDMaterial*>::size_type i = 0; i < m_materials.size(); i ++) {
		HRESULT hr = m_materials[i]->UpdateVertexBuffer();
		if(FAILED(hr)) {
			return hr;
		}
	}
	return S_OK;
}

// [VXV܂B
HRESULT MMDModel::UpdateMotion(float elapsedFrame)
{
	if(m_pVertexList == NULL) {
		return S_FALSE;
	}
	MMD_VERTEX_DESC** ppOriginalDescs = NULL;
	HRESULT hr = m_pVertexList->GetVertexDescs(& ppOriginalDescs);
	if(FAILED(hr)) {
		return hr;
	}

	// \
	if(m_morps.size() > 0) {
		hr = m_morps[0]->Set(ppOriginalDescs);
		if(FAILED(hr)) {
			return hr;
		}
	}

	for(vector<MMDMorp*>::size_type i = 0; i < m_morps.size(); i ++) {
		hr = m_morps[i]->ApplyMotion(elapsedFrame, ppOriginalDescs);
		if(FAILED(hr)) {
			return hr;
		}
	}

	// Bone
	for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
		hr = m_bones[i]->UpdateMotion(elapsedFrame);
		if(FAILED(hr)) {
			return hr;
		}
	}
	for(vector<MMDBone*>::size_type i = 0; i < m_bones.size(); i ++) {
		hr = m_bones[i]->UpdateMatrix();
		if(FAILED(hr)) {
			return hr;
		}
	}

	// IK
	for(vector<MMDIK*>::size_type i = 0; i < m_iks.size(); i ++) {
		HRESULT hr = m_iks[i]->Update();
		if(FAILED(hr)) {
			return hr;
		}
	}
	return S_OK;
}
