#pragma once

#include "../BaseTypes.h"
#include "../io/d3dxhelper.h"
#include "../io/common.h"
#include "../io/MMD_IO.h"
#include "../io/clsPMDFile.h"
#include "../io/clsVMDFile.h"
#include "MMDVertexList.h"

class BezierCurve;

/**
* Bone̎łB
*/
class MMDBone
{
private:

	class Motion;

	/**
	* [Vmr܂B
	*/
	static bool Less(const Motion* pLeft, const Motion* pRight);

public:

	/**
	* \z܂B
	*/
	MMDBone(const PMD_BONE_RECORD* pBone, HRESULT* phr);

	/**
	* j܂B
	*/
	virtual ~MMDBone();

	/**
	* ܂B
	*/
	HRESULT Init(vector<MMDBone*>& pBones);

	/**
	* Ԃݒ肵܂B
	*/
	HRESULT SetVisible(bool visible);

	/**
	* Ԃ擾܂B
	*/
	HRESULT IsVisible(bool* pVisible);

	/**
	* vZp̍sȂǂZbg܂B
	*/
	HRESULT Reset();

	/**
	* [Vj܂B
	*/
	HRESULT ClearMotion();

	/**
	* O擾܂B
	*/
	HRESULT GetName(string& name);

	/**
	* ǉΏۂǂ𔻒肵܂B
	*/
	HRESULT IsTarget(const VMD_MOTION_RECORD* pMotion);

	/**
	* [Vǉ܂B
	*/
	HRESULT AddMotion(const VMD_MOTION_RECORD* pMotion);

	/**
	* XLjOsXV܂B
	*/
	HRESULT UpdateSkinning();

	/**
	* XLjOs𒸓_ɓKp܂B(1_)
	*/
	HRESULT ApplySkinning(MMD_VERTEX_TEXUSE* pDest, const MMD_VERTEX_TEXUSE* pOriginal);

	/**
	* XLjOs𒸓_ɓKp܂B(2_)
	*/
	HRESULT ApplySkinning(MMD_VERTEX_TEXUSE* pDest, const MMD_VERTEX_TEXUSE* pOriginal, MMDBone* pBone, float bweight);

	/**
	* [V̂߂̏sȂ܂B
	*/
	HRESULT PrepareMotion();

	/**
	* [VXV܂B
	*/
	HRESULT UpdateMotion(float elapsedFrame);

	/**
	* ϊpsXV܂B
	*/
	HRESULT UpdateMatrix();

	/**
	* őt[擾܂B
	*/
	HRESULT GetMaxFrame(int* pFrameNo);

	/**
	* IK֘Ãp[^XV܂B
	*/
	HRESULT UpdateIKLimitAngle();

	/**
	* IKLimitAnglesǂ𔻒肵܂B
	*/
	HRESULT IsIKLimitAngle(bool* pValue);

	/**
	* ϊs񂩂xNg擾܂B
	*/
	HRESULT GetPositionFromLocal(MMD_VECTOR3* pVector);

	/**
	* ϊs擾܂B
	*/
	HRESULT GetLocal(MMD_MATRIX* pMatrix);

	/**
	* sړE]擾܂B
	*/
	HRESULT GetVectors(MMD_VECTOR3* pPos, MMD_VECTOR4* pQt);

	/**
	* sړE]ݒ肵܂B
	*/
	HRESULT SetVectors(const MMD_VECTOR3* pPos, const MMD_VECTOR4* pQt);

private:

	/**
	* ɑΉ郂[V擾܂B
	*/
	HRESULT FindMotion(float elapsedTime, Motion** ppBegin, Motion** ppEnd, float* pOffset);

	/**
	* Boneێ܂B
	*/
	PMD_BONE_RECORD m_bone;

	/**
	* eێ܂B
	*/
	MMDBone* m_pParent;

	/**
	* qێ܂B
	*/
	MMDBone* m_pChild;

	/**
	* ItZbgێ܂B
	*/
	MMD_VECTOR3 m_offset;

	/**
	* tϊsێ܂B
	*/
	MMD_MATRIX m_invTransform;

	/**
	* sړێ܂B
	*/
	MMD_VECTOR3 m_effectPosition;

	/**
	* ]ێ܂B
	*/
	MMD_VECTOR4 m_effectRotation;

	/**
	* ϊsێ܂B
	*/
	MMD_MATRIX m_effectLocal;

	/**
	* XL̏psێ܂B
	*/
	MMD_MATRIX m_effectSkinning;

	/**
	* IKɊւtOێ܂B
	*/
	bool m_bIKLimitAngle;

	/**
	* ֘A[Vێ܂B
	*/
	vector<Motion*> m_motions;

	/**
	* Ԃێ܂B
	*/
	bool m_bVisible;

	/**
	* [VǗ܂B
	*/
	class Motion
	{
	public:

		/**
		* \z܂B
		*/
		Motion(const VMD_MOTION_RECORD* pMotion, HRESULT* phr);

		/**
		* j܂B
		*/
		virtual ~Motion();

		/**
		* t[ԍ擾܂B
		*/
		HRESULT GetFrameNo(int* pFrameNo) const;

		/**
		* ړxNgƃNH[^jI擾܂B
		*/
		HRESULT GetVectors(MMD_VECTOR3* pPos, MMD_VECTOR4* pQt);

		/**
		* ԏsȂ܂B
		*/
		HRESULT Lerp(MMD_VECTOR3* pDest, const MMD_VECTOR3* pValue1, const MMD_VECTOR3* pValue2, float weight);

		/**
		* ԏsȂ܂B
		*/
		HRESULT Lerp(MMD_VECTOR4* pDest, const MMD_VECTOR4* pValue1, const MMD_VECTOR4* pValue2, float weight);

	private:

		/**
		* [Vێ܂B
		*/
		VMD_MOTION_RECORD m_motion;

		/**
		* ʒuێ܂B
		*/
		MMD_VECTOR3 m_pos;

		/**
		* NH[^jIێ܂B
		*/
		MMD_VECTOR4 m_qt;

		/**
		* ẌړȐێ܂B
		*/
		BezierCurve* m_pXBez;

		/**
		* ŸړȐێ܂B
		*/
		BezierCurve* m_pYBez;

		/**
		* Z̈ړȐێ܂B
		*/
		BezierCurve* m_pZBez;

		/**
		* ]ʂ̈ړȐێ܂B
		*/
		BezierCurve* m_pRotBez;

	};
};
