#include "StdAfx.h"
#include "MMDIK.h"
#include <math.h>
#include "MMDBone.h"
#include "../CalcUtil.h"

// IKmr܂B
bool MMDIK::Compare(const MMDIK* pLeft, const MMDIK* pRight)
{
	if(pLeft == NULL || pRight == NULL) {
		return false;
	}
	int sortVal1 = 0;
	HRESULT hr = pLeft->GetSortValue(&sortVal1);
	if(FAILED(hr)) {
		return false;
	}
	int sortVal2 = 0;
	hr = pRight->GetSortValue(&sortVal2);
	if(FAILED(hr)) {
		return false;
	}
	if(sortVal1 < sortVal2) {
		return true;
	}
	if(sortVal1 > sortVal2) {
		return false;
	}
	return false;
}

// \z܂B
MMDIK::MMDIK(const PMD_IK_RECORD* pIK, HRESULT* phr)
{
	m_pTarget = NULL;
	m_pEffect = NULL;
	m_fact = 0.0f;
	m_bEnabled = true;

	if(pIK == NULL) {
		*phr = E_POINTER;
		return;
	}
	if(pIK->link.size() == 0) {
		*phr = E_INVALIDARG;
		return;
	}
	m_ik = *pIK;
}

// j܂B
MMDIK::~MMDIK()
{
	m_pTarget = NULL;
	m_pEffect = NULL;
}

// ܂B
HRESULT MMDIK::Init(vector<MMDBone*>& bones)
{
	m_pTarget = bones[m_ik.parent];
	m_pEffect = bones[m_ik.to];
	m_fact = m_ik.fact * 3.1415926f;

	for(vector<WORD>::size_type i = 0; i < m_ik.link.size(); i ++) {
		MMDBone* pBone = bones[m_ik.link[i]];
		m_bones.push_back(pBone);
		HRESULT hr = pBone->UpdateIKLimitAngle();
		if(FAILED(hr)) {
			return hr;
		}
	}

	return S_OK;
}

// O擾܂B
HRESULT MMDIK::GetTargetName(string& name)
{
	if(m_pTarget == NULL) {
		return E_UNEXPECTED;
	}
	return m_pTarget->GetName(name);
}

// Lǂݒ肵܂B
HRESULT MMDIK::SetEnabled(bool enabled)
{
	m_bEnabled = enabled;
	return S_OK;
}

// Lǂ擾܂B
HRESULT MMDIK::IsEnabled(bool* pEnabled)
{
	if(pEnabled == NULL) {
		return E_POINTER;
	}
	*pEnabled = m_bEnabled;
	return S_OK;
}

// XV܂B
HRESULT MMDIK::Update()
{
	if(m_pTarget == NULL || m_pEffect == NULL) {
		return E_UNEXPECTED;
	}
	if(m_bEnabled == false) {
		return S_OK;
	}
	HRESULT hr = S_OK;

	for(vector<MMDBone*>::reverse_iterator it = m_bones.rbegin(); it != m_bones.rend(); it ++) {
		hr = (*it)->UpdateMatrix();
		if(FAILED(hr)) {
			return hr;
		}
	}

	hr = m_pEffect->UpdateMatrix();
	if(FAILED(hr)) {
		return hr;
	}

	MMD_VECTOR3 targetOriginalPosition;
	hr = m_pTarget->GetPositionFromLocal(& targetOriginalPosition);
	if(FAILED(hr)) {
		return hr;
	}

	for(int it = 0; it < m_ik.count; it ++) {
		int index = 0;
		for(vector<MMDBone*>::iterator bit = m_bones.begin(); bit != m_bones.end(); bit ++, index ++) {
			MMD_VECTOR3 effectPosition;
			hr = m_pEffect->GetPositionFromLocal(& effectPosition);
			if(FAILED(hr)) {
				return hr;
			}

			MMDBone* pBone = *bit;
			MMD_MATRIX matBone;
			MMD_MATRIX matInvBone;
			hr = pBone->GetLocal(& matBone);
			if(FAILED(hr)) {
				return hr;
			}

			hr = CalcUtil::Inverse(& matInvBone, & matBone);
			if(FAILED(hr)) {
				return hr;
			}

			hr = CalcUtil::Transform(& effectPosition, & effectPosition, & matInvBone);
			if(FAILED(hr)) {
				return hr;
			}

			MMD_VECTOR3 targetPosition;
			hr = CalcUtil::Transform(& targetPosition, & targetOriginalPosition, & matInvBone);
			if(FAILED(hr)) {
				return hr;
			}

			MMD_VECTOR3 diff;
			hr = CalcUtil::Subtract(& diff, & effectPosition, & targetPosition);
			if(FAILED(hr)) {
				return hr;
			}

			float dp = 0.0f;
			hr = CalcUtil::DotProduct(& dp, & diff, & diff);
			if(FAILED(hr)) {
				return hr;
			}

			if(dp < 0.0000001f) {
				return S_OK;
			}

			hr = CalcUtil::Normalize(& effectPosition, & effectPosition);
			if(FAILED(hr)) {
				return hr;
			}

			hr = CalcUtil::Normalize(& targetPosition, & targetPosition);
			if(FAILED(hr)) {
				return hr;
			}

			hr = CalcUtil::DotProduct(& dp, & effectPosition, & targetPosition);
			if(FAILED(hr)) {
				return hr;
			}

			float rotAngle = acosf(dp);
			if( 0.00000001f < fabsf( rotAngle ) )
			{
				if( rotAngle < -m_fact ) {
					rotAngle = -m_fact;
				} else if( m_fact < rotAngle ) {
					rotAngle = m_fact;
				}

				MMD_VECTOR3 rotAxis;
				hr = CalcUtil::CrossProduct(& rotAxis, & effectPosition, & targetPosition);
				if(FAILED(hr)) {
					return hr;
				}

				hr = CalcUtil::DotProduct(& dp, & rotAxis, & rotAxis);
				if(FAILED(hr)) {
					return hr;
				}

				if(dp < 0.0000001f) {
					continue;
				}

				hr = CalcUtil::Normalize(& rotAxis, & rotAxis);
				if(FAILED(hr)) {
					return hr;
				}

				MMD_VECTOR4 rotQuat;
				hr = CalcUtil::CreateAxis(& rotQuat, & rotAxis, rotAngle);
				if(FAILED(hr)) {
					return hr;
				}

				bool limitAngle = false;
				hr = pBone->IsIKLimitAngle(& limitAngle);
				if(FAILED(hr)) {
					return hr;
				}
				if(limitAngle) {
					hr = CalcUtil::LimitAngle(& rotQuat, & rotQuat);
					if(FAILED(hr)) {
						return hr;
					}
				}

				hr = CalcUtil::Normalize(& rotQuat, & rotQuat);
				if(FAILED(hr)) {
					return hr;
				}

				MMD_VECTOR4 destRotation;
				MMD_VECTOR3 destPosition;
				hr = pBone->GetVectors(& destPosition, & destRotation);
				if(FAILED(hr)) {
					return hr;
				}

				hr = CalcUtil::Multiply(& destRotation, & destRotation, & rotQuat);
				if(FAILED(hr)) {
					return hr;
				}

				hr = CalcUtil::Normalize(& destRotation, & destRotation);
				if(FAILED(hr)) {
					return hr;
				}

				hr = pBone->SetVectors(& destPosition, & destRotation);
				if(FAILED(hr)) {
					return hr;
				}

				for(int j = index; j >= 0; j --) {
					hr = m_bones[j]->UpdateMatrix();
					if(FAILED(hr)) {
						return hr;
					}
				}

				hr = m_pEffect->UpdateMatrix();
				if(FAILED(hr)) {
					return hr;
				}
			}
		}
	}

	return S_OK;
}

// \[gp̒l擾܂B
HRESULT MMDIK::GetSortValue(int* pNum) const
{
	if(pNum == NULL) {
		return E_POINTER;
	}
	if(m_ik.link.size() == 0) {
		return E_UNEXPECTED;
	}
	*pNum = m_ik.link[0];
	return S_OK;
}
