#include "StdAfx.h"
#include "MMDMorp.h"
#include "../CalcUtil.h"

// IuWFNgmr܂B
bool MMDMorp::Less(const Motion* pLeft, const Motion* pRight)
{
	if(pLeft == NULL || pRight == NULL) {
		return false;
	}
	int f1 = 0;
	int f2 = 0;
	HRESULT hr = pLeft->GetFrameNo(& f1);
	if(FAILED(hr)) {
		return false;
	}
	
	hr = pRight->GetFrameNo(& f2);
	if(FAILED(hr)) {
		return false;
	}

	if(f1 < f2) {
		return true;
	}
	return false;
}

// \z܂B
MMDMorp::MMDMorp(const PMD_MORP_RECORD* pMorp, const PMD_MORP_RECORD* pMorpBase, HRESULT* phr)
{
	if(pMorp == NULL) {
		*phr = E_POINTER;
		return;
	}

	m_morp = *pMorp;
	m_morp.mv.clear();
	if(pMorpBase == NULL) {
		for(vector<PMD_MORP_VERTEX_RECORD>::size_type i = 0; i < pMorp->mv.size(); i ++) {
			PMD_MORP_VERTEX_RECORD source = pMorp->mv[i];
			PMD_MORP_VERTEX_RECORD vert;
			vert.vec[0] = source.vec[0];
			vert.vec[1] = source.vec[1];
			vert.vec[2] = source.vec[2];
			vert.no = source.no;
			m_morp.mv.push_back(vert);
		}
	}else{
		for(vector<PMD_MORP_VERTEX_RECORD>::size_type i = 0; i < pMorp->mv.size(); i ++) {
			PMD_MORP_VERTEX_RECORD source1 = pMorp->mv[i];
			PMD_MORP_VERTEX_RECORD source2 = pMorpBase->mv[source1.no];
			PMD_MORP_VERTEX_RECORD vert;
			vert.vec[0] = (source1.vec[0] + source2.vec[0]);
			vert.vec[1] = (source1.vec[1] + source2.vec[1]);
			vert.vec[2] = (source1.vec[2] + source2.vec[2]);
			vert.no = source2.no;
			m_morp.mv.push_back(vert);
		}
	}
}

// j܂B
MMDMorp::~MMDMorp()
{
	ClearMotion();
}

// [Vj܂B
HRESULT MMDMorp::ClearMotion()
{
	for(vector<Motion>::size_type i = 0; i < m_motions.size(); i ++) {
		Motion* pMotion = m_motions[i];
		delete pMotion;
		m_motions[i] = NULL;
	}
	m_motions.clear();
	return S_OK;
}

// O擾܂B
HRESULT MMDMorp::GetName(string& name)
{
	char morpName[32];
	strncpy(morpName, m_morp.name, 20);
	morpName[20] = '\0';
	name = morpName;
	return S_OK;
}

// őt[擾܂B
HRESULT MMDMorp::GetMaxFrame(int* pFrameNo)
{
	if(pFrameNo == NULL) {
		return E_POINTER;
	}
	if(m_motions.size() == 0) {
		return S_FALSE;
	}
	return m_motions[m_motions.size() - 1]->GetFrameNo(pFrameNo);
}

// ǉΏۂǂ𔻒肵܂B
HRESULT MMDMorp::IsTarget(const VMD_MORP_RECORD* pMotion)
{
	if(pMotion == NULL) {
		return E_POINTER;
	}
	if(strncmp(m_morp.name, pMotion->name, 15) == 0) {
		return S_OK;
	}
	return S_FALSE;
}

// [Vǉ܂B
HRESULT MMDMorp::AddMotion(const VMD_MORP_RECORD* pMotion)
{
	if(pMotion == NULL) {
		return E_POINTER;
	}
	HRESULT hr = S_OK;
	Motion* pMot = new Motion(pMotion, & hr);
	if(FAILED(hr)) {
		delete pMot;
		pMot = NULL;
		return hr;
	}
	m_motions.push_back(pMot);
	return S_OK;
}

// [V܂B
HRESULT MMDMorp::PrepareMotion()
{
	std::sort(m_motions.begin(), m_motions.end(), MMDMorp::Less);
	return S_OK;
}

// [V𒸓_ɓKp܂B
HRESULT MMDMorp::ApplyMotion(float elapsedFrame, MMD_VERTEX_DESC** pOriginalVertexes)
{
	Motion* pBegin = NULL;
	Motion* pEnd = NULL;
	float offset = 0.0f;
	HRESULT hr = FindMotion(elapsedFrame, & pBegin, & pEnd, & offset);
	if(FAILED(hr)) {
		return hr;
	}
	if(hr == S_FALSE) {
		return S_OK;
	}

	if(pEnd == NULL) {
		float weight = 0.0f;
		hr = pBegin->GetRate(& weight);
		if(FAILED(hr)) {
			return hr;
		}

		return Blend(pOriginalVertexes, weight);
	}else{
		float weight1 = 0.0f;
		hr = pBegin->GetRate(& weight1);
		if(FAILED(hr)) {
			return hr;
		}

		float weight2 = 0.0f;
		hr = pEnd->GetRate(& weight2);
		if(FAILED(hr)) {
			return hr;
		}

		float weight = weight1 * (1.0f - offset) + weight2 * offset;

		return Blend(pOriginalVertexes, weight);
	}
	return S_OK;
}

// _ɓKp܂B
HRESULT MMDMorp::Set(MMD_VERTEX_DESC** pOriginalVertexes)
{
	for(vector<PMD_MORP_VERTEX_RECORD>::size_type i = 0; i < m_morp.mv.size(); i ++) {
		PMD_MORP_VERTEX_RECORD v = m_morp.mv[i];
		int jno = v.no;
		pOriginalVertexes[jno]->faced.point.x = v.vec[0];
		pOriginalVertexes[jno]->faced.point.y = v.vec[1];
		pOriginalVertexes[jno]->faced.point.z = v.vec[2];
	}
	return S_OK;
}

// _Blend[hœKp܂B
HRESULT MMDMorp::Blend(MMD_VERTEX_DESC** pOriginalVertexes, float weight)
{
	if(weight == 1.0f) {
		return Set(pOriginalVertexes);
	}

	MMD_VECTOR3 vec;
	for(vector<PMD_MORP_VERTEX_RECORD>::size_type i = 0; i < m_morp.mv.size(); i ++) {
		PMD_MORP_VERTEX_RECORD v = m_morp.mv[i];
		vec.x = v.vec[0];
		vec.y = v.vec[1];
		vec.z = v.vec[2];
		int jno = v.no;
		HRESULT hr = CalcUtil::Lerp(& pOriginalVertexes[jno]->faced.point, & pOriginalVertexes[jno]->faced.point, & vec, weight);
		if(FAILED(hr)) {
			return hr;
		}
	}
	return S_OK;
}

// [V܂B
HRESULT MMDMorp::FindMotion(float elapsedTime, Motion** ppBegin, Motion** ppEnd, float* pOffset)
{
	if(ppBegin == NULL || ppEnd == NULL || pOffset == NULL) {
		return E_POINTER;
	}
	if(m_motions.size() == 0) {
		return S_FALSE;
	}
	int fr = 0;
	HRESULT hr = m_motions[0]->GetFrameNo(& fr);
	if(elapsedTime <= fr) {
		*ppBegin = m_motions[0];
		return S_OK;
	}
	
	hr = m_motions[m_motions.size() - 1]->GetFrameNo(& fr);
	if(elapsedTime >= fr) {
		*ppBegin = m_motions[m_motions.size() - 1];
		return S_OK;
	}
	for(vector<Motion>::size_type i = 0; i < m_motions.size() - 1; i ++) {
		Motion* pSMotion = m_motions[i];
		Motion* pEMotion = m_motions[i + 1];

		int fr1 = 0;
		int fr2 = 0;
		hr = pSMotion->GetFrameNo(& fr1);
		if(FAILED(hr)) {
			return hr;
		}
		hr = pEMotion->GetFrameNo(& fr2);
		if(FAILED(hr)) {
			return hr;
		}

		if(fr1 <= elapsedTime && elapsedTime <= fr2) {
			*ppBegin = pSMotion;
			*ppEnd = pEMotion;
			*pOffset = (elapsedTime - fr1) / (fr2 - fr1);
			return S_OK;
		}
	}
	return S_FALSE;
}

// \z܂B
MMDMorp::Motion::Motion(const VMD_MORP_RECORD* pMotion, HRESULT* phr)
{
	if(pMotion == NULL) {
		*phr = E_POINTER;
		return;
	}

	m_motion = *pMotion;
}

// j܂B
MMDMorp::Motion::~Motion()
{
	;
}

// t[ԍ擾܂B
HRESULT MMDMorp::Motion::GetFrameNo(int* pFrameNo) const
{
	if(pFrameNo == NULL) {
		return E_POINTER;
	}
	*pFrameNo = m_motion.frame_no;
	return S_OK;
}

// Kp擾܂B
HRESULT MMDMorp::Motion::GetRate(float* pRate)
{
	if(pRate == NULL) {
		return E_POINTER;
	}
	*pRate = m_motion.factor;
	return S_OK;
}
